#ifndef CBITMAP_HPP
    #define CBITMAP_HPP

/**************************************************************************
*                                                                         *
* Name   : cBitmap.hpp               ( Implementation:              )     *
*                                                                         *
* Klasse(n) deklariert: BITMAP_C                                          *
* Objekt(e) definiert : -                                                 *
*                                                                         *
* Aufgabe :                                                               *
*                                                                         *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include "cGlobal.hpp"

// --- globale Konstanten ---------------------------------------------

//#define EMS_BANK_SIZE  0x10000L
// durch drei Teilbar, damit Punkte nicht ber Boundaries lappen knnen:
#define EMS_BANK_SIZE  0xF000L

// --- globale Typen --------------------------------------------------

typedef unsigned char LUT_COLOR_T;

typedef struct {
    BYTE R, G, B;
} RGB_COLOR_T;


typedef union {
    long int   L;
    struct { BYTE R, G, B;    LUT_COLOR_T LUT; } B;
    struct { RGB_COLOR_T RGB; LUT_COLOR_T LUT; } C;
} COLOR_U;

typedef RGB_COLOR_T LUT_T;             // Eintrag in der LUT

/*------------------------------------------------------------------------+
|                                                                         |
| Klasse : BITMAP_C                                                       |
|                                                                         |
+------------------------------------------------------------------------*/

class BITMAP_C {

// --- Variablen: -----------------------------------------------------

private:

protected:
    BOOL      Open;

    BOOL      IsRGB;
    int       NCols;
    int       NRows;
    int       LineLength;
    int       BytesPerPix;
    int       LUTSize;

    long      Size;                // Bildgre in Byte
    int       Banks;               // Anzahl der EMS Blcke ( 0..n * 64kB
		                   //     + 0..1 * 16k, 32 oder 48k )
    HANDLE    *HandleListP;        // * HANDLE [0..Banks-1]
    BYTE      **BankPListP;        // Ptr auf Bank 0..(Banks-1)

public:

// --- Methoden -------------------------------------------------------

public:                                // Konstruktor, Destruktor
    BITMAP_C ( BOOL anIsRGB,
	     int aBytesPerPix,
	     int aLUTSize,
	     int aCols,
	     int aRows );
    ~BITMAP_C ();

private:
    BOOL      initCard (BOOL     anIsRGB,  int aBytesPerPix,
                        int aLUTSize, int aCols,
                        int aRows);
    BYTE*     map (int aBank);

protected:

public:
    BOOL        open()        { return Open; };
    BOOL        isRGB()       { return IsRGB; };
    int         bytesPerPix() { return BytesPerPix; };
    int         getLUTSize()  { return LUTSize; };

    void        getLine   (int y, BYTE *BufP );
    void        putLine   (int y, BYTE *BufP );
    void        putBuffer (long Offset, BYTE *BufP, unsigned BufLen);
    void        getBuffer (long Offset, BYTE *BufP, unsigned BufLen);
    void        putPixLUT (int X, int Y, LUT_COLOR_T Color);
    LUT_COLOR_T getPixLUT (int X, int Y);
    void        putPixRGB (int X, int Y, RGB_COLOR_T &Color);
    RGB_COLOR_T getPixRGB (int X, int Y);

    void        clear();
    BOOL        writeBMP24 (char *aFName);

}; // class BITMAP_C


#endif CBITMAP_HPP

