#ifndef CCEFREAD_HPP
  #define CCEFREAD_HPP

/**************************************************************************
*                                                                         *
* Name   : cCefRead.hpp              ( Implementation:                )   *
*                                                                         *
* Klasse(n) definiert :  CEFREAD_C                                        *
* Objekt(e) deklariert:  -                                                *
*                                                                         *
* Abgeleitet von      :  -                                                *
*                                                                         *
* Aufgabe : Einlesen von .CEF Dateien. (Komprimiertes binrformat fr     *
*           farbige, 3D-Polylines )                                       *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include <stdio.h>

#include "cMath.hpp"
#include "cPolyLin.hpp"


// --- externe Klassen: -----------------------------------------------

// --- globale Konstanten ---------------------------------------------

#define MAX_CAMERAS 2

extern char *CefErrorMsg[];

// --- globale Typen --------------------------------------------------

typedef enum {
    CEF_OK         = 0,
    CEF_NOTFOUND,
    CEF_EOF,
    CEF_CORRUPTED
} CEF_STATUS_E;

typedef enum {                         // Transformationsarten fr Edgetransformation
    CEF_MATRIX_ABS44,
    CEF_MATRIX_REL44,
    CEF_TRANSL_ABS,
    CEF_TRANSL_REL,
    CEF_ROTAT_ABS,
    CEF_ROTAT_REL
} CEF_TRANSF_E;

typedef enum {                         // Blocktypen
    CEF_NO_KEY       = 0x0000,
    CEF_HEADER_KEY   = 0x0001,
    CEF_F_COMM_KEY   = 0x0101,
    CEF_E_TRANSF_KEY = 0x0002,
    CEF_E_DATA_KEY   = 0x0003,
    CEF_E_COMM_KEY   = 0x0103,
    CEF_CAMTRAN_KEY  = 0x0004,
    CEF_TABLECMD_KEY = 0x0005,
    CEF_EOF_KEY      = 0xFFFF
} CEF_KEY_E;

/*------------------------------------------------------------------------+
|                                                                         |
| Klasse : CEFREAD_C                                                      |
|                                                                         |
+------------------------------------------------------------------------*/

class CEFREAD_C {

// --- Variablen: -----------------------------------------------------

private:

protected:
    FILE         *CefIn;
    FILE         *DebugOut;
    CEF_STATUS_E Status;
    BOOL         DebugIt;
    BOOL         HeaderRead;
    BOOL         CamTranOk;

    MATRIX44_T   CamTran [MAX_CAMERAS];
    MATRIX44_T   CurTran;

    int          SimEdgeCnt;

public:
    long         ByteCnt;
    int          EdgeCnt;
    long         VertexTotal;
    long         FileSize;

// --- Methoden -------------------------------------------------------

public:                                // Konstruktor, Destruktor
    CEFREAD_C();
    virtual ~CEFREAD_C();

private:

protected:
    void debugMsg                      (char* Format, ...);
    void dumpMatrix44                  (MATRIX44_T aMatrix);
    void dumpVector4                   (VECTOR4_T aVector);

    BOOL readByte                      (BYTE  &aByte);
    BOOL readWord                      (WORD  &aWord);
    BOOL readDWord                     (DWORD &aDWord);
    BOOL readPfloat                    (Pfloat &aFloat);
    BOOL readVector                    (int N, Pfloat *aVectorP);
    BOOL readMatrix44                  (MATRIX44_T M);
    BOOL readString                    (char* StrP, int NMax);
    BOOL nextAlign();

    BOOL readHeader();
    BOOL readFileComment               (char *aStr, int NMax);
    BOOL readCameraTransformation();
    BOOL readEdgeTransformation();
    BOOL readPlData                    (VERTEXARRAY_C *aPLP);
    BOOL readEdgeComment               (char *aStr, int NMax);
    BOOL nextKey                       (WORD &aKey);

    CEF_KEY_E    peekBlock();
    BOOL         skipBlock();

    BOOL filter                        (Pptco3 &VP);

    BOOL simulateCylinder              (VERTEXARRAY_C *aPLP);
    BOOL simulatePlane                 (VERTEXARRAY_C *aPLP);
    BOOL simulatePlane2                (VERTEXARRAY_C *aPLP);
    BOOL simulateSphere                (VERTEXARRAY_C *aPLP);
    BOOL simulateCone                  (VERTEXARRAY_C *aPLP);
    BOOL simulateBox                   (VERTEXARRAY_C *aPLP);

public:
    BOOL         isOpen() { return (CefIn!=0); }
    CEF_STATUS_E status() { return Status; }

    void         init();
    CEF_STATUS_E open (char *FName, BOOL aDebugIt);
    void         close();
    BOOL         readNextPL (VERTEXARRAY_C *aPLP);

}; // class CEFREAD_C


#endif CCEFREAD_HPP
