#ifndef CGLOBAL_HPP
  #define CGLOBAL_HPP

/**************************************************************************
*                                                                         *
* Name   : cGlobal.hpp                 ( Implementation:                ) *
*                                                                         *
* Klasse(n) definiert : GLOBAL_C                                          *
* Objekt(e) deklariert: Global                                            *
*                                                                         *
* Abgeleitet von      : -                                                 *
*                                                                         *
* Aufgabe : Comp 3D, globale Funktionen, Variablen, Konstanten, ...       *
*                                                                         *
* Include windows.h first !                                               *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include <stdio.h>

// #include "cTypes.hpp"


// --- externe Klassen: -----------------------------------------------

class CEFREAD_C;
class GLUE_C;
class OPTIONS_C;
class REDUCE_C;
class STATUSWINDOW_C;
class TEXTURE_C;
class TRISET_C;
class MAINWINDOW_C;
class GEOMETRYOUT_C;

// --- globale Konstanten ---------------------------------------------

// --- globale Typen --------------------------------------------------


/*------------------------------------------------------------------------+
|                                                                         |
| Klasse : GLOBAL_C                                                       |
|                                                                         |
+------------------------------------------------------------------------*/

class GLOBAL_C {

// --- Variablen: -----------------------------------------------------

protected:
    HANDLE       Instance;
    HANDLE       PrevInstance;
    char         *CmdLine;
    int          CmdShow;

    int          WaitCnt;
    long         LastTicks;
    int          BusyCnt;

    FILE         *LogFile;
    char         LogFileName[80];
    FILE         *StatsDataFile;
    char         StatsFileName[80];

public:
    POINT        ScreenRes;
    POINT        CharSize;
    BOOL         CancelRequest;

// Klassen:
    // Die Kon- und Destruktoren der nachfolgenden Klassen sollten keine globalen Objekte
    // benutzen !!
    CEFREAD_C      *CefReadP;
    GLUE_C         *GlueP;             // Zuordnung Windowhandle <-> This-Pointer
    MAINWINDOW_C   *MainWndP;
    OPTIONS_C      *OptionsP;
    REDUCE_C       *ReduceP;
    STATUSWINDOW_C *StatWinP;
    TEXTURE_C      *TextureP;
    TRISET_C       *TrisetP;
    GEOMETRYOUT_C  *GeomOutP;

// --- Methoden -------------------------------------------------------

public:                                // Konstruktor, Destruktor
    GLOBAL_C ();
    virtual ~GLOBAL_C();

private:

protected:

public:
    HANDLE       getInstance()              { return Instance; }
    HWND         getMainWnd();

    BOOL         isInited()                 { return Instance!=0; }
    BOOL         waiting()                  { return WaitCnt>0; }
    BOOL         busy()                     { return BusyCnt>0; }

    BOOL         init  (HANDLE aInstance, HANDLE aPrevInstance,
                        LPSTR aCmdLine, int aCmdShow);
    void         reInit();

    void         yield();

    BOOL         openLog();
    BOOL         openStatsFile();

    // Fehler- und Debugmeldungen:

    void         beep();
    BOOL         logMsg   (char *Format, ...);   // -> log file
    BOOL         statsOut (char *Format, ...);   // -> Quattro pro input file
    void         debugMsg (char *Format, ...);   // -> debug window
    void         debugInfo(char *Format, ...);   // -> Messagebox (abschaltbar)
    void         info     (char *Format, ...);   // -> Messagebox
    void         warning  (char *Format, ...);   // -> Messagebox, beep
    void         fatal    (char *Format, ...);   // -> Messagebox: abort, ignore

    // Busyflag

    void         setBusy()                  { BusyCnt++; }
    void         clearBusy()                { BusyCnt--; }

    // Sanduhr cursor

    void         wait();
    void         noWait();

    // Stoppuhr:

    long         getTimeMark();
    float        elapsed (long LastTimeMark);
    float        elapsed();
    char *       elapsedStrHMSH (float aSecs);
    char *       elapsedStrHMS  (float aSecs);
    char *       elapsedStrMSH  (float aSecs);

    // Handler fr WM_COMPACTING

    void         wmCompacting   (WORD wParam);

}; // class GLOBAL_C

/*------------------------------------------------------------------------+
|                                                                         |
| Object public machen:                                                   |
|                                                                         |
+------------------------------------------------------------------------*/

extern GLOBAL_C Global;

#endif CGLOBAL_HPP
