#ifndef CMATH_HPP
  #define CMATH_HPP

/**************************************************************************
*                                                                         *
* Name   : cMath.hpp                 ( Implementation:                )   *
*                                                                         *
* Klasse(n) definiert :  -                                                *
*                                                                         *
* Aufgabe : Mathematische Typen, Konstanten und Routinen                  *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include <values.h>                    // MAXFLOAT, MINFLOAT, ...

#include "cTypes.hpp"

// --- globale Typen --------------------------------------------------

typedef Ppoint3 VECTOR3_T;
typedef Pfloat  VECTOR4_T  [4];
typedef Pfloat  MATRIX44_T [4][4];

// --- globale Konstanten ---------------------------------------------

extern MATRIX44_T Zero44;              // Alle Elemente 0
extern MATRIX44_T Identity44;          // Einheitsmatrix
extern Pfloat     Pi;
extern Pfloat     Sqr2;                // sqrt (2)
extern Pfloat     Sqr3;                // sqrt (3)
extern Pfloat     Ln2;                 // Log. naturalis (2)


// --- Macros ---------------------------------------------------------
// (vgl Glassner: Graphics GEMS s.631 )

#define MIN(a,b)          (((a)<(b))?(a):(b))
#define MAX(a,b)          (((a)>(b))?(a):(b))

// swappen ohne temporre Variable (alle Datentypen)
#define SWAP(a,b)         { a^=b; b^=a; a^=b; }

// linear interpolieren: a=0 -> low; a=1 -> high
#define LERP(a,low,high)  ((low)+(((high)-(low))*(a)))

// clippen: a<min -> min; a>max -> max; sonst a
#define CLAMP(a,min,max) ((a)<(min)?(min):(a)>(max)?(max):(a))

#define SORT2(a,b)       {if((a)>(b))SWAP((a),(b));else;}

#define SORT3(a,b,c)     { SORT2((a),(b)); SORT2((a),(c)); SORT2((b),(c)); }


// --- Funktionen -----------------------------------------------------

// Matrizenrechnung:

void      makeNull44     (MATRIX44_T aMatrix);
void      makeIdentity44 (MATRIX44_T aMatrix);
void      copyMatrix44   (MATRIX44_T Source, MATRIX44_T Dest);
void      multMatrix44   (MATRIX44_T Result, MATRIX44_T A, MATRIX44_T B);

// Vektorrechnung

BOOL      ptsEqual       (Ppoint3 &P1, Ppoint3 &P2);
BOOL      isNull3        (Ppoint3 &P);
Ppoint3   add3           (Ppoint3 &P1, Ppoint3 &P2);
Ppoint3   sub3           (Ppoint3 &P1, Ppoint3 &P2);
Pfloat    abs3           (Ppoint3 &P);
Ppoint3   crossProduct   (Ppoint3 &P1, Ppoint3 &P2);

// Geometrische tools:

void      transformPt44  (VECTOR4_T Result, MATRIX44_T M, VECTOR4_T Pt);
void      transformXY44  (Ppoint3 &Result, MATRIX44_T M, int X, int Y);

Pfloat    sqrDistance3   (Ppoint3 &P1, Ppoint3 &P2);
Pfloat    distance3      (Ppoint3 &P1, Ppoint3 &P2);
Pfloat    distancePtLn3  (Ppoint3 &P,  Ppoint3 &P1, Ppoint3 &P2);
Pfloat    distancePtPlane3 (Ppoint3 &P,  Ppoint3 &P1, Ppoint3 &P2, Ppoint3 &P3);
Pfloat    distancePtTri3 (Ppoint3 &P,  Ppoint3 &P1, Ppoint3 &P2, Ppoint3 &P3);
Pfloat    areaTri3       (Ppoint3 &P1, Ppoint3 &P2, Ppoint3 &P3);
Pfloat    volumeTetra    (Ppoint3 &P1, Ppoint3 &P2, Ppoint3 &P3, Ppoint3 &P4);

#endif CMATH_HPP
