#ifndef COPTIONS_HPP
  #define COPTIONS_HPP

/**************************************************************************
*                                                                         *
* Name   : cOptions.hpp              ( Implementation:                )   *
*                                                                         *
* Klasse(n) definiert :  OPTIONS_C                                        *
* Objekt(e) deklariert:  -                                                *
* Abgeleitet von      :  -                                                *
*                                                                         *
* Aufgabe : Enthlt Variablen fr die Comp3D Optionen, die Initialisierung*
*           und die Dialoge, um sie zu verndern.                         *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include "cFileBox.hpp"


// --- externe Klassen: -----------------------------------------------

class INIFILE_C;


// --- globale Konstanten: --------------------------------------------

extern char *ISTypeStr[];
extern char *MapMethodTypeStr[];
extern char *SimulationModeStr[];


// --- globale Typen: -------------------------------------------------

// ID's der Items der Optionsdialoge:

typedef enum {

    OPTS_PB_INFILE      = 1000,        // Input
    OPTS_ST_INFILE      = 1010,

    OPTS_CB_FILTER      = 2000,        // Filterung
    OPTS_PB_FILTER      = 2010,
    OPTS_ST_FILTER      = 2050,
    OPTS_ED_FACTR       = 2100,
    OPTS_ED_FACTG       = 2110,
    OPTS_ED_FACTB       = 2120,
    OPTS_ED_MINY        = 2200,
    OPTS_ED_MAXY        = 2210,

    OPTS_CB_REDUCPL     = 3000,        // Reduce Polyline
    OPTS_PB_REDUCPL     = 3010,
    OPTS_ST_REDUCPL     = 3015,
    OPTS_ED_MAXERRORPL  = 3020,
    OPTS_ED_MESHSIZEPL  = 3030,

    OPTS_CB_REDUCTRI    = 4000,        // Reduce Triset
    OPTS_PB_REDUCTRI    = 4010,
    OPTS_ST_REDUCTRI    = 4015,
    OPTS_ED_MAXERRORTRI = 4020,
    OPTS_ED_MESHSIZETRI = 4030,
    OPTS_ED_PROPORTION  = 4040,
    OPTS_ED_COMPACTTHR  = 4100,
    OPTS_ED_STOPTHR     = 4110,
    OPTS_CB_AVOIDDOUBLES= 4120,
    OPTS_ED_HITRATEIVAL = 4130,
    OPTS_CB_ANNEALING   = 4200,
    OPTS_ED_ANNEALFACT  = 4210,
    OPTS_ED_PROPFACT    = 4220,
    OPTS_ED_SAVESTART   = 4230,
    OPTS_ED_SAVEPERCENT = 4240,
    OPTS_CB_KEEPEDGE    = 4250,
    OPTS_ED_ERRORFACT   = 4270,
    OPTS_ED_VERTFACT    = 4280,
    OPTS_ED_AREAFACT    = 4290,
    OPTS_ED_TIMELIMIT   = 4300,


    OPTS_CB_TEXTURE     = 5000,        // Texturemapping
    OPTS_PB_TEXTURE     = 5010,
    OPTS_ST_TEXTURE     = 5015,
    OPTS_PB_TEXFILE     = 5020,
    OPTS_ST_TEXFILE     = 5030,
    OPTS_RB_CYLINDER    = 5100,
    OPTS_RB_SPHERE      = 5110,
    OPTS_RB_CENTROID    = 5200,
    OPTS_RB_ISN         = 5210,
    OPTS_ED_H           = 5300,
    OPTS_ED_U           = 5310,
    OPTS_ED_V           = 5320,
    OPTS_RB_TEXBMP      = 5400,
    OPTS_RB_TEXRLE      = 5410,
    OPTS_CB_INTERPOL    = 5500,

    OPTS_CB_GEOMETRY    = 6000,        // Output Geometrie
    OPTS_PB_GEOMETRY    = 6010,
    OPTS_ST_GEOMETRY    = 6015,
    OPTS_ST_GEOMFILE    = 6020,
    OPTS_PB_GEOMFILE    = 6030,
    OPTS_CB_GEOMSMALL   = 6100,
    OPTS_CB_GEOMCOMPR   = 6110,
    OPTS_RB_DATANONE    = 6200,
    OPTS_RB_DATARGB     = 6210,
    OPTS_RB_DATAUV      = 6220,
    OPTS_RB_FORMATDXF   = 6300,
    OPTS_RB_FORMATZOF   = 6310,
    OPTS_CB_SPLITSET    = 6400,
    OPTS_ED_VERTWEIGHT  = 6500,
    OPTS_ED_FACEWEIGHT  = 6510,
    OPTS_ED_MAXWEIGHT   = 6520,

    DBG_ED_UPDSTATSIVAL = 7000,        // Debugging
    DBG_ED_UPDHISTIVAL  = 7010,
    DBG_ED_LOGHISTIVAL  = 7020,
    DBG_CB_ENABINFOS    = 7030,
    DBG_CB_MARKFACES    = 7040,
    DBG_CB_CHKDOUBLES   = 7050,

    OPTS_ED_HASHSIZE    = 8000,        // sonstige
    OPTS_CB_CLOSEOBJ    = 8010

} OPTIONSDLGID_E;

// andere Aufzhlungstypen:

typedef enum {
    SURF_SPHERE,                       // Kugel
    SURF_CYLINDER                      // Zylinder
} ISTYPE_E;                            // Intermediate Surface Type

typedef enum {
    MAP_CENTROID,                      // Mittelpunkt
    MAP_ISN                            // Intermediate Surface Normal
} MAPMETHOD_E;                         // Mapping method type

typedef enum {
    TEX_BMP,
    TEX_RLE
} TEXFILETYPE_E;                       // Dateiformat fr Textur

typedef enum {
    GEOM_DATANONE,
    GEOM_DATARGB,
    GEOM_DATAUV
} GEOMFILETYPE_E;                       // Datenstruktur fr Geometrie

typedef enum {
    GEOM_DXF,
    GEOM_ZOF
} GEOMFILEFORMAT_E;                     // Dateiformat fr Geometrie

typedef enum {
    SIM_NONE,
    SIM_PLANE,
    SIM_PLANE2,
    SIM_SPHERE,
    SIM_CONE,
    SIM_CYLINDER,
    SIM_BOX
} SIMULATIONMODE_E;

/*------------------------------------------------------------------------+
|                                                                         |
| Klasse : OPTIONS_C                                                      |
|                                                                         |
+------------------------------------------------------------------------*/

class OPTIONS_C
{

// --- Variablen: -----------------------------------------------------

private:

protected:
    FILEBOX_C      FileBox;
    INIFILE_C      *IniFileP;

public:

    // Input:

    char           InFile   [80];      // File name

    // Filterung:

    BOOL           Filtering;
    float          FactR;              // Farbkorrektur
    float          FactG;
    float          FactB;
    float          ClipMinY;           // Y clipping
    float          ClipMaxY;

    // Reduce Polyline:

    BOOL           ReducePL;           // einzelne Polylines reduzieren ?
    float          MaxErrorPL;         // Maximale Abweichung der reduzierten Flche
    float          MeshSizePL;         // Minimaler Abstand zwischen zwei Punkten der Flche

    // Reduce Triset:

    BOOL           ReduceTri;          // Triset reduzieren ?
    float          MaxErrorTri;        // Maximale Abweichung der reduzierten Flche
    float          MeshSizeTri;        // Minimaler Abstand zwischen zwei Punkten der Flche
    float          WorstProportion;    // Max. Verhltnis von lngster Seite zur Hhe im Dreieck
    BOOL           KeepEdgePoints;     // Randpunkte nicht lschen ?

    long           CompactThresh;
    long           StopThresh;
    long           HitRateInterval;
    BOOL           AvoidDoubles;

    BOOL           Annealing;
    float          AnnealFactor;
    float          TimeLimitHours;
    float          VertexFactor;
    float          ErrorFactor;
    float          ProportionFactor;
    float          AreaFactor;
    float          AutoSavePercentage;
    float          AutoSaveStartTime;

    // Texturemapping:

    BOOL           TextureMapping;
    char           TexFile  [80];
    ISTYPE_E       Surface;
    MAPMETHOD_E    MapMethod;
    BOOL           Interpolate;        // Farben der Texturdreiecke Interpolieren ?
    int            U;                  // Maximale U Koordinate der Textur
    int            V;                  // Maximale V Koordinate der Textur
    float          H;                  // Hhe des Objektes (fr Berechnung von U und V)
    TEXFILETYPE_E  TexFileType;        // BMP oder RLE

    // Geometry Output:

    BOOL           WriteGeometry;
    char           GeomFile [80];
    GEOMFILETYPE_E GeomFileType;       // Zusatzdaten ? (keine, UV oder RGB)
    GEOMFILEFORMAT_E
                   GeomFileFormat;     // Ausgabeformat (.zof, .dxf)
    BOOL           GeomSmall;          // Folgende Nullen abschneiden ?
    BOOL           GeomCompress;
    BOOL           SplitTriset;
    float          VertexWeight;
    float          FaceWeight;
    float          MaxWeight;

    // Debugging:

    int            UpdateStatusInterval;
    BOOL           LogCefRead;
    BOOL           HistoryGraph;
    BOOL           HistoryIconic;
    int            UpdateHistoryInterval;
    BOOL           LogHistory;
    int            LogHistoryInterval;
    BOOL           LogSettings;
    BOOL           EnableInfos;
    BOOL           MarkFaces;
    BOOL           CheckDoubleEdges;
    BOOL           CrtioIconic;
    BOOL           StatusIconic;

    // Sonstige:

    SIMULATIONMODE_E
                   SimulationMode;     // Sollen die Eingabedaten simuliert werden ?
    long           VertHashTabSize;    // Gre der Vertex Hash-Tabelle
    BOOL           CloseObject;        // 1. und letzten Polygonzug schliessen ?


// --- Methoden -------------------------------------------------------

public:                                // Konstruktor, Destruktor
    OPTIONS_C();
    virtual ~OPTIONS_C();

private:

protected:

public:

    // Message server:

    void    newFName           (HWND hDlg, WORD wParam);
    int     updateConvReducDlg (HWND hDlg);
    int     updateInputDlg     (HWND hDlg);
    int     updateFilterDlg    (HWND hDlg);
    int     updateReducePLDlg  (HWND hDlg);
    int     updateReduceTriDlg (HWND hDlg);
    int     updateTextureDlg   (HWND hDlg);
    int     updateDebugDlg     (HWND hDlg);
    int     updateOutputDlg    (HWND hDlg);

    // allgemeine Funktionen:

    BOOL    initOptions();
    BOOL    makeFNames();
    BOOL    saveOptions();
    void    logSettings();

    // Dialogboxen:

    int     execConvReducDlg   (HWND hCaller);
    int     execInputDlg       (HWND hCaller);
    int     execFilterDlg      (HWND hCaller);
    int     execReducePLDlg    (HWND hCaller);
    int     execReduceTriDlg   (HWND hCaller);
    int     execTextureDlg     (HWND hCaller);
    int     execDebugDlg       (HWND hCaller);
    int     execOutputDlg      (HWND hCaller);

}; // class OPTIONS_C

#endif COPTIONS_HPP
