#ifndef CREDUCE_HPP
  #define CREDUCE_HPP

/**************************************************************************
*                                                                         *
* Name   : cReduce.hpp               ( Implementation:                )   *
*                                                                         *
* Klasse(n) definiert :  REDUCE_C                                         *
* Objekt(e) deklariert:  -                                                *
* Abgeleitet von      :  -                                                *
*                                                                         *
* Aufgabe :                                                               *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include "cTriset.hpp"
#include "cPolyLin.hpp"

// --- externe Klassen: -----------------------------------------------

// --- globale Konstanten: --------------------------------------------

// max. zulssige Polygongre:
#define MAXPGSIZE  20

// --- globale Typen: -------------------------------------------------

/*------------------------------------------------------------------------+
|                                                                         |
| Klasse : REDUCE_C                                                       |
|                                                                         |
+------------------------------------------------------------------------*/

class REDUCE_C
{

// --- Variablen: -----------------------------------------------------

private:

protected:
    INDEXARRAY_C TempIdxList;

public:
    long         PlKickCnt;            // Anzahl der gelschten Punkte (reducePL)

    float        MaxErrorPL;           // max. erlaubte Abweichung (ReducePL)
    float        MaxMeshSizePL;        // max. Kantenlnge der Dreiecke (ReducePL)
    float        MaxErrorTri;          // max. erlaubte Abweichung (ReduceTriset)
    float        MaxMeshSizeTri;       // max. Kantenlnge der Dreiecke (ReduceTriset)
    float        WorstProportion;      // Proportion des schmalsten Dreiecks

    long         Processed;
    long         CurVert;              // Index des aktuellen Vertex
    VERTEX_T     *CurVertP;            // Pointer auf aktuellen Vertex
    BOOL         CurVertValid;         // Ist der aktuelle Vertex gltig ?
    BOOL         CurVertVisible;       // Ist der aktuelle Vertex sichtbar ?
    INDEXARRAY_C PgIdx;                // Indizes der Nachbarpunkte von CurVert
    BOOL         PgClosed;             // Polygon geschlossen ?
    INDEXARRAY_C OldSurfIdx;           // Indizes der Nachbarfaces von CurVert
    float        CurEnergy;            // aktuelle Energie des Trisets
    float        SavedEnergy;          // Energie des gesicherten Trisets

    Pfloat       BestVolume;           // delta Volumen der besten Tesselierung
    Pfloat       BestMinArea;          // Flche der besten Tesselierung

    POLYGON_C    PG;                   // Kopie von
    FACEARRAY_C  BestSurf;             // Beste Tesselierung
    BOOL         BestSurfValid;        // ist die beste Tesselierung gltig ?
    FACEARRAY_C  CurSurf;              // Aktuelle Tesselierung

// --- Methoden -------------------------------------------------------

public:                                // Konstruktor, Destruktor
    REDUCE_C();
    virtual ~REDUCE_C();

protected:
    Pfloat greatestDistancePtLn      (Pptco3 *V, int A, int B);

    void   distributeHiddenVerts     (long         Center,
                                      INDEXARRAY_C *OldSurfIdx,
                                      FACEARRAY_C  *CurSurf,
                                      Pfloat &Error, Pfloat &AvgError);

    int    addFaceToCurSurf          (int I, BOOL &OneFaceInvalid, Pfloat &Volume);

    BOOL   hasBestSurfDoubleEdges();
    void   tesselateBest             (int N, Pfloat AddVolume, BOOL OneFaceInvalid);
    BOOL   findBestTesselation();
    BOOL   tesselateValid            (int N);
    BOOL   findValidTesselation();

    BOOL   findUndelTesselation      (long VertexIdx);
    float  energy();
    float  deltaEnergy();
    BOOL   metropolis                (float DeltaEnergy, float Temp);
    void   reduceDirect();
    void   reduceAnneal();

public:
    void   init();

    float  proportion                (Ppoint3 *V1P, Ppoint3 *V2P,Ppoint3 *V3P);
    BOOL   isTriValid                (long I1, long I2, long I3);

    // Reduktion eines Polygonzuges:

    int    reducePL                  (VERTEXARRAY_C &PLA);

    // Reduktion des Trisets:

    void   reduceTriset();

}; // class REDUCE_C


#endif CREDUCE_HPP
