#ifndef CTRISET_HPP
  #define CTRISET_HPP

/**************************************************************************
*                                                                         *
* Name   :                           ( Implementation:                )   *
*                                                                         *
* Klasse(n) definiert :  TRISET_C, FACEARRAY_C                            *
* Objekt(e) deklariert:  -                                                *
* Abgeleitet von      :  -                                                *
*                                                                         *
* Aufgabe :                                                               *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include "cPolyLin.hpp"

// --- externe Klassen: -----------------------------------------------

// --- globale Konstanten: --------------------------------------------

#define NOINDEX 0x8FFFFFFEL

static const int      MaxVertexArrays       = 500;
static const int      MaxFaceArrays         = 500;
static const int      MaxVertHashArrays     =  60;
static const int      InitialIndexArraySize =   0;
static const int      IndexArrayIncrement   =   5;

// --- globale Typen: -------------------------------------------------

typedef struct {

    Pptco3       PC;                   // Koordinaten und Farbe
    long         FI;                   // Index des Vater-Face

} VERTEX_T;


typedef struct {

    long         Index[3];             // Indizes der 3 Eckpunkte
    int          MaxN;                 // Gre von *HVP in Elementen
    int          N;                    // Anzahl von Elementen in *HVP
    long         *HVP;                 // Buffer fr Indizes der Hidden Vertizes

} FACE_S;


/*------------------------------------------------------------------------+
|                                                                         |
| Klasse : FACEARRAY_C                                                    |
|                                                                         |
|          Dynamisches Array von FACE_S. Die berlagerten Methoden        |
|          stellen sicher, da die Hidden Vertex Listen bercksichtigt    |
|          werden.                                                        |
|                                                                         |
+------------------------------------------------------------------------*/


class FACEARRAY_C :public DYNARRAY_C<FACE_S> {

// --- Variablen: -----------------------------------------------------

protected:
public:

// --- Methoden -------------------------------------------------------

public:                                // Konstruktor, Destruktor
    FACEARRAY_C  (int aMax, int aGrowBy);
    virtual ~FACEARRAY_C();

protected:

public:
    BOOL      newSize        (int nFaces);
    // warning '.. hides virtual function ..' abschalten:
    #pragma warn -hid
    BOOL      copyFrom       (FACEARRAY_C *SourceFAP);
    #pragma warn .hid

    BOOL      deleteLast();
    void      clear();

}; // class FACEARRAY_C



// --- globale Konstanten: --------------------------------------------

// ACHTUNG: Pfloat kann 4 oder 8 Bytes lang sein. ..ArraySize mu < 64k sein !

static const int      VertsPerArray        = 2048;
static const int      Log2VertsPerArray    = 11;           // 2^11 = 2048
static const long     VertexOffsMask       = 0x07FFL;      // 11 Bits gesetzt
static const unsigned VertexArraySize      = VertsPerArray * sizeof (VERTEX_T) + 15;

static const int      FacesPerArray        = 2048;
static const int      Log2FacesPerArray    = 11;           // 2^11 = 2048
static const long     FaceOffsMask         = 0x07FFL;      // 11 Bits gesetzt
static const unsigned FaceArraySize        = FacesPerArray * sizeof (FACE_S) + 15;

static const int      VertHashEntriesPerArray = 16000;
static const unsigned VertHashArraySize    = VertHashEntriesPerArray
                                             * sizeof (long) + 15;

/*------------------------------------------------------------------------+
|                                                                         |
| Klasse : TRISET_C                                                       |
|                                                                         |
+------------------------------------------------------------------------*/

class TRISET_C
{

// --- Variablen: -----------------------------------------------------

private:

protected:
    BOOL        Swapped;               // -> connectIntervall
    long        StartSearch;           // -> connectIntervall
    BOOL        Echo;
    char        BackupFSpec[80];

public:
    long        VertexCnt;
    long        VisibleVertexCnt;
    VERTEX_T    *VertArrayList [MaxVertexArrays];
    long        FaceCnt;
    long        ValidFaceCnt;
    FACE_S      *FaceArrayList [MaxFaceArrays];

    long        VertHashSize;
    long        *VertHashArrayList [MaxVertHashArrays];

    BOOL        MapTriangles;          // Flags zur Ablaufsteuerung
    BOOL        StoreTriangles;

// --- Methoden -------------------------------------------------------

public:                                // Konstruktor, Destruktor
    TRISET_C();
    virtual ~TRISET_C();

private:

protected:
    long    addVertex            (Pptco3 &Pt, long aStartSearch = 0L);
    void    add3Vertices         (Pptco3 &PC1, Pptco3 &PC2, Pptco3 &PC3,
                                 long   &I1,  long   &I2,  long   &I3,
                                 BOOL   &MatchedAll,
                                 long   aStartSearch);


    // Konvertierung der Edges (erste Stufe: Polylines):

    BOOL    nearestPlVertex      (Ppoint3 &RefVert, VERTEXARRAY_C &Pl,
                                  int Start, int End);
    BOOL    divideQuad           (Pptco3 &A1, Pptco3 &A2,
                                  Pptco3 &B1, Pptco3 &B2);
    void    connectIntervall     (VERTEXARRAY_C &Pl1, int A1, int A2,
                                  VERTEXARRAY_C &Pl2, int B1, int B2);


public:
    BOOL    echo                 (BOOL OnOff);
    BOOL    echo()                   { return Echo; }
    void    init();
    void    updateStatus();

    // elementare Methoden fr Zugriff auf BigArrays:

    VERTEX_T*getVertP            (long VertIdx);
    Ppoint3 *getVertPtP          (long VertIdx);
    Pptco3  *getVertPtCoP        (long VertIdx);
    FACE_S  *getFaceP            (long FaceIdx);
    long    *getFaceIdxP         (long FaceIdx);

    void    markFace             (long Fidx, BYTE R, BYTE G, BYTE B);
    BOOL    isFaceDeleted        (long FaceIdx);
    BOOL    existFace            (long I1, long I2, long I3);
    long    findOrientatedEdge   (long VIdx1, long VIdx2, long FirstFaceIdx = 0L);
    long    findDoubleEdge       (long VIdx1, long VIdx2, long VIdx3,
                                  long FirstFaceIdx = 0L);
    BOOL    existDoubleEdgeExcept(long VIdx1, long VIdx2, long VIdx3,
                                  INDEXARRAY_C *FaceIdxList);


    // Methoden zur Verwaltung des Trisets:

    BOOL    createVertHashTab    (long MaxEntries);
    void    destroyVertHashTab();

    void    compactFaces();
    void    clear();

    long    appendVertex         (Pptco3 &Pt);
    long    addVertexHash        (Pptco3 &V);

    long    appendFace           (long I1, long I2, long I3);
    void    delFace              (long FaceIdx);

    BOOL    addTriangle          (Pptco3 &V1, Pptco3 &V2, Pptco3 &V3);
    BOOL    processFace          (Pptco3 &V1, Pptco3 &V2, Pptco3 &V3);

    // Methoden fr Zugriff auf Hidden Vertex Listen:

    BOOL    addHV                (long FaceIdx, long VertIdx);
    BOOL    adjustHVSize         (long FaceIdx);
    BOOL    newHVSize            (long FaceIdx, int aVertexCnt);

    // Aus zwei Polylines einen Face-Streifen erzeugen:

    BOOL    addStripToTriset     (VERTEXARRAY_C &Pl1, VERTEXARRAY_C &Pl2);

    // Modifizierung des Trisets (wird beim Reduzieren benutzt):

    BOOL    getNeighbours        (long Center, INDEXARRAY_C *PGP,
                                               INDEXARRAY_C *FAP,
                                               BOOL &IsClosed);

    BOOL    replacePatch         (INDEXARRAY_C *OldFaceIdxP,
                                  FACEARRAY_C  *NewFaceP);

    // allgemeine Funktionen:

    BOOL    write();
    BOOL    backup();
    void    dumpVertexInfo       (long VertexIndex);
    void    dumpFaceInfo         (long FaceIndex);
    BOOL    isEqualOrientated();
    BOOL    integrityCheck();

}; // class TRISET_C

/*------------------------------------------------------------------------+
|                                                                         |
| Globale Funktionen fr Zugriff auf FACE_S und die damit verknpften     |
| Hidden Vertex Listen. Knnen sowohl fr                                 |
|     TRISET_C.FaceArrayList[].HVP... als auch fr                        |
|     FACEARRAY_C.Val[].HVP                                               |
| benutzt werden.                                                         |
|                                                                         |
+------------------------------------------------------------------------*/

BOOL    _isFaceDeleted        (long *VP);
void    _delFace              (FACE_S *FP);

BOOL    _addHV                (FACE_S *FP, long VertIdx);
void    _clearHV              (FACE_S *FP);
BOOL    _adjustHVSize         (FACE_S *FP);
BOOL    _newHVSize            (FACE_S *FP, int aVertexCnt);


#endif CTRISET_HPP
