/**************************************************************************
*                                                                         *
* Name   : cBitmap.cpp                ( Headerdatei:                 )    *
*                                                                         *
* Klasse(n) definiert : BITMAP_C                                          *
* Objekt(e) deklariert: -                                                 *
* Abgeleitet von      :                                                   *
*                                                                         *
* Aufgabe:                                                                *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include "cTypes.hpp"

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <mem.h>

#include "cBitmap.hpp"
#include "cStatWin.hpp"

// --- lokale Konstanten:----------------------------------------------

#define BM_DEBUG

// --- lokale Typen: --------------------------------------------------

// --- lokale Variablen: ----------------------------------------------


/*------------------------------------------------------------------------+
|                                                                         |
| lokale Funktionen                                                       |
|                                                                         |
+------------------------------------------------------------------------*/

/*------------------------------------------------------------------------+
|                                                                         |
| Protected:                                                              |
|                                                                         |
+------------------------------------------------------------------------*/

BOOL BITMAP_C::initCard (BOOL anIsRGB, int aBytesPerPix, int aLUTSize,
                          int aCols, int aRows)

{
    Open        = FALSE;


    IsRGB       = anIsRGB;
    BytesPerPix = aBytesPerPix;
    if (BytesPerPix>3) BytesPerPix = 3;

    LUTSize     = aLUTSize;
    NCols       = aCols;
    NRows       = aRows;

    LineLength  = NCols * BytesPerPix;
    Size        = (long)NRows * LineLength;
    Banks       = 0;
    HandleListP = NULL;
    BOOL Error = FALSE;            // Alles Ok ?

    Global.StatWinP->status ("creating bitmap ...");
    long EMS_Free = GetFreeSpace(0);

    if (Size <= EMS_Free) {

	unsigned Banks64 = (unsigned) (Size / EMS_BANK_SIZE);
	unsigned Banks16 = (unsigned) (((Size % EMS_BANK_SIZE) + 0x3FFF ) / 0x4000);

        Banks = Banks64;
        if (Banks16>0)
            Banks++;

        HandleListP = new HANDLE [Banks];
        BankPListP  = new BYTE*  [Banks];
        if (!HandleListP || !BankPListP) {
            return FALSE;
        }

        for (int I=0; I<Banks; I++) {   // Alle Handles lschen
            HandleListP [I] = 0;
        }

        for (I=0; I<Banks64; I++) {    // 64kB Blcke allocieren
            HandleListP [I] = GlobalAlloc (0, EMS_BANK_SIZE - 1);
            if (HandleListP [I]==NULL) {
                Error=TRUE;
                break;
            }
            Global.StatWinP->updateDlg();
        }

        if (Banks16>0)                 // Rest allocieren
            HandleListP [I] = GlobalAlloc (0, Banks16 * 0x4000 - 1);

        Global.StatWinP->updateDlg();

        for (I=0; I<Banks; I++) {
            if (HandleListP [I]==NULL)
                Error = TRUE;
            else
                BankPListP [I] = (BYTE*) GlobalLock ( HandleListP [I] );
        }
    }
    Global.StatWinP->status ("");
    Global.StatWinP->updateDlg();
    Open = !Error;
    return Open;
}

BYTE* BITMAP_C::map (int aBank)
{
    if ((aBank<0) || (aBank>=Banks)) {
        Global.warning ("BITMAP_C: illegal bank number: %i", aBank);
        return BankPListP [0];
    }
    return BankPListP [aBank];
}

/*------------------------------------------------------------------------+
|                                                                         |
| Public:                                                                 |
|                                                                         |
+------------------------------------------------------------------------*/

BITMAP_C::BITMAP_C (BOOL anIsRGB, int aBytesPerPix, int aLUTSize, int aCols, int aRows)
{
    if ( ! initCard ( anIsRGB, aBytesPerPix, aLUTSize, aCols, aRows ) )
        Global.warning ("BITMAP_C: Error initializing Bitmap");
}

BITMAP_C::~BITMAP_C ()
{
    if (HandleListP && Open) {
        for (int I=0; I<Banks; I++) {
            GlobalUnlock (HandleListP [I]);
            if ( GlobalFree (HandleListP [I]) )
                Global.warning ("Error disposing Bitmap!");
            HandleListP [I]=0;
        }
        delete HandleListP;
    }
    if (BankPListP) delete BankPListP;
}

void BITMAP_C::getLine (int Y, BYTE *BufP)
{
    long OffsY    = (long)Y * LineLength;
    int  Bank     = int (OffsY / EMS_BANK_SIZE);
    long BankOffs = OffsY % EMS_BANK_SIZE;

    BYTE *BaseP = map (Bank); // Bank einblenden

    if (BaseP==NULL) return;

    if (BankOffs + LineLength > EMS_BANK_SIZE) {    // Zeile ber zwei Bnke ?

	unsigned Rest = (unsigned) (BankOffs + LineLength - EMS_BANK_SIZE);

        #ifdef BM_DEBUG
            if ( ( BankOffs+LineLength-Rest > GlobalSize (HandleListP [Bank]) ) ) {
                Global.warning ("BITMAP_C -> getLine : range violation !\n");
                return;
            }
        #endif
	movmem (BaseP + unsigned (BankOffs), BufP, LineLength-Rest);

        BaseP = map (Bank + 1);

        if (BaseP==NULL) return;

	movmem (BaseP, BufP + LineLength - Rest, Rest);
        }
    else {
        #ifdef BM_DEBUG
            if ( ( BankOffs+LineLength-1 > GlobalSize (HandleListP [Bank]) ) ) {
                Global.warning ("BITMAP_C -> getLine: range violation !\n");
                return;
            }
        #endif
	movmem (BaseP + unsigned(BankOffs), BufP, LineLength);
    }
}

void BITMAP_C::putLine (int Y, BYTE *BufP)
{
    unsigned long OffsY    = (long)Y * LineLength;
    unsigned      Bank     = (unsigned) (OffsY / EMS_BANK_SIZE);
    unsigned long BankOffs = OffsY % EMS_BANK_SIZE;

    BYTE *BaseP = map (Bank); // Bank einblenden

    if (BaseP==NULL) return;

    if (BankOffs + LineLength > EMS_BANK_SIZE) {    // Zeile ber zwei Bnke ?

	unsigned Rest = unsigned (BankOffs + LineLength - EMS_BANK_SIZE);

        #ifdef BM_DEBUG
            if ( ( BankOffs+LineLength-Rest > GlobalSize (HandleListP [Bank]) ) ) {
                Global.warning ("BITMAP_C -> putLine : range violation !\n");
                return;
            }
        #endif
	movmem (BufP, BaseP + unsigned(BankOffs), LineLength-Rest);

        BaseP = map (Bank + 1);

        if (BaseP==NULL) return;

        movmem (BufP + LineLength - Rest, BaseP, Rest);
        }
    else {
        #ifdef BM_DEBUG
            if ( ( BankOffs+LineLength-1 > GlobalSize (HandleListP [Bank]) ) ) {
                Global.warning ("BITMAP_C -> putLine : range violation !\n");
                return;
            }
        #endif
	movmem (BufP, BaseP + unsigned(BankOffs), LineLength);
    }
}

void BITMAP_C::putBuffer (long Offset, BYTE *BufP, unsigned BufLen)
{
    if (!Open) return;

    unsigned      Bank     = (unsigned) (Offset / EMS_BANK_SIZE);
    unsigned long BankOffs = Offset % EMS_BANK_SIZE;

    BYTE *BaseP = map (Bank); // Bank einblenden

    if (BaseP==NULL) return;

    if (BankOffs + BufLen > EMS_BANK_SIZE) {    // Block ueber zwei Baenke ?

	unsigned Rest = unsigned (BankOffs + BufLen - EMS_BANK_SIZE);

        #ifdef BM_DEBUG
            if ( ( BankOffs+BufLen-Rest > GlobalSize (HandleListP [Bank]) ) ) {
                Global.warning ("BITMAP_C -> putLine : range violation !\n");
                return;
            }
        #endif
	movmem (BufP, BaseP + unsigned(BankOffs), BufLen-Rest);

        BaseP = map (Bank + 1);

        if (BaseP==NULL) return;

        movmem (BufP + BufLen - Rest, BaseP, Rest);
        }
    else {
        #ifdef BM_DEBUG
            if ( ( BankOffs+BufLen-1 > GlobalSize (HandleListP [Bank]) ) ) {
                Global.warning ("BITMAP_C -> putLine : range violation !\n");
                return;
            }
        #endif
	movmem (BufP, BaseP + unsigned(BankOffs), BufLen);
    }
}

LUT_COLOR_T BITMAP_C::getPixLUT (int X, int Y)
{
    if (!Open) return 0;

    unsigned long Offs     = (long)Y * LineLength + X * BytesPerPix;
    unsigned      Bank     = unsigned (Offs / EMS_BANK_SIZE);
    unsigned long BankOffs = Offs % EMS_BANK_SIZE;

    unsigned char *BaseP = map (Bank); // Bank einblenden

    if (BaseP==NULL) return 0;

    #ifdef BM_DEBUG
        if ( ( BankOffs > GlobalSize (HandleListP [Bank]) ) ) {
            Global.warning ("BITMAP_C -> getPixLUT : range violation !\n");
            return 0;
        }
    #endif

    return *(BaseP + unsigned(BankOffs));
}

void BITMAP_C::putPixLUT (int X, int Y, LUT_COLOR_T Color)
{
    if (!Open) return;

    unsigned long Offs     = (long)Y * LineLength + X * BytesPerPix;
    unsigned      Bank     = unsigned (Offs / EMS_BANK_SIZE);
    unsigned long BankOffs = Offs % EMS_BANK_SIZE;

    unsigned char *BaseP = map (Bank); // Bank einblenden

    if (BaseP==NULL) return;

    #ifdef BM_DEBUG
        if ( ( BankOffs > GlobalSize (HandleListP [Bank]) ) ) {
            Global.warning ("BITMAP_C -> putPixLUT : range violation !\n");
            return;
        }
    #endif

    *(BaseP +  unsigned (BankOffs)) = Color;
}

void BITMAP_C::putPixRGB (int X, int Y, RGB_COLOR_T &Color)
{
    if (!Open) return;

    long          Offs     = (long)Y * LineLength + X * BytesPerPix;
    unsigned      Bank     = unsigned (Offs / EMS_BANK_SIZE);
    long          BankOffs = Offs % EMS_BANK_SIZE;

    BYTE *BaseP = map (Bank); // Bank einblenden

    if (BaseP==NULL) return;

    if ( ( BankOffs+3 > GlobalSize (HandleListP [Bank]) ) ) {
        Global.warning ("BITMAP_C -> putPixRGB (%i, %i): range violation", X, Y);
        return;
    }

    #ifdef BM_DEBUG
        if (BankOffs>(0xF000UL-3))
            Global.warning ("BITMAP_C -> putPixRGB (%i, %i): range violation", X, Y);
    #endif

    // *** ACHTUNG: Wenn Records an Wortgrenzen ausgerichtet werden, hat RGB_COLOR_T
    // die Gre 4 Byte! Dann wird evtl. ber die Segmentgrenze geschrieben
    // Deswegen funktioniert das folgende nicht:
    //     *((RGB_COLOR_T*)(BaseP + unsigned(BankOffs))) = Color;

    BYTE *BP = BaseP + unsigned(BankOffs);
    BP[0] = Color.R; BP[1] = Color.G; BP[2] = Color.B;
}

static RGB_COLOR_T DummyRGB = { 0, 0, 0 };

RGB_COLOR_T BITMAP_C::getPixRGB (int X, int Y)
{
    if (!Open) return DummyRGB;

    unsigned long Offs     = (long)Y * LineLength + X * BytesPerPix;
    unsigned      Bank     = unsigned (Offs / EMS_BANK_SIZE);
    unsigned long BankOffs = Offs % EMS_BANK_SIZE;

    BYTE *BaseP = map (Bank); // Bank einblenden

    if (BaseP==NULL) return DummyRGB;

    if ( ( BankOffs+3 > GlobalSize (HandleListP [Bank]) ) ) {
        Global.warning ("BITMAP_C -> getPixRGB (%i,%i): range violation !", X, Y);
        return DummyRGB;
    }

    // *** ACHTUNG: Wenn Records an Wortgrenzen ausgerichtet werden, hat RGB_COLOR_T
    // die Gre 4 Byte! Dann wird evtl. ber die Segmentgrenze gelesen
    // Deswegen funktioniert das folgende nicht:
    //     return *((RGB_COLOR_T*)(BaseP + unsigned(BankOffs)));

    BYTE        *BP = BaseP + unsigned(BankOffs);
    RGB_COLOR_T RGB;
    RGB.R = BP[0]; RGB.G = BP[1];RGB.B = BP[2];
    return RGB;
}

void BITMAP_C::clear()
{
    BYTE *BufP = (BYTE*)malloc (LineLength);
    if (!BufP) return;

    memset (BufP, 0, LineLength);
    for (int I=0; I<NRows; I++)
        putLine (I, BufP);

    free (BufP);
}


BOOL BITMAP_C::writeBMP24 (char *FName)
{
    BITMAPFILEHEADER BF;
    BITMAPINFOHEADER BI;
    BOOL             Error = FALSE;

    BYTE *BufP = (BYTE*)malloc (LineLength+15);
    if (!BufP) return FALSE;

    FILE *F = fopen (FName, "wb");
    if (!F) return FALSE;

    BF.bfType      = 0x4D42;
    BF.bfSize      = (unsigned long)sizeof (BF) +
                     (unsigned long)sizeof (BI) +
                     (unsigned long)NRows*LineLength;
    BF.bfReserved1 = 0;
    BF.bfReserved2 = 0;
    BF.bfOffBits   = (long)sizeof (BF) + sizeof (BI);

    BI.biSize          = sizeof (BI);
    BI.biWidth         = NCols;
    BI.biHeight        = NRows;
    BI.biPlanes        = 1;
    BI.biBitCount      = 24;
    BI.biCompression   = 0L;
    BI.biSizeImage     = 0L;
    BI.biXPelsPerMeter = 0L;
    BI.biYPelsPerMeter = 0L;
    BI.biClrUsed       = 0L;
    BI.biClrImportant  = 0L;

    if ( ! fwrite (&BF, sizeof (BF), 1, F) ) Error=TRUE;
    if ( ! fwrite (&BI, sizeof (BI), 1, F) ) Error=TRUE;

    for (int I=0; I<NRows; I++) {
        getLine (I, BufP);
        for (int X=0; X<NCols; X++) {                      // RGB -> BGR
            int  J=X*3;
            BYTE T=BufP [J]; BufP [J]=BufP[J+2]; BufP[J+2]=T;
        }
        if ( !fwrite (BufP, LineLength, 1, F) ) {
            Error = TRUE;
            break;
        }
    }

    fclose (F);
    free (BufP);
    return !Error;
}

