/**************************************************************************
*                                                                         *
* Name   : cCefRead.cpp                ( Headerdatei:                 )   *
*                                                                         *
* Klasse(n) definiert : CEFREAD_C                                         *
* Objekt(e) deklariert: -                                                 *
*                                                                         *
* Abgeleitet von      : -                                                 *
*                                                                         *
* Aufgabe:                                                                *
*                                                                         *
**************************************************************************/

// --- Defines: -------------------------------------------------------

// low level debugging (jedes gelesene Byte in Debugdatei ausgeben) ?
// #define CEF_PRINTBYTES

// weden in der Eingabedatei doubles oder floats benutzt ?
#define CEF_INPUTDOUBLE

// --- Includes: ------------------------------------------------------

#include "cTypes.hpp"

#include <stdarg.h>
#include <stdio.h>
#include <dir.h>
#include <string.h>
#include <mem.h>

#include <stdlib.h>      // Simulation mode
#include <math.h>        // Simulation mode

#include "cCefRead.hpp"
#include "cGlobal.hpp"
#include "cOptions.hpp"
#include "cPolyLin.hpp"
#include "cStatWin.hpp"
#include "cTools.hpp"


// --- lokale Konstanten:----------------------------------------------

static const ESCAPE_NIBBLE =  8;
static const ALIGNMENT     = 16;

// --- public Konstanten:----------------------------------------------

char *CefErrorMsg[] = {
    "Ok",
    "File not found",
    "EoF reached",
    "File corrupted"
};


// --- lokale Typen: --------------------------------------------------

// --- lokale Variablen: ----------------------------------------------


/*------------------------------------------------------------------------+
|                                                                         |
| lokale Funktionen                                                       |
|                                                                         |
+------------------------------------------------------------------------*/

/*------------------------------------------------------------------------+
|                                                                         |
| protected Funtionen                                                     |
|                                                                         |
+------------------------------------------------------------------------*/

// statc char MsgBuffer [150];

void CEFREAD_C::debugMsg (char* Format, ...)
{
    va_list ArgP;

    if (DebugIt) {
        va_start (ArgP, Format);
        vfprintf (DebugOut, Format, ArgP);
        va_end (ArgP);
    }
}

void CEFREAD_C::dumpMatrix44 (MATRIX44_T M)
{
    if (!DebugIt) return;
    for (int I=0; I<4; I++)
        dumpVector4 ( M[I] );
    debugMsg ("\n");
}

void CEFREAD_C::dumpVector4 (VECTOR4_T V)
{
    if (!DebugIt) return;
    debugMsg ("%10.5f %10.5f%10.5f%10.5f\n", V[0], V[1], V[2], V[3]);
}

BOOL CEFREAD_C::readByte (BYTE &aByte)
{
    int I;
    I = getc (CefIn);
    if (I==EOF) {
        Status=CEF_EOF;
        return FALSE;
    }
    aByte = (BYTE) I;
    Status=CEF_OK;
    ByteCnt++;
    #ifdef CEF_PRINTBYTES
        debugMsg ("%02X ", aByte);
    #endif
    return TRUE;
}

BOOL CEFREAD_C::readWord (WORD  &aWord)
{
    aWord = (WORD) getw (CefIn);
    if ( ferror (CefIn) ) {
        Status=CEF_EOF;
        return FALSE;
    }
    Status=CEF_OK;
    ByteCnt += sizeof(WORD);
    #ifdef CEF_PRINTBYTES
        debugMsg ("%04X ", aWord);
    #endif
    return TRUE;
}

BOOL CEFREAD_C::readDWord  (DWORD &aDWord)
{
    WORD Low, High;

    aDWord = 0;

    if ( ! readWord (Low) )  return FALSE;
    if ( ! readWord (High) ) return FALSE;
    aDWord = ( (DWORD)High<<16) | Low;
    #ifdef CEF_PRINTBYTES
        debugMsg ("%08lX ", aDWord);
    #endif
    return TRUE;
}

BOOL CEFREAD_C::readPfloat (Pfloat &aFloat)
{
    BYTE C;

    #ifdef CEF_INPUTDOUBLE

        double D;

        for (int I=0; I<sizeof (double); I++) {
            if ( ! readByte (C) ) return FALSE;      // Schleife vorzeitig verlassen
            *((char*)(&D)+I) = C;
        }
        aFloat = D;
        #ifdef CEF_PRINTBYTES
            debugMsg ("%10.5lf ", aFloat);
        #endif

    #else

        for (int I=0; I<sizeof (Pfloat); I++) {
            if ( ! readByte (C) ) return FALSE;      // Schleife vorzeitig verlassen
            *((char*)(&aFloat)+I) = C;
        }
        #ifdef CEF_PRINTBYTES
            debugMsg ("%10.5f ", aFloat);
        #endif

    #endif
    return TRUE;
}

BOOL CEFREAD_C::readVector (int N, Pfloat *aVectorP)
{
    for (int I=0; I<N; I++)
        if ( ! readPfloat (aVectorP [I]) ) return FALSE;
    dumpVector4 (aVectorP);
    return TRUE;
}

BOOL CEFREAD_C::readMatrix44 (MATRIX44_T M)
{
    for (int I=0; I<4; I++)
        for (int J=0; J<4; J++)
            if ( ! readPfloat (M [I][J]) ) return FALSE;
    dumpMatrix44 (M);
    return TRUE;
}

BOOL CEFREAD_C::readString (char* StrP, int NMax)
{
    BYTE C;
    int  Cnt = 0;

    while (Cnt<NMax) {
        if ( !readByte (C) ) {
            *StrP = '\0';
            return FALSE;
        }
        *StrP = C;
        if (C=='\0') break;
        StrP++;
        Cnt++;
    };
    return TRUE;
}

BOOL CEFREAD_C::nextAlign()
{
    BYTE Dummy;

    while (ByteCnt % ALIGNMENT)        // lesen bis zur nchsten Paragraphengrenze
        if ( ! readByte (Dummy) )
            return FALSE;
    return TRUE;
}

BOOL CEFREAD_C::readHeader()
{
    DWORD Stamp;
    WORD  Version;
    WORD  Key;
    DWORD BlockSize;
    WORD  Rows;
    WORD  Columns;

    if ( ! readDWord (Stamp) )     return FALSE;
    if ( ! readWord  (Version) )   return FALSE;
    if ( ! readWord  (Key) )       return FALSE;
    if ( ! readDWord (BlockSize) ) return FALSE;
    if ( ! readWord  (Rows) )      return FALSE;
    if ( ! readWord  (Columns) )   return FALSE;

    if (Stamp != 0x1A464543L)    { Status = CEF_CORRUPTED; return FALSE; }
    if (Key   != CEF_HEADER_KEY) { Status = CEF_CORRUPTED; return FALSE; }

    debugMsg ("\nHeader (Version %04X, Paragraphs: %lu) Rows:%u, Columns: %u\n",
              Version, BlockSize, Rows, Columns);

    return nextAlign();
}

BOOL CEFREAD_C::readFileComment (char* StrP, int NMax)
{
    if (!readString (StrP, NMax)) return FALSE;

    debugMsg ("\nFile comment:\"%s\"\n", StrP);
    return TRUE;
}

BOOL CEFREAD_C::readEdgeComment (char* StrP, int NMax)
{
    if (!readString (StrP, NMax)) return FALSE;

    debugMsg ("\nEdge comment:\"%s\"\n", StrP);
    return TRUE;
}

BOOL CEFREAD_C::readCameraTransformation()
{
    WORD Cnt;

    readWord (Cnt);

    debugMsg ("\nReading %u camera transformations\n", Cnt);

    if (Cnt>MAX_CAMERAS) return FALSE;

    for (int I=0; I<Cnt; I++)
        if ( !readMatrix44 ( CamTran[I] ) )
            return FALSE;
    CamTranOk = TRUE;
    return TRUE;
}

BOOL CEFREAD_C::readEdgeTransformation()
{
    WORD         Type;
    MATRIX44_T   M44;

    if ( ! readWord (Type) ) return FALSE;

    debugMsg ("\nReading edge transformation type %04X\n", Type);

    switch ( (CEF_TRANSF_E) Type ) {

        case CEF_MATRIX_ABS44:
        case CEF_MATRIX_REL44:
            readMatrix44 (M44);
            break;
        case CEF_TRANSL_ABS:
        case CEF_TRANSL_REL:
        case CEF_ROTAT_ABS:
        case CEF_ROTAT_REL:
    default:
        debugMsg ("\nUnknown transformation type (%4X)\n", Type);
        return FALSE;
    }
    if (Status) return FALSE;

    switch (Type) {
        case CEF_MATRIX_ABS44:
            multMatrix44 (CurTran, M44, CamTran[0]);          // T' = Tpos * Tcam
            break;
        case CEF_MATRIX_REL44:
            multMatrix44 (CurTran, M44, CurTran);             // T' = Tpos * T
            break;
        case CEF_TRANSL_ABS:
        case CEF_TRANSL_REL:
        case CEF_ROTAT_ABS:
        case CEF_ROTAT_REL:
    default:
        debugMsg ("\nUnknown transformation type (%4X)\n", Type);
    }
    debugMsg ("CurTran:\n");
    dumpMatrix44 (CurTran);
    return TRUE;
}

BOOL CEFREAD_C::readPlData (VERTEXARRAY_C *PlP)

// Eine Polyline einlesen. Fr jeden Punkt wird die Filterfunktion aufgerufen.

{
    signed char  dX, dR, dG, dB;
    WORD         W;
    BYTE         B;
    int          CurX, CurY;
    Pcobundl     CurC;
    Pptco3       Vert;

    WORD Cnt    = 0;
    WORD EscCnt = 0;
    BOOL Abs    = TRUE;

    if (PlP==NULL) return FALSE;

    do {
        if (Abs) {

            readWord (W); CurX = (int)W;
            readWord (W); CurY = (int)W;
            readByte (B); CurC.x = B;
            readByte (B); CurC.y = B;
            readByte (B); CurC.z = B;

            debugMsg ("abs: ");

            Abs = FALSE;               // Umschalten auf relative (komprimierte) Speicherung
            }
        else {

            readByte (B);

            if ( (B >> 4) == ESCAPE_NIBBLE) {

                if ( (B & 0x0F) == 0x0F ) {

                    break;             // Ende des Datensatzes
                    }
                else {

                    Abs=TRUE;          // Umschalten auf absolute Speicherung
                    EscCnt++;
                    continue;
                }
                }
            else {

                dX = (signed char) (B>>4);
                if (dX>7) dX = dX - 16;

                dR = (signed char)(B & 0x0F);
                if (dR>7) dR = dR - 16;

                readByte (B);

                dG = (signed char) (B>>4);
                if (dG>7) dG = dG - 16;

                dB = (signed char) (B & 0x0F);
                if (dB>7) dB = dB - 16;

                CurX   += dX;
                CurY   += 1;
                CurC.x += dR;
                CurC.y += dG;
                CurC.z += dB;

                debugMsg ("rel: ");
            }
        }

        if (Status) break;

        // Koordinaten transformieren, Farben normalisieren:

        transformXY44  (Vert.point, CurTran, CurX, CurY);
        Vert.colour.x = (Pfloat)CurC.x/255;
        Vert.colour.y = (Pfloat)CurC.y/255;
        Vert.colour.z = (Pfloat)CurC.z/255;

        debugMsg ("[%3u, %3u] (%u,%u,%u) -> [%8.4f,%8.4f,%8.4f] ",
                  CurX, CurY, (unsigned)CurC.x, (unsigned)CurC.y, (unsigned)CurC.z,
                  Vert.point.x, Vert.point.y,Vert.point.z );

        // ggf. Filterung durchfhren:

        if (filter (Vert)) {

            PlP->append (Vert);

            debugMsg ("(%4.2f, %4.2f, %4.2f)\n",
                      Vert.colour.x, Vert.colour.y, Vert.colour.z);
            Cnt++;
            }
        else {
            debugMsg ("-> FILTERED !\n");
        }

    } while ( TRUE );

    EdgeCnt++;
    VertexTotal += Cnt;

    Global.StatWinP->InVerts = VertexTotal;
    Global.StatWinP->InPls   = EdgeCnt;

    debugMsg ("\nPolyline #%i had %u vertices (%u Escapes).\n\n", EdgeCnt, Cnt, EscCnt);

    return (Status==CEF_OK);
}


BOOL CEFREAD_C::nextKey ( WORD &aKey)
{
    if (!nextAlign()) return FALSE;
    return readWord (aKey);
}

CEF_KEY_E CEFREAD_C::peekBlock()
{
    WORD Key;
    long CurPos;

    CurPos = ftell (CefIn);

    if (!nextAlign()) return CEF_NO_KEY;

    if (!readWord (Key)) Key = CEF_NO_KEY;

    fseek (CefIn, CurPos, SEEK_SET);
    ByteCnt -= sizeof (WORD);

    return (CEF_KEY_E)Key;
}

BOOL CEFREAD_C::skipBlock()
{
    DWORD Paragraphs;
    WORD  Key;
    BYTE  B;

    if (!readWord (Key))         return FALSE;
    if (!readDWord (Paragraphs)) return FALSE;
    if (Paragraphs<1)            return FALSE;

    for (long I=0L; I<(Paragraphs*16-6); I++)
        if (!readByte (B)) return FALSE;

    return TRUE;
}

/*------------------------------------------------------------------------+
|                                                                         |
| Filterung der Eingangsdaten                                             |
|                                                                         |
+------------------------------------------------------------------------*/

BOOL CEFREAD_C::filter (Pptco3 &Vert)
// Vertexkoordinaten und Farben werden ggf. verndert.
// Gibt FALSE zurck, wenn der Punkt gelscht werden soll.
{
    OPTIONS_C *OP = Global.OptionsP;

    if (!OP->Filtering) return TRUE;

    // clipping:

    Pfloat Y = Vert.point.y;

    if ( (Y < OP->ClipMinY) || (Y > OP->ClipMaxY) ) return FALSE;

    // color filtering:

    Vert.colour.x = CLAMP (Vert.colour.x * OP->FactR, 0.0, 1.0);
    Vert.colour.y = CLAMP (Vert.colour.y * OP->FactG, 0.0, 1.0);
    Vert.colour.z = CLAMP (Vert.colour.z * OP->FactB, 0.0, 1.0);

    return TRUE;
}

/*------------------------------------------------------------------------+
|                                                                         |
| Simulation der Eingangsdaten                                            |
|                                                                         |
+------------------------------------------------------------------------*/


BOOL CEFREAD_C::simulateCylinder (VERTEXARRAY_C *aPLP)
{
    Pptco3 Vert;
    Pfloat X, Y, Z;

    const int    nEdges        = 10;
    const int    nVertsPerEdge = 10;
    const Pfloat DeltaPhi      = 5.0 * 3.14 / 180;    // 5 Grad

    if (SimEdgeCnt >= nEdges) return FALSE;

    Pfloat Phi = SimEdgeCnt * DeltaPhi;

    Y = 0;

    for (int I=0; I<nVertsPerEdge; I++) {

        float R;

        switch (SimEdgeCnt) {

            case 0:
            case 1:
            case 4:
                R = 6; break;
            case 2:
            case 3:
                if (I<=(nVertsPerEdge/2)) R=6;
                else if (I<(3*nVertsPerEdge/4)) R+=.5;
                else R-=.5;
                break;
            case 5:
                R = 6.5; break;

            case 6:
            case 7:
                R = 7.0; break;
            default:
                R = 7.0 + (float)(random (21) - 10) / 10.0;        // 6 +- 1.0
        }

        X = R * cos (Phi);
        Y = I;
        Z = R * sin (Phi);

        Vert.point.x = X / 10;
        Vert.point.y = Y / 10;
        Vert.point.z = Z / 10;

        Vert.colour.x = 0.8;
        Vert.colour.y = 0.8;
        Vert.colour.z = 0.8;
        VertexTotal ++;

        aPLP->append (Vert);
    }

    EdgeCnt++;
    SimEdgeCnt++;
    Global.StatWinP->InVerts    = VertexTotal;
    Global.StatWinP->InPls      = EdgeCnt;
    Global.StatWinP->Percentage =  100.0 * EdgeCnt / nEdges;
    return TRUE;
}

BOOL CEFREAD_C::simulatePlane (VERTEXARRAY_C *aPLP)
{
    Pptco3 Vert;

    const int    nEdges        = 10;
    const int    nVertsPerEdge = 10;

    if (SimEdgeCnt >= nEdges) return FALSE;

    Pfloat Y   = 0;
    Pfloat X, Z;

    for (int I=0; I<nVertsPerEdge; I++) {

        X = 0.0;
        Y = I;

        if (SimEdgeCnt%2==0)
            if (SimEdgeCnt%4==0)
                Y*=2;
            else
                Y+=5;

        Z = EdgeCnt;

        Vert.point.x = X / 10;
        Vert.point.y = Y / 10;
        Vert.point.z = Z / 10;

        Vert.colour.x = 0.8;
        Vert.colour.y = 0.8;
        Vert.colour.z = 0.8;

        VertexTotal ++;
        aPLP->append (Vert);
    }

    EdgeCnt++;
    SimEdgeCnt++;
    Global.StatWinP->InVerts    = VertexTotal;
    Global.StatWinP->InPls      = EdgeCnt;
    Global.StatWinP->Percentage =  100.0 * EdgeCnt / nEdges;
    return TRUE;
}

BOOL CEFREAD_C::simulatePlane2 (VERTEXARRAY_C *aPLP)
{
    Pptco3 Vert;

    const int    nEdges        = 20;
    const int    nVertsPerEdge = 20;

    if (SimEdgeCnt >= nEdges) return FALSE;

    Pfloat X, Y, Z;

    for (int I=0; I<nVertsPerEdge; I++) {

        X = 0.5 - (float)random (100) / 100;
        Y = (float)I;
        Z = (float)EdgeCnt;

        Vert.point.x = X / 10;
        Vert.point.y = Y / 10;
        Vert.point.z = Z / 10;

        Vert.colour.x = 0.8;
        Vert.colour.y = 0.8;
        Vert.colour.z = 0.8;

        VertexTotal ++;
        aPLP->append (Vert);
    }

    EdgeCnt++;
    SimEdgeCnt++;
    Global.StatWinP->InVerts    = VertexTotal;
    Global.StatWinP->InPls      = EdgeCnt;
    Global.StatWinP->Percentage =  100.0 * EdgeCnt / nEdges;
    return TRUE;
}

BOOL CEFREAD_C::simulateBox (VERTEXARRAY_C *aPLP)

// Wrfel, oben und unten offen, Kantenlnge 10mm

{
    Pptco3 Vert;
    const int nEdges        = 40;
    const int nVertsPerEdge = 10;

    if (SimEdgeCnt >= nEdges) return FALSE;

    Pfloat X, Y, Z;

    if (SimEdgeCnt>=0.75*nEdges)      { X = -5;                    Z =  5 - SimEdgeCnt % 10; }
    else if (SimEdgeCnt>=0.5*nEdges)  { X =  5 - SimEdgeCnt % 10;  Z =  5; }
    else if (SimEdgeCnt>=0.25*nEdges) { X =  5;                    Z = SimEdgeCnt % 10 - 5; }
    else                              { X = SimEdgeCnt % 10 - 5;   Z = -5; }

    for (int I=0; I<nVertsPerEdge; I++) {

        Y = I;

        Vert.point.x = X / 10;
        Vert.point.y = Y / 10;
        Vert.point.z = Z / 10;

        Vert.colour.x = (X+5.0) / 10.0;
        Vert.colour.y = Y       / nVertsPerEdge;
        Vert.colour.z = (Z+5.0) / 10.0;

        if (X==0) {
            Vert.colour.x = 1; Vert.colour.y = Vert.colour.z = 0;
        }
        if (Y==5.0) {
            Vert.colour.x = Vert.colour.z = 0; Vert.colour.y = 1;
        }
        if (Z==0) {
            Vert.colour.x = Vert.colour.y = 0; Vert.colour.z = 1;
        }

        VertexTotal ++;
        aPLP->append (Vert);
    }

    EdgeCnt++;
    SimEdgeCnt++;
    Global.StatWinP->InVerts    = VertexTotal;
    Global.StatWinP->InPls      = EdgeCnt;
    Global.StatWinP->Percentage =  100.0 * EdgeCnt / nEdges;
    return TRUE;
}

BOOL CEFREAD_C::simulateCone (VERTEXARRAY_C *aPLP)

// Kegelstumpf, R = 1.0 - 6.0
{
    Pptco3 Vert;
    Pfloat X, Y, Z, R;

    const int nEdges        = 20;
    const int nVertsPerEdge = 20;

    if (SimEdgeCnt >= nEdges) return FALSE;

    Pfloat Phi = (Pfloat)SimEdgeCnt * (6.28 / nEdges);

    for (int I=0; I<nVertsPerEdge; I++) {

        R = 1.0 + (Pfloat)I * (5.0/nVertsPerEdge);
        X = R * cos (Phi);
        Y = I;
        Z = R * sin (Phi);

        Vert.point.x = X / 10;
        Vert.point.y = Y / 10;
        Vert.point.z = Z / 10;

        Vert.colour.x = Phi / 6.28;
        Vert.colour.y = (float)I / nVertsPerEdge;
        Vert.colour.z = 1.0;
        if ( (I%5==0) || (SimEdgeCnt%5==0)) {
            Vert.colour.x = Vert.colour.z = 1.0; Vert.colour.y = 0.0;
        }

        VertexTotal ++;
        aPLP->append (Vert);
    }

    EdgeCnt++;
    SimEdgeCnt++;
    Global.StatWinP->InVerts    = VertexTotal;
    Global.StatWinP->InPls      = EdgeCnt;
    Global.StatWinP->Percentage =  100.0 * EdgeCnt / nEdges;
    return TRUE;
}

BOOL CEFREAD_C::simulateSphere (VERTEXARRAY_C *aPLP)
{
    Pptco3 Vert;
    Pfloat X, Y, Z;

    const int    nEdges        = 20;
    const int    nVertsPerEdge = 20;
    const Pfloat Radius        = 10.0;

    if (SimEdgeCnt >= nEdges) return FALSE;

    Pfloat Phi = (Pfloat)SimEdgeCnt * (6.28 / nEdges);

    for (int I=-(nVertsPerEdge/2)+1; I<(nVertsPerEdge/2); I++) {

        Y = I + (nVertsPerEdge/2);
        Pfloat DY = sqrt (Radius*Radius - I*I);
        X = DY * cos (Phi);
        Z = DY * sin (Phi);

        Vert.point.x = X / 10;
        Vert.point.y = Y / 10;
        Vert.point.z = Z / 10;

        Vert.colour.x = Phi / 6.28;
        Vert.colour.y = (float)Y / nVertsPerEdge;
        Vert.colour.z = Phi / 6.28;
        if ( (I%5==0) || (SimEdgeCnt%5==0)) {
            Vert.colour.x = Vert.colour.z = 1.0; Vert.colour.y = 0.0;
        }

        VertexTotal ++;
        aPLP->append (Vert);
    }

    EdgeCnt++;
    SimEdgeCnt++;
    Global.StatWinP->InVerts    = VertexTotal;
    Global.StatWinP->InPls      = EdgeCnt;
    Global.StatWinP->Percentage =  100.0 * EdgeCnt / nEdges;
    return TRUE;
}


/*------------------------------------------------------------------------+
|                                                                         |
| public Methoden                                                         |
|                                                                         |
+------------------------------------------------------------------------*/

CEFREAD_C::CEFREAD_C ()        // Konstruktor
{
    init();
}

CEFREAD_C::~CEFREAD_C()            // Destruktor
{
    close();
}

void CEFREAD_C::init()
{
    CefIn          = NULL;
    DebugOut       = NULL;
    Status         = CEF_OK;
    DebugIt        = FALSE;
    EdgeCnt        = 0;
    ByteCnt        = 0L;
    VertexTotal    = 0L;
    HeaderRead     = FALSE;
    CamTranOk      = FALSE;
    FileSize       = 0L;

    SimEdgeCnt     = 0;

    for (int I=0; I<MAX_CAMERAS; I++)
        makeIdentity44 (CamTran [I]);
    makeIdentity44 (CurTran);
}

CEF_STATUS_E CEFREAD_C::open (char *anFSpec, BOOL aDebugIt)
{
    if (Global.OptionsP->SimulationMode != SIM_NONE) {
        strcpy (Global.StatWinP->InName,
                SimulationModeStr [Global.OptionsP->SimulationMode]);

        return CEF_OK;
    }

    if ( isOpen() ) return CEF_OK;

    DebugIt = aDebugIt;

    // Dateinamen ins Statusfenster bringen:

    char Name[15], Ext[5];

    fnsplit (anFSpec, NULL, NULL, Name, Ext);
    strcpy (Global.StatWinP->InName, strcat (Name,Ext));
    Global.StatWinP->updateDlg();

    // Datei ffnen:

    CefIn = fopen (anFSpec, "rb");               // read only, binary mode

    if (DebugIt) {
        DebugOut = fopen ("\\CefRead.DBG", "wt");  // write only, create/overwrite, text mode
        if ( ! DebugOut ) DebugIt = FALSE;
    }

    if (! CefIn) return CEF_NOTFOUND;

    if (readHeader()) HeaderRead = TRUE;

    FileSize = fileSize (CefIn);

    return Status;
}

void CEFREAD_C::close()
{
    if ( ! isOpen() ) return;

    debugMsg ("\nclose(): File pointer at: %ld\n", ftell (CefIn));
    if (DebugOut) fclose (DebugOut);
    DebugOut   = NULL;
    DebugIt    = FALSE;

    fclose (CefIn);
    CefIn      = NULL;

    ByteCnt    = 0;
    Status     = CEF_OK;
    HeaderRead = FALSE;
}

static char DummyStr [200];

BOOL CEFREAD_C::readNextPL (VERTEXARRAY_C *aPlP)
{
    WORD      Key;
    DWORD     Paragraphs;
    DWORD     LastByteCnt;
    DWORD     ActualParagraphs;

    // Sollen die Eingabedaten simuliert werden ?

    if (Global.OptionsP->SimulationMode != SIM_NONE) {

        Status = CEF_OK;
        BOOL CC;

        Global.StatWinP->status ("Simulating input: %s",
                                  SimulationModeStr [Global.OptionsP->SimulationMode]);

        switch (Global.OptionsP->SimulationMode) {

            case SIM_PLANE:    CC = simulatePlane    (aPlP); break;
            case SIM_PLANE2:   CC = simulatePlane2   (aPlP); break;
            case SIM_SPHERE:   CC = simulateSphere   (aPlP); break;
            case SIM_CONE:     CC = simulateCone     (aPlP); break;
            case SIM_CYLINDER: CC = simulateCylinder (aPlP); break;
            case SIM_BOX:      CC = simulateBox      (aPlP); break;
        default:
            Global.warning ("readNextPl: illegal simulation mode (%i)",
                            Global.OptionsP->SimulationMode);
        }

        return CC;
    }

    // Normalerweise:

    if ( ! isOpen() ) return FALSE;

    Global.StatWinP->status ("reading CEF ...");

    BOOL      Done       = FALSE;            // Polyline gelesen ?
    BOOL      EdgeTranOk = FALSE;            // Edge Transformation gelesem ?

    do {
        if (!nextKey (Key))          return FALSE;
        if (!readDWord (Paragraphs)) return FALSE;
        LastByteCnt = ByteCnt;

        switch ( (CEF_KEY_E)Key ) {
            case CEF_HEADER_KEY:
                debugMsg ("\nWARNING: Multiple header declaration:\n");
                readHeader();
                break;
            case CEF_F_COMM_KEY:
                readFileComment (DummyStr, sizeof (DummyStr));
                break;
            case CEF_E_COMM_KEY:
                readEdgeComment (DummyStr, sizeof (DummyStr));
                break;
            case CEF_TABLECMD_KEY:
                BYTE B;
                if (Paragraphs<1) break;
                for (long I=0L; I<(Paragraphs*16-6); I++) if (!readByte (B)) break;
                break;
            case CEF_E_TRANSF_KEY:
                if (EdgeTranOk) debugMsg ("\nWARNING: Multiple transformations for this edge (concatenating)\n");
                readEdgeTransformation();
                EdgeTranOk = TRUE;
                break;
            case CEF_E_DATA_KEY:
                if (!EdgeTranOk) debugMsg ("\nWARNING: No transformation for this edge (using last one)\n");
                readPlData (aPlP);
                Done = TRUE;
                break;
            case CEF_CAMTRAN_KEY:
                if (CamTranOk) debugMsg ("\nWARNING: Overriding old camera transformation\n");
                readCameraTransformation();
                break;
            case CEF_EOF_KEY:
                close();
                return FALSE;
                //break;
        default:
            debugMsg ("\nUnknown Key (%4X) at file position %lu\n", Key, ftell (CefIn));
            Status = CEF_CORRUPTED;
        }

        ActualParagraphs = ((ByteCnt-(LastByteCnt-6))+ALIGNMENT-1) / ALIGNMENT;

        if ( ActualParagraphs != Paragraphs )
            debugMsg ("\nWARNING: BlockSize (%lu) does not match declaration (%lu)\n",
                       ActualParagraphs, Paragraphs);
    } while ( !Done && (Status==CEF_OK) );

    Global.StatWinP->status ("");
    if (FileSize>0)
        Global.StatWinP->Percentage =  100.0 * ByteCnt / FileSize;

    if (Status!=CEF_OK)
        Global.warning ("Error reading CEF (%i) at file position %lu\n", Status, ftell (CefIn));

    return Done;
}

