/**************************************************************************
*                                                                         *
* Name   : cFileBox.cpp                ( Headerdatei:                 )   *
*                                                                         *
* Klasse(n) definiert : FILEBOX_C                                         *
* Objekt(e) deklariert: -                                                 *
* Abgeleitet von      : -                                                 *
*                                                                         *
* Aufgabe:                                                                *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include "cTypes.hpp"

#include <dir.h>
#include <string.h>

#include "cGlobal.hpp"
#include "cFileBox.hpp"

extern "C" {
    #include "\sdk\include\commdlg.h"
}

// --- lokale Konstanten:----------------------------------------------

// --- lokale Typen: --------------------------------------------------

// --- lokale Variablen: ----------------------------------------------

/*------------------------------------------------------------------------+
|                                                                         |
| lokale Funktionen                                                       |
|                                                                         |
+------------------------------------------------------------------------*/

/*------------------------------------------------------------------------+
|                                                                         |
| Message handler                                                         |
|                                                                         |
+------------------------------------------------------------------------*/

/*------------------------------------------------------------------------+
|                                                                         |
| Methoden der Klasse FILEBOX_C                                           |
|                                                                         |
+------------------------------------------------------------------------*/

FILEBOX_C::FILEBOX_C()             // Konstruktor
{
    Mode = FBX_OPENDLG;

    strcpy (Drive,  "");
    strcpy (Dir,    "");
    strcpy (Name,   "");
    strcpy (Ext,    "");

    strcpy (DefExt, "");

    Flags = OFN_HIDEREADONLY;
}

FILEBOX_C::~FILEBOX_C()            // Destruktor
{
}

void FILEBOX_C::FSpecToFName (char *lpszFSpec, char *lpszFName)
{
    char Name [10];
    char Ext  [5];

    fnsplit (lpszFSpec, NULL, NULL, Name, Ext);
    strcpy (lpszFName, Name);
    strcat (lpszFName, Ext);
}

BOOL FILEBOX_C::fileExist (char *lpszFSpec)
{
    OFSTRUCT FileInfoRec;

    return ( OpenFile (lpszFSpec, &FileInfoRec, OF_EXIST) != -1 );
}

BOOL FILEBOX_C::fileDlg (HWND hCaller, char *lpszHeader,
                          char FSpec[], int FSpecSize,
                          char *szFilter[], DWORD &CurrentFilter,
                          char *lpszDefExt)
{
    OPENFILENAME of;

//  char szDirName[256];                /* directory name array  */

    // Get the system directory name and store in szDirName

//  GetSystemDirectory ((LPSTR)szDirName, 255);

    // Initialize the OPENFILENAME members

    of.lStructSize              = sizeof (OPENFILENAME);
    of.hwndOwner                = hCaller;                 //  -> parent window
    of.lpstrFilter              = szFilter[0];             //
    of.lpstrCustomFilter        = (LPSTR) NULL;            // <-> History list buffer
    of.nMaxCustFilter           = 0L;                      //  -> History list buffer size
    of.nFilterIndex             = CurrentFilter;           // <-> currently selected filter
    of.lpstrFile                = FSpec;                   // <-> drive + path + name + ext
    of.nMaxFile                 = FSpecSize;               //
    of.lpstrFileTitle           = NULL;                    // <-  name + ext
    of.nMaxFileTitle            = 0;                       //  ->
    of.lpstrInitialDir          = NULL; //szDirName;       //  ->
    of.lpstrTitle               = lpszHeader;              //  -> window header
    of.Flags                    = Flags;                   // <-> Dlg Style (may set OFN_EXTENSIONDIFFERENT)
    of.nFileOffset              = 0;                       //  -> Offs of file name in lpstrFile
    of.nFileExtension           = 0;                       //  -> Offs of file ext  in lpstrFile
    of.lpstrDefExt              = (LPSTR) lpszDefExt;      //  -> keine default extension
    of.lCustData                = (DWORD) 0L;              //  ->
    of.lpfnHook                 = NULL;                    //
    of.lpTemplateName           = (LPSTR) NULL;            //


    BOOL Status = (Mode==FBX_OPENDLG) ? GetOpenFileName (&of) : GetSaveFileName (&of);


    CurrentFilter = of.nFilterIndex;

    if (Status)
        fnsplit (FSpec, Drive, Dir, Name, Ext);
    else {
        DWORD CC = CommDlgExtendedError();
        if ( CC != 0 )
            Global.warning ("FileBox: CommDlgExtendedError: %li", CC);
    }

    return Status;
}

BOOL FILEBOX_C::openFileDlg (HWND hCaller, char *lpszHeader,
                             char FSpec[], int FSpecSize,
                             char *szFilter[], DWORD &CurrentFilter,
                             char *lpszDefExt)
{
    Mode  = FBX_OPENDLG;
    Flags = OFN_HIDEREADONLY | OFN_PATHMUSTEXIST;

    return fileDlg (hCaller, lpszHeader, FSpec, FSpecSize, szFilter,
                    CurrentFilter, lpszDefExt);
}

BOOL FILEBOX_C::openExistingFileDlg (HWND hCaller, char *lpszHeader,
                                     char FSpec[], int FSpecSize,
                                     char *szFilter[], DWORD &CurrentFilter,
                                     char *lpszDefExt)
{
    BOOL     Done         = FALSE;
    char     TempStr[256];

    Mode  = FBX_OPENDLG;
    Flags = OFN_HIDEREADONLY | OFN_FILEMUSTEXIST;

    strncpy (TempStr, FSpec, sizeof (TempStr));

    do {
        BOOL Status = fileDlg (hCaller, lpszHeader, FSpec, FSpecSize,
                               szFilter, CurrentFilter, lpszDefExt);

        if (!Status) {                 // "Cancel" oder FSpec-Buffer zu klein ?
            strncpy (FSpec, TempStr, FSpecSize);
            return Status;
        }

        if ( ! fileExist (FSpec) )
            Global.warning ("File does not exist: '%s'", FSpec);
        else
            Done =TRUE;

    } while (!Done);

    return (DWORD)0;
}

BOOL FILEBOX_C::saveFileDlg (HWND hCaller, char *lpszHeader,
                              char FSpec[], int FSpecSize,
                              char *szFilter[], DWORD &CurrentFilter,
                              char *lpszDefExt)
{
    Mode  = FBX_SAVEDLG;
    Flags = OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT; // | OFN_NOREADONLYRETURN;

    return fileDlg (hCaller, lpszHeader, FSpec, FSpecSize, szFilter,
                    CurrentFilter, lpszDefExt);
}
