/**************************************************************************
*                                                                         *
* Name   : cGeomOut.cpp                ( Headerdatei:                 )   *
*                                                                         *
* Klasse(n) definiert : GEOMETRYOUT_C                                     *
* Objekt(e) deklariert: -                                                 *
* Abgeleitet von      : -                                                 *
*                                                                         *
* Aufgabe:                                                                *
*                                                                         *
*          Da Pfloat entweder float ODER double sein kann, wird intern    *
*          mit double gerechnet.                                          *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include "cTypes.hpp"

#include <dir.h>
#include <string.h>

#include "cGlobal.hpp"
#include "cGeomOut.hpp"
#include "cMath.hpp"
#include "cOptions.hpp"
#include "cStatWin.hpp"
#include "cTexture.hpp"
#include "cTools.hpp"
#include "cTriset.hpp"

// --- lokale Konstanten:----------------------------------------------

// benutzt Arena double floats ?

#undef ARENAUSESDOUBLE


// --- lokale Typen: --------------------------------------------------

// --- lokale Variablen: ----------------------------------------------


/*------------------------------------------------------------------------+
|                                                                         |
| lokale Funktionen                                                       |
|                                                                         |
+------------------------------------------------------------------------*/


/*------------------------------------------------------------------------+
|                                                                         |
| Methoden der Klasse GEOMETRYOUT_C                                       |
|                                                                         |
+------------------------------------------------------------------------*/

GEOMETRYOUT_C::GEOMETRYOUT_C()             // Konstruktor
{
    Zof = NULL;
    strcpy (LayerName, "0");
    strcpy (NameBody, "DUMMY");
}

GEOMETRYOUT_C::~GEOMETRYOUT_C()            // Destruktor
{
}

BOOL GEOMETRYOUT_C::open (char* aFName, BOOL aSmall)
{
    Zof = fopen (aFName, "wt");        // open text file for writing
    Small = aSmall;

    fnsplit (aFName, NULL, NULL, NameBody, NULL);
    strupr (NameBody);

    return (Zof!=NULL);
}

void GEOMETRYOUT_C::close()
{
    if (Zof) fclose (Zof);
    Zof=NULL;
}

/*------------------------------------------------------------------------+
|                                                                         |
| ZOF Format                                                              |
|                                                                         |
+------------------------------------------------------------------------*/

BOOL GEOMETRYOUT_C::openPattern (char* aPatternName)
{
    if (!Zof) return FALSE;
    int Stat = fprintf (Zof, "PATTERN \"%s\"\n", aPatternName);
    return Stat!=EOF;
}

BOOL GEOMETRYOUT_C::closePattern ()
{
    if (!Zof) return FALSE;
    int Stat = fprintf (Zof, "END\n");
    return Stat!=EOF;
}

BOOL GEOMETRYOUT_C::openGroup (char* aName)
{
    if (!Zof) return FALSE;
    int Stat = fprintf (Zof, "GROUP \"%s\"\n", aName);
    return Stat!=EOF;
}

BOOL GEOMETRYOUT_C::closeGroup()
{
    if (!Zof) return FALSE;
    int Stat = fprintf (Zof, "END\n");
    return Stat!=EOF;
}

BOOL GEOMETRYOUT_C::writeObject (char* anObjectName, char* aPatternName)
{
    if (!Zof) return FALSE;
    int Stat = fprintf (Zof, "OBJECT \"%s\" ISA \"%s\"\n", anObjectName, aPatternName);
    return Stat!=EOF;
}

BOOL GEOMETRYOUT_C::writePoint3 (Ppoint3 &P)
{
    int Stat;

    if (!Zof) return FALSE;

    if (Small) {
        fprintf (Zof, "%s\t", stripZeros((double)P.x));
        fprintf (Zof, "%s\t", stripZeros((double)P.y));
        Stat = fprintf (Zof, "%s\n", stripZeros((double)P.z));
        }
    else
        Stat = fprintf (Zof, "%10.5lf\t%10.5lf\t%10.5lf\n",
                             (double)P.x, (double)P.y, (double)P.z);

    return Stat!=EOF;
}

BOOL GEOMETRYOUT_C::writePt3UV (Ppoint3 &P, Pfloat U, Pfloat V)
{
    int Stat;

    if (!Zof) return FALSE;

    if (Small) {
        fprintf (Zof, "%s\t", stripZeros((double)P.x));
        fprintf (Zof, "%s\t", stripZeros((double)P.y));
        fprintf (Zof, "%s\t", stripZeros((double)P.z));
        fprintf (Zof, "%s\t", stripZeros((double)U));
        Stat = fprintf (Zof, "%s\n", stripZeros((double)V));
        }
    else
        Stat = fprintf (Zof, "%10.5lf\t%10.5lf\t%10.5lf\t%10.5lf\t%10.5lf\n",
                             (double)P.x, (double)P.y, (double)P.z,
                             (double)U, (double)V);

    return Stat!=EOF;
}

BOOL GEOMETRYOUT_C::writePtCo3 (Pptco3 &V)
{
    int Stat;

    if (!Zof) return FALSE;

    if (Small) {
        fprintf (Zof, "%s\t", stripZeros((double)V.point.x));
        fprintf (Zof, "%s\t", stripZeros((double)V.point.y));
        fprintf (Zof, "%s\t", stripZeros((double)V.point.z));
        fprintf (Zof, "%s\t", stripZeros((double)V.colour.x));
        fprintf (Zof, "%s\t", stripZeros((double)V.colour.y));
        Stat = fprintf (Zof, "%s\n", stripZeros((double)V.colour.z));
        }
    else
        Stat = fprintf (Zof, "%10.5lf\t%10.5lf\t%10.5lf\t%10.5lf\t%10.5lf\t%10.5lf\n",
                              (double)V.point.x, (double)V.point.y, (double)V.point.z,
                              (double)V.colour.x, (double)V.colour.y, (double)V.colour.z);
    return Stat!=EOF;

}

#pragma argsused
BOOL GEOMETRYOUT_C::writeHeaderZof (TRISET_C *TP)
{
    int  Stat;

    if (!Zof) return FALSE;

    // Material definieren:

    Stat = fprintf (Zof, "MATERIAL \"%s%s\"\n", NameBody, "_MAT");
    if (Stat==EOF) return FALSE;

    Stat = fprintf (Zof, "0.7 0.7 0.7  ! Interior color\n");
    Stat = fprintf (Zof, "0.0 0.0 0.0  ! Line color\n");
    Stat = fprintf (Zof, "0.3 0.3 0.4  ! Material reflection coefficients: ambient, diffuse, specular\n");
    Stat = fprintf (Zof, "1.0 1.0 1.0  ! reflection highlight colour\n");
    Stat = fprintf (Zof, "30           ! exponent\n");
    Stat = fprintf (Zof, "0            ! transparency\n");

    switch (Global.OptionsP->GeomFileType) {

        case GEOM_DATAUV:
            Stat = fprintf (Zof, "1            ! Texture flag\n");
            Stat = fprintf (Zof, "\"%s\"\n", Global.OptionsP->TexFile);
            Stat = fprintf (Zof, "SOLID\nNOEDGES\nFLAT\n\n");
            break;
        case GEOM_DATANONE:
        case GEOM_DATARGB:
        default:
            Stat = fprintf (Zof, "0            ! Texture flag\n");
            Stat = fprintf (Zof, "SOLID\nEDGES\nGOURAUD\n\n");
            break;
    }

    return (Stat!=EOF);
}

BOOL GEOMETRYOUT_C::flushZof (TRISET_C *TP)
{
    int  Stat;
    long I;
    int  FaceFlag;
    int  VertexFlag;
    char Str [50];

    if (!Zof) return FALSE;

    sprintf (Str, "%s%s %i", NameBody, "_PAT", FlushCnt+1);
    openPattern (Str);

    FaceFlag = PFA_NONE;

    switch (Global.OptionsP->GeomFileType) {

        case GEOM_DATANONE:
            Global.StatWinP->status ("writing ZOF (XYZ) ...");
            VertexFlag     = PVERT_NONE;
            break;
        case GEOM_DATARGB:
            Global.StatWinP->status ("writing ZOF (XYZ, RGB) ...");
            VertexFlag = PVERT_COLOUR;
            break;
        case GEOM_DATAUV:
            Global.StatWinP->status ("writing ZOF (XYZ, UV) ...");
            VertexFlag = PVERT_DATA;
            break;
        default:
            Global.warning ("Unsuported geometry file format (%i)",
                            Global.OptionsP->GeomFileType);
            break;
    }

    long FaceCnt = TP->FaceCnt;
    long VertCnt = TP->VertexCnt;

    Stat = fprintf (Zof, "TRISET3 %i %i \n %li %li \n",
                         FaceFlag, VertexFlag, FaceCnt, VertCnt);
    if (Stat==EOF) return FALSE;

    if (TP->echo())
        Global.StatWinP->OutSets++;

    Global.StatWinP->updateDlg();

    int Umax = Global.OptionsP->U;
    int Vmax = Global.OptionsP->V;

    for (I=0; I<VertCnt; I++) {

        Pptco3 *V1 = TP->getVertPtCoP (I);

        switch (Global.OptionsP->GeomFileType) {

            case GEOM_DATANONE:
                if ( ! writePoint3 (V1->point) ) return FALSE;
                break;

            case GEOM_DATARGB:
                if ( ! writePtCo3 (*V1) ) return FALSE;
                break;

            case GEOM_DATAUV:

                int U, V;

                // Texturkoordinaten berechnen:

                Global.TextureP->xyz2uv (V1->point, U, V);

                // fr ARENA: Nullpunkt nach oben links legen, auf 1.0 normalisieren:

                if ( ! writePt3UV (V1->point, (Pfloat)U/Umax, 1.0-(Pfloat)V/Vmax) )
                    return FALSE;
                break;
        }
    }

    for (I=0; I<FaceCnt; I++) {

        long *F  = TP->getFaceIdxP (I);

        if (Small)
            Stat = fprintf (Zof, "%li %li %li\n", F[0]+1, F[1]+1, F[2]+1);
        else
            Stat = fprintf (Zof, "%5li\t%5li\t%5li\n", F[0]+1, F[1]+1, F[2]+1);
        if (Stat==EOF) return FALSE;
    }

    Global.StatWinP->status ("");

    closePattern();

    return Stat!=EOF;
}

#pragma argsused
BOOL GEOMETRYOUT_C::writeFooterZof (TRISET_C *TP)
{
    int  Stat;
    char Str[100];

    sprintf (Str, "%s%s", NameBody, "_GRP");
    openGroup (Str);

    for (int J=1; J<=FlushCnt; J++) {


        Stat = fprintf (Zof, "PART \"%s%s %i\" ISA \"%s%s %i\"\n",
                             NameBody, "_PART", J,
                             NameBody, "_PAT", J
                       );

        if (Stat==EOF) return FALSE;
    }
    closeGroup();

    sprintf (Str, "%s%s", NameBody, "_GRP");
    if ( ! writeObject (NameBody, Str) ) return FALSE;

    fprintf (Zof, "BEGIN\n");
    fprintf (Zof, "\tMATERIAL \"%s%s\"\n", NameBody, "_MAT");
    fprintf (Zof, "\tMOVE  2  1  0\n");
    fprintf (Zof, "\tSCALE 1.2 1.2 1.2\n");
    Stat = fprintf (Zof, "END\n");

    if (Stat==EOF) return FALSE;

    return TRUE;
}

/*------------------------------------------------------------------------+
|                                                                         |
| DXF Format                                                              |
|                                                                         |
+------------------------------------------------------------------------*/

BOOL GEOMETRYOUT_C::writePoint3Dxf (Ppoint3 &P)
{
    int Stat;

    if (!Zof) return FALSE;

    Stat = fprintf (Zof, "0\nVERTEX\n8\n%s\n", LayerName);
    if (Stat==EOF) return FALSE;

    if (Small) {
        fprintf (Zof, "10\n%s\n", stripZeros((double)P.x));
        fprintf (Zof, "20\n%s\n", stripZeros((double)P.y));
        fprintf (Zof, "30\n%s\n", stripZeros((double)P.z));
        Stat = fprintf (Zof, "70\n192\n");
        }
    else {
        fprintf (Zof, "10\n%10.5lf\n20\n%10.5lf\n30\n%10.5lf\n",
                      (double)P.x, (double)P.y, (double)P.z);
        Stat = fprintf (Zof, "70\n192\n\n");
    }

    return Stat!=EOF;
}

BOOL GEOMETRYOUT_C::writeFaceDxf (long I1, long I2, long I3)
{
    int Stat;

    if (!Zof) return FALSE;

    Stat = fprintf (Zof, "0\nVERTEX\n8\n%s\n", LayerName);
    if (Stat==EOF) return FALSE;
    Stat = fprintf (Zof, "10\n0\n20\n0\n30\n0\n70\n128\n");
    if (Stat==EOF) return FALSE;
    Stat = fprintf (Zof, "71\n%li\n72\n%li\n73\n%li\n",
                         I1 + 1, I2 + 1, I3 + 1);
    if (Stat==EOF) return FALSE;

    return Stat!=EOF;
}

#pragma argsused
BOOL GEOMETRYOUT_C::writeHeaderDxf (TRISET_C *TP)
{
    int Stat;

    if (!Zof) return FALSE;

    Stat = fprintf (Zof, "0\nSECTION\n2\nENTITIES\n");

    return (Stat!=EOF);
}

BOOL GEOMETRYOUT_C::flushDxf (TRISET_C *TP)
{
    int  Stat;
    long I;

    // Triset header:

    Stat = fprintf (Zof, "0\nPOLYLINE\n8\n%s\n66\n1\n70\n64\n", LayerName);
    if (Stat==EOF) return FALSE;
    Stat = fprintf (Zof, "71\n%li\n72\n%li\n", TP->VertexCnt, TP->FaceCnt);
    if (Stat==EOF) return FALSE;

    // Vertexliste:

    for (I=0; I<TP->VertexCnt; I++) {

        writePoint3Dxf (*TP->getVertPtP (I));
    }

    // Faceliste

    for (I=0; I<TP->FaceCnt; I++) {

        long *FIP = TP->getFaceIdxP (I);

        writeFaceDxf (FIP[0], FIP[1], FIP[2]);
    }

    // Triset footer:

    Stat = fprintf (Zof, "0\nSEQEND\n8\n%s\n", LayerName);
    if (Stat==EOF) return FALSE;

    return TRUE;
}

#pragma argsused
BOOL GEOMETRYOUT_C::writeFooterDxf (TRISET_C *TP)
{
    int Stat;

    if (!Zof) return FALSE;

    Stat = fprintf (Zof, "0\nENDSEC\n0\nEOF\n");

    return (Stat!=EOF);
}

/*------------------------------------------------------------------------+
|                                                                         |
| Alle Formate (Dispatcher)                                               |
|                                                                         |
+------------------------------------------------------------------------*/

BOOL GEOMETRYOUT_C::writeHeader(TRISET_C *TP)
{
    BOOL Stat = FALSE;

    switch (Global.OptionsP->GeomFileFormat) {

        case GEOM_ZOF:
            Stat = writeHeaderZof (TP);
            break;
        case GEOM_DXF:
            Stat = writeHeaderDxf (TP);
            break;
    }
    return Stat;
}

BOOL GEOMETRYOUT_C::flush (TRISET_C *TP)
{
    BOOL Stat = FALSE;

    switch (Global.OptionsP->GeomFileFormat) {

        case GEOM_ZOF:
            Stat = flushZof (TP);
            break;
        case GEOM_DXF:
            Stat = flushDxf (TP);
            break;
    }
    FlushCnt++;
    return Stat;
}

BOOL GEOMETRYOUT_C::writeFooter (TRISET_C *TP)
{
    BOOL Stat = FALSE;

    switch (Global.OptionsP->GeomFileFormat) {

        case GEOM_ZOF:
            Stat = writeFooterZof (TP);
            break;
        case GEOM_DXF:
            Stat = writeFooterDxf (TP);
            break;
    }
    return Stat;
}

BOOL GEOMETRYOUT_C::writeTriset (TRISET_C *TP)
{
    float VertexWeight, FaceWeight, MaxWeight;

    FlushCnt = 0;

    writeHeader (TP);

    TRISET_C *NewTP = new TRISET_C;

    if (!NewTP) return FALSE;

    // Hashing einschalten
    //(Abschtzung der Gre: (2*sichtbare Vertizes) ~ FaceCnt):

    NewTP->createVertHashTab (TP->FaceCnt);

    Global.StatWinP->status ("creating compacted triset ...");


    if ( Global.OptionsP->SplitTriset ) {

        VertexWeight = Global.OptionsP->VertexWeight;
        FaceWeight   = Global.OptionsP->FaceWeight;
        MaxWeight    = Global.OptionsP->MaxWeight;
        }
    else {

        VertexWeight = FaceWeight = 0.0;
        MaxWeight    = 1.0;
    }

    // Triset kompaktieren:

    for (long I=0; I<TP->FaceCnt; I++) {

        long    *FIP = TP->getFaceIdxP (I);

        if ( !_isFaceDeleted (FIP) ) {

            NewTP->addTriangle (*TP->getVertPtCoP (FIP[0]),
                                *TP->getVertPtCoP (FIP[1]),
                                *TP->getVertPtCoP (FIP[2]) );
        }

        // Triset schreiben, wenn maximale Gre erreicht:

        if ( ((NewTP->FaceCnt + 1) * FaceWeight) +
             ((NewTP->VertexCnt + 3) * VertexWeight) > MaxWeight) {

            if ( ! flush (NewTP) ) {
                Global.warning ("Error writing triset");
                return FALSE;
            }

            NewTP->clear();
            Global.StatWinP->status ("creating compacted triset ...");
        }
    }

    if ( (NewTP->FaceCnt>0) ) {

        if ( ! flush (NewTP) ) {
            Global.warning ("Error writing triset");
            return FALSE;
        }
    }

    delete NewTP;

    writeFooter (TP);

    if (Global.OptionsP->LogSettings)
        Global.logMsg ("Sets written\t: %i\n", FlushCnt);

    Global.StatWinP->status ("");
    if (TP->echo())
        Global.StatWinP->Percentage = 100.0;
    Global.StatWinP->updateDlg();

    return TRUE;
}
