/**************************************************************************
*                                                                         *
* Name   : cGlobal.cpp                ( Headerdatei:                 )    *
*                                                                         *
* Klasse(n) definiert : GLOBAL_C                                          *
* Objekt(e) deklariert: Global                                            *
*                                                                         *
* Abgeleitet von      : -                                                 *
*                                                                         *
* Aufgabe:                                                                *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include "cTypes.hpp"

#include <dir.h>
#include <math.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <time.h>

#include "cGlobal.hpp"

#include "\src\lib\crtio.h"

#include "cCefRead.hpp"
#include "cGlue.hpp"
#include "cOptions.hpp"
#include "cMainWin.hpp"
#include "cReduce.hpp"
#include "cStatWin.hpp"
#include "cTexture.hpp"
#include "cTriset.hpp"
#include "cGeomOut.hpp"


// --- lokale Konstanten:----------------------------------------------

// --- lokale Typen: --------------------------------------------------

// --- lokale Variablen: ----------------------------------------------


/*------------------------------------------------------------------------+
|                                                                         |
| lokale Funktionen                                                       |
|                                                                         |
+------------------------------------------------------------------------*/

/*------------------------------------------------------------------------+
|                                                                         |
| Methoden der Klasse GLOBAL_C                                            |
|                                                                         |
+------------------------------------------------------------------------*/

GLOBAL_C::GLOBAL_C ()
{
    Instance      = 0;
    CancelRequest = FALSE;
    BusyCnt       = 0;
    WaitCnt       = 0;

    GlueP         = new GLUE_C();

    CefReadP      = NULL;
    MainWndP      = NULL;
    OptionsP      = NULL;
    ReduceP       = NULL;
    StatWinP      = NULL;
    TextureP      = NULL;
    TrisetP       = NULL;
    GeomOutP     = NULL;

    LogFile       = NULL;
    strcpy (LogFileName, "");
    StatsDataFile = NULL;
    strcpy (StatsFileName, "");

    HWND       hDtWnd = GetDesktopWindow();
    HDC        hDtDC  = GetDC (hDtWnd);
    TEXTMETRIC TM;
    SelectObject (hDtDC, GetStockObject (OEM_FIXED_FONT) );
    GetTextMetrics (hDtDC, &TM);
    ScreenRes.x = GetDeviceCaps (hDtDC, HORZRES);
    ScreenRes.y = GetDeviceCaps (hDtDC, VERTRES);
    ReleaseDC      (hDtWnd, hDtDC);
    CharSize.x = TM.tmMaxCharWidth;
    CharSize.y = TM.tmHeight+TM.tmExternalLeading+TM.tmInternalLeading;
}

GLOBAL_C::~GLOBAL_C()
{
    if (CefReadP)    delete CefReadP;
    if (ReduceP)     delete ReduceP;
    if (StatWinP)    delete StatWinP;
    if (TextureP)    delete TextureP;
    if (TrisetP)     delete TrisetP;

    if (GeomOutP) {
        if (GeomOutP->isOpen()) GeomOutP->close();
        delete GeomOutP;
    }
    if (MainWndP)    delete MainWndP;

    if (OptionsP)    delete OptionsP;  // andere Destruktoren ndern  Parameter
    if (GlueP)       delete GlueP;     // zum Schlu (wird in den anderen Destruktoren benutzt!)

    if (LogFile)       fclose (LogFile);
    if (StatsDataFile) fclose (StatsDataFile);
}

HWND GLOBAL_C::getMainWnd()
{
    return (MainWndP!=NULL) ? MainWndP->hWnd : HWND (0);
}

BOOL GLOBAL_C::init  (HANDLE aInstance, HANDLE aPrevInstance,
                      LPSTR aCmdLine, int aCmdShow)
{
    Instance      = aInstance;
    PrevInstance  = aPrevInstance;
    CmdLine       = aCmdLine;
    CmdShow       = aCmdShow;

    MainWndP      = new MAINWINDOW_C;

    // wenn das Fenster kein Icon sein soll, dann ganz gro:

    if ( (CmdShow==SW_SHOW) || (CmdShow==SW_SHOWNORMAL) ||
         (CmdShow==SW_RESTORE) || (CmdShow==SW_SHOWNA) ||
          (CmdShow==SW_SHOWNOACTIVATE) )
        aCmdShow = SW_SHOWMAXIMIZED;

    MainWndP->openWnd (aCmdShow);      // folgende Konstruktoren brauchen evtl.
                                       // das Window Handle

    CefReadP      = new CEFREAD_C;
    OptionsP      = new OPTIONS_C;
    ReduceP       = new REDUCE_C;
    StatWinP      = new STATUSWINDOW_C;
    TextureP      = new TEXTURE_C;
    TrisetP       = new TRISET_C;
    GeomOutP      = new GEOMETRYOUT_C;

    TrisetP->echo (TRUE);
    MainWndP->updateMenu();            // Settings aus *OptionsP bernehmen

    return TRUE;
}

void GLOBAL_C::reInit()

// Globale Objekte auf neue Konvertierung vorbereiten:

{
    StatWinP->init();
    CefReadP->init();
    ReduceP->init();
    TextureP->init();
    TrisetP->init();

    char Name[15], Ext[5];

    fnsplit (OptionsP->GeomFile, NULL, NULL, Name, Ext);
    strcpy (StatWinP->OutName, strcat (Name,Ext));
    fnsplit (OptionsP->InFile, NULL, NULL, Name, Ext);
    strcpy (StatWinP->InName, strcat (Name,Ext));

    openLog();
    openStatsFile();
}

void GLOBAL_C::yield()
{
    kbhit();
}

/*------------------------------------------------------------------------+
|                                                                         |
| Ausgabe von Strings (Infos, Warnings, Debugging, ...)                   |
|                                                                         |
+------------------------------------------------------------------------*/

void GLOBAL_C::debugMsg (char *Format, ...)
{
    char    ErrMsg[255];
    va_list ArgP;

    va_start (ArgP, Format);

    if (strlen (Format)+1 < sizeof (ErrMsg))
        vsprintf (ErrMsg, Format, ArgP);

    va_end (ArgP);

    donotyield (TRUE);
    printf ("%s", ErrMsg);
    donotyield (FALSE);
}

void GLOBAL_C::debugInfo (char *Format, ...)
{
    if (OptionsP) if (!OptionsP->EnableInfos) return;

    char    Msg[255];
    va_list ArgP;

    va_start (ArgP, Format);

    if (strlen (Format)+1 < sizeof (Msg))
        vsprintf (Msg, Format, ArgP);

    va_end (ArgP);

    HWND hWindow    = GetTopWindow(getMainWnd());
    int  MsgBoxType = MB_APPLMODAL | MB_OK | MB_ICONEXCLAMATION;
    MessageBox (hWindow, Msg, "Debug info", MsgBoxType);
}

BOOL GLOBAL_C::openLog()
{
    char datebuf[9], timebuf[9];

    if (LogFile) fclose (LogFile);

    // Name und Verzeichnis wie Ausgabedatei:

    strncpy (LogFileName, OptionsP->GeomFile, sizeof (LogFileName));
    strcat (LogFileName, ".");
    strcpy (strchr (LogFileName, '.'), ".log");  // extension .LOG

    LogFile = fopen (LogFileName, "wt");         // overwrite existing
    if (!LogFile) return FALSE;

    // Header schreiben:

    _strdate (datebuf);
    _strtime (timebuf);
    fprintf (LogFile, "Comp3D Logfile, Date: %s  Time: %s\n\n",
                      datebuf,timebuf);

    fclose (LogFile);
    LogFile = NULL;

    return TRUE;
}

void GLOBAL_C::beep()
{
    MessageBeep (0);
}

BOOL GLOBAL_C::logMsg (char *Format, ...)

// String in Log-File schreiben

{
    char    Msg [255];
    va_list ArgP;

    // beim ersten mal Datei ffnen:

    if (!LogFile)
        LogFile = fopen (LogFileName, "a+t");    // append

    if (!LogFile) return FALSE;

    va_start (ArgP, Format);

    if (strlen (Format)+1 < sizeof (Msg))
        vsprintf (Msg, Format, ArgP);

    va_end (ArgP);

    fprintf (LogFile, "%s", Msg);

    // aus Sicherheitsgrnden wieder schlieen:

    fclose (LogFile);
    LogFile = NULL;

    return TRUE;
}

BOOL GLOBAL_C::openStatsFile()
{
    char datebuf[9], timebuf[9];

    if (StatsDataFile) fclose (StatsDataFile);

    // Name und Verzeichnis wie Ausgabedatei:

    strncpy (StatsFileName, OptionsP->GeomFile, sizeof (StatsFileName));
    strcat (StatsFileName, ".");
    strcpy (strchr (StatsFileName, '.'), ".prn");// extension .PRN

    StatsDataFile = fopen (StatsFileName, "wt"); // overwrite existing
    if (!StatsDataFile) return FALSE;

    // Header schreiben:

    _strdate (datebuf);
    _strtime (timebuf);
    fprintf (StatsDataFile, "Comp3D runtime stats , Date: %s  Time: %s\n\n",
                            datebuf,timebuf);

    fprintf (StatsDataFile, "\"%s\"\n", OptionsP->InFile);
    fprintf (StatsDataFile, "\"MaxError: %f, ReducePL: %s\"\n\n",
             OptionsP->MaxErrorTri,
             OptionsP->ReducePL ? "TRUE" : "FALSE");
    fprintf (StatsDataFile, "\"Del.Verts\", \"Effizienz\", \"Temperatur\", \"Energy\"\n");
    fclose (StatsDataFile);
    StatsDataFile = NULL;

    return TRUE;
}

BOOL GLOBAL_C::statsOut (char *Format, ...)

// Ausgabe von Laufzeitstatistik. Diese Datei kann von Quattro Pro
// importiert und grafisch aufbereitet werden.

{
    char    Msg [255];
    va_list ArgP;

    // beim ersten mal Datei ffnen:

    if (!StatsDataFile)
        StatsDataFile = fopen (StatsFileName, "a+t");      // append

    if (!StatsDataFile) return FALSE;

    va_start (ArgP, Format);

    if (strlen (Format)+1 < sizeof (Msg))
        vsprintf (Msg, Format, ArgP);

    va_end (ArgP);

    fprintf (StatsDataFile, "%s", Msg);

    // aus Sicherheitsgrnden wieder schlieen:

    fclose (StatsDataFile);
    StatsDataFile = NULL;

    return TRUE;
}

void GLOBAL_C::info (char *Format, ...)
{
    char    Msg[255];
    va_list ArgP;

    va_start (ArgP, Format);
    if (strlen (Format)+1 < sizeof (Msg))
        vsprintf (Msg, Format, ArgP);
    va_end (ArgP);

    HWND hWindow    = GetTopWindow(getMainWnd());
    int  MsgBoxType = MB_APPLMODAL | MB_OK | MB_ICONEXCLAMATION;
    MessageBox (hWindow, Msg, "INFO", MsgBoxType);
}

void GLOBAL_C::warning (char *Format, ...)
{
    char    Msg[255];
    va_list ArgP;

    va_start (ArgP, Format);
    if (strlen (Format)+1 < sizeof (Msg))
        vsprintf (Msg, Format, ArgP);
    va_end (ArgP);

    beep();

    HWND hWindow = GetTopWindow(getMainWnd());
    int MsgBoxType = MB_APPLMODAL | MB_OK | MB_ICONSTOP;
    MessageBox (hWindow, Msg, "WARNING", MsgBoxType);
}

void GLOBAL_C::fatal (char *Format, ...)
{
    char    ErrMsg[255];
    va_list ArgP;

    va_start (ArgP, Format);
    vsprintf (ErrMsg, Format, ArgP);
    va_end (ArgP);

    strcat (ErrMsg, "\nCANCEL will terminate program");

//  FatalAppExit (0, ErrMsg);

    beep();

    HWND hWindow = GetTopWindow(getMainWnd());
    int MsgBoxType = MB_APPLMODAL | MB_OKCANCEL | MB_DEFBUTTON1 | MB_ICONHAND;
    int Answer = MessageBox (hWindow, ErrMsg, "INTERNAL ERROR", MsgBoxType);

    if (Answer==IDCANCEL) exit (EXIT_FAILURE);
}

void GLOBAL_C::wait()
{
    WaitCnt++;
    SetCursor (LoadCursor (0, waiting() ? IDC_WAIT : IDC_ARROW ) );
}

void GLOBAL_C::noWait()
{
    WaitCnt--;
    SetCursor (LoadCursor (0, waiting() ? IDC_WAIT : IDC_ARROW ) );
}

long GLOBAL_C::getTimeMark()
{
    LastTicks = GetTickCount();

    return LastTicks;
}

float GLOBAL_C::elapsed (long LastTimeMark)
{
    return (float)(GetTickCount()-LastTimeMark)/1000.0;
}

float GLOBAL_C::elapsed()
{
    return elapsed (LastTicks);
}

static char ElapsedStr[20];

char *GLOBAL_C::elapsedStrHMSH (float aSecs)
{
    int Hour = aSecs/3600.0;
    int Min  = aSecs/60.0 - Hour*60;
    int Sec  = (int)aSecs % 60;
    int S100 = (int) fmod (100.0 * aSecs, 100.0);
    sprintf (ElapsedStr, "%i:%02i:%02i.%02i", Hour, Min, Sec, S100);
    return ElapsedStr;
}

char *GLOBAL_C::elapsedStrHMS (float aSecs)
{
    int Hour = aSecs/3600.0;
    int Min  = aSecs/60.0 - Hour*60;
    int Sec  = (int)aSecs % 60;
    sprintf (ElapsedStr, "%i:%02i:%02i", Hour, Min, Sec);
    return ElapsedStr;
}

char *GLOBAL_C::elapsedStrMSH (float aSecs)
{
    int Min  = aSecs/60.0;
    int Sec  = (int)aSecs % 60;
    int S100 = (int) fmod (100.0 * aSecs, 100.0);
    sprintf (ElapsedStr, "%i:%02i.%02i", Min, Sec, S100);
    return ElapsedStr;
}

static EnableWmCompactingInfo = FALSE;

void GLOBAL_C::wmCompacting (WORD wParam)
{

    if (TrisetP)
        TrisetP->compactFaces();

    if ( ! EnableWmCompactingInfo) return;

    char Msg[255];
    HWND hWindow    = GetTopWindow(getMainWnd());
    int  MsgBoxType = MB_APPLMODAL | MB_OKCANCEL | MB_ICONSTOP | MB_DEFBUTTON1;

    sprintf (Msg,
        "Compacting (%li%%), %i tasks running\nCANCEL to suppress this info",
        (long)100*wParam/0x10000L, GetNumTasks() );

    if (MessageBox (hWindow, Msg, "System memory is low", MsgBoxType) == IDCANCEL)

        EnableWmCompactingInfo = FALSE;

}

/*------------------------------------------------------------------------+
|                                                                         |
| Objekt deklarieren                                                      |
|                                                                         |
+------------------------------------------------------------------------*/

GLOBAL_C Global;
