/**************************************************************************
*                                                                         *
* Name   : cHistWin.cpp                ( Headerdatei:                 )   *
*                                                                         *
* Klasse(n) definiert : HISTORYWINDOW_C                                   *
* Objekt(e) deklariert: -                                                 *
* Abgeleitet von      : -                                                 *
*                                                                         *
* Aufgabe:                                                                *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include "cTypes.hpp"

#include <math.h>
#include <mem.h>
#include <stdlib.h>
#include <string.h>
#include <values.h>

#include "cGlobal.hpp"
#include "cGlue.hpp"
#include "cMath.hpp"
#include "cHistWin.hpp"

// --- lokale Konstanten:----------------------------------------------

// --- lokale Typen: --------------------------------------------------

// --- lokale Variablen: ----------------------------------------------


/*------------------------------------------------------------------------+
|                                                                         |
| lokale Funktionen                                                       |
|                                                                         |
+------------------------------------------------------------------------*/


/*------------------------------------------------------------------------+
|                                                                         |
| Message handler                                                         |
|                                                                         |
+------------------------------------------------------------------------*/

long  FAR PASCAL _export histWndProc (HWND hWnd, UINT Message, UINT wParam, LONG lParam)
{
    HISTORYWINDOW_C *P;

    if (Message==WM_CREATE)
        Global.GlueP->addGlue (hWnd, ((CREATESTRUCT*)lParam)->lpCreateParams);

    P = (HISTORYWINDOW_C*) Global.GlueP->getThisFromWnd (hWnd);

    if (!P) return DefWindowProc (hWnd, Message, wParam, lParam);

    switch (Message) {

        case WM_DESTROY:
            Global.GlueP->delGlue (hWnd);
            break;
        case WM_PAINT:
            P->paint (hWnd);
            break;
	default:
	    return DefWindowProc (hWnd, Message, wParam, lParam);
    }
    return 0; // Message processed
}

/*------------------------------------------------------------------------+
|                                                                         |
| Window Initialisierung                                                  |
|                                                                         |
+------------------------------------------------------------------------*/

static char ClassName[] = "Comp3DHistWinClass";

int HISTORYWINDOW_C::registerWindow (HANDLE hInstance)
{
    WNDCLASS WndClass;
    WORD     Style = CS_BYTEALIGNCLIENT | CS_OWNDC | CS_HREDRAW | CS_VREDRAW;

    WndClass.style         = Style;
    WndClass.lpfnWndProc   = histWndProc;
    WndClass.cbClsExtra    = 0;
    WndClass.cbWndExtra    = 0;
    WndClass.hInstance     = hInstance;
    WndClass.hIcon         = NULL;
    WndClass.hCursor       = LoadCursor(NULL, IDC_ARROW);
    WndClass.hbrBackground = hBkBrush;
    WndClass.lpszMenuName  = NULL;
    WndClass.lpszClassName = ClassName;

    return (RegisterClass(&WndClass));
}

HWND HISTORYWINDOW_C::createWindow (HANDLE  hInstance, int CmdShow, HWND hParent)
{
    HWND  hWnd;

    char *Title = "History Graph";
    long Style  = WS_CHILD | WS_OVERLAPPEDWINDOW | WS_CLIPSIBLINGS;

    hWnd = CreateWindow (

	ClassName,                            // Class name
        Title,
        Style,
	CW_USEDEFAULT, CW_USEDEFAULT,         // Window position
        400, 200,                             // Window size
	hParent,                              // Parent window handle
	NULL,                                 // use class menu
	hInstance,                            // Task handle
	(LPSTR)this                           // passed in lParam of WM_CREATE message
    );

    if (!hWnd) return 0;

    ShowWindow   (hWnd, CmdShow);
    UpdateWindow (hWnd);
    return hWnd;
}


/*------------------------------------------------------------------------+
|                                                                         |
| Methoden der Klasse HISTORYWINDOW_C                                     |
|                                                                         |
+------------------------------------------------------------------------*/

static HistoryWindowCRegistered = FALSE;

HISTORYWINDOW_C::HISTORYWINDOW_C (HWND hParent, int CmdShow, int MaxN)
{
    int I;

    hWnd          = NULL;
    hDC           = NULL;

    hBkBrush      = CreateSolidBrush (RGB (255, 255, 255));          // wei
    hPen[0]       = CreatePen (PS_SOLID, 1, RGB ( 255,   0,   0));   // rot
    hPen[1]       = CreatePen (PS_SOLID, 1, RGB (   0, 255,   0));   // grn
    hPen[2]       = CreatePen (PS_SOLID, 1, RGB (   0,   0, 255));   // blau
    hPen[3]       = CreatePen (PS_SOLID, 1, RGB ( 255,   0, 255));   // magenta
    hPen[4]       = CreatePen (PS_SOLID, 1, RGB (   0, 255, 255));   // cyan
    hAxisPen      = CreatePen (PS_DOT,   1, RGB (  64,  64,  64));   // dk grau

    strncpy (Legend[0], "Deleted", sizeof (Legend[0]));
    strncpy (Legend[1], "Efficiency", sizeof (Legend[1]));
    strncpy (Legend[2], "Temp.", sizeof (Legend[2]));
    strncpy (Legend[3], "Energy", sizeof (Legend[3]));
    strncpy (Legend[4], "", sizeof (Legend[4]));

    for (I=0; I<NHISTORIES; I++)
        HistBuffer [I] = NULL;

    UpdateTicks   = 2000;
    ElapsedTicks  = 0;

    MaxCnt        = MaxN;

    NCnt          = 0;
    TotalCnt      = 0;
    setRangeY (MAXFLOAT, MINFLOAT);
    Filter        = 0x0001;

    if (! HistoryWindowCRegistered)
        if ( registerWindow (Global.getInstance()) ) {

            HistoryWindowCRegistered = TRUE;
        }
    hWnd = createWindow (Global.getInstance(), CmdShow, hParent);
    if (!hWnd) return;

    hDC = GetDC (hWnd);
    if (!hDC) return;

    for (I=0; I<NHISTORIES; I++) {

        HistBuffer[I] = (float*) calloc (MaxCnt, sizeof (float));
    }
}

HISTORYWINDOW_C::~HISTORYWINDOW_C()            // Destruktor
{
    if (hBkBrush) DeleteObject (hBkBrush);
    if (hPen[0])  DeleteObject (hPen[0]);
    if (hPen[1])  DeleteObject (hPen[1]);
    if (hPen[2])  DeleteObject (hPen[2]);
    if (hPen[3])  DeleteObject (hPen[3]);
    if (hPen[4])  DeleteObject (hPen[4]);
    if (hAxisPen) DeleteObject (hAxisPen);

    for (int I=0; I<NHISTORIES; I++) {

        if (HistBuffer [I]) {

            free (HistBuffer [I]);
            HistBuffer [I] = NULL;
        }
    }
    if (IsWindow (hWnd)) {

        if (hDC) {

            ReleaseDC (hWnd, hDC);
            hDC  = NULL;
        }
        DestroyWindow (hWnd);
    }
    hWnd = NULL;
}

void HISTORYWINDOW_C::clear()
{
    Filter   = 0x0001;
    NCnt     = 0;
    TotalCnt = 0;
    repaint();
}

void HISTORYWINDOW_C::paint (HWND hWnd)
{
    PAINTSTRUCT PS;
    RECT        R;

    if (!hDC) {

        BeginPaint (hWnd, &PS);
        EndPaint (hWnd, &PS);
        return;
    }

    BeginPaint (hWnd, &PS);

    GetClientRect (hWnd, &R);
    int         Width  = R.right-R.left;
    int         Height = R.bottom-R.top;

    SetBkMode (hDC, TRANSPARENT);

    if (HistBuffer[0] && NCnt>1) {

        float ScaleY = ScaleH * (float)(Height-1);
        float ScaleX = (float)(Width-1)/(NCnt-1);

        // Achsen:

        SelectObject (hDC, hAxisPen);
        MoveTo (hDC, 0, Height-1 - ScaleY*(50.0-MinVal));             // 50%
        LineTo (hDC, ScaleX*NCnt, Height-1 - ScaleY*(50.0-MinVal));
        MoveTo (hDC, 0, Height-1 - ScaleY*(100.0-MinVal));            // 100%
        LineTo (hDC, ScaleX*NCnt, Height-1 - ScaleY*(100.0-MinVal));
        SelectObject (hDC, GetStockObject (BLACK_PEN));
        MoveTo (hDC, 0, Height-1 - ScaleY*(MinVal));                  // 0%
        LineTo (hDC, ScaleX*NCnt, Height-1 - ScaleY*(MinVal));

        // Kurvenzge:

        for (int I=0; I<NHISTORIES; I++) {

            SelectObject (hDC, hPen[I]);
            MoveTo (hDC, 0, Height-1 - ScaleY*(HistBuffer[I][0]-MinVal));

            for (int X=1; X<NCnt; X++) {

                LineTo (hDC, ScaleX * X, Height-1 - ScaleY*(HistBuffer[I][X]-MinVal));
            }
        }
    }

    // Legende:

    int X = Width / NHISTORIES / 10;
    int Y = 10;

    if (!IsIconic (hWnd)) {

        SetTextColor (hDC, RGB (255, 0, 0));
        TextOut (hDC, X, Y, Legend[0], strlen (Legend[0]));
        X += Width / NHISTORIES;
        SetTextColor (hDC, RGB (0, 255, 0));
        TextOut (hDC, X, Y, Legend[1], strlen (Legend[1]));
        X += Width / NHISTORIES;
        SetTextColor (hDC, RGB (0, 0, 255));
        TextOut (hDC, X, Y, Legend[2], strlen (Legend[2]));
        X += Width / NHISTORIES;
        SetTextColor (hDC, RGB (255, 0, 255));
        TextOut (hDC, X, Y, Legend[3], strlen (Legend[3]));
        X += Width / NHISTORIES;
        SetTextColor (hDC, RGB (0, 255, 255));
        TextOut (hDC, X, Y, Legend[4], strlen (Legend[4]));
    }

    EndPaint (hWnd, &PS);
}

void HISTORYWINDOW_C::repaint()
{
    if (hWnd) {
        InvalidateRect (hWnd, NULL, TRUE);
        UpdateWindow (hWnd);
    }
}

BOOL HISTORYWINDOW_C::compactVals()

// Immer zwei Werte werden auf eine Position gemittelt. der 1. Wert bleibt
// erhalten. Der Filter wird verndert, so da nur noch halb soviel Punkte
// wi vorher durchgelassen werden.

{
    if ( Filter&0x8000 ) return FALSE;

    Filter = Filter << 1;

    for (int I=1; I<((MaxCnt+1)/2); I++) {

        int I2 = 2*I;

        HistBuffer[0][I] = (HistBuffer[0][I2] + HistBuffer[0][I2-1]) / 2.0;
        HistBuffer[1][I] = (HistBuffer[1][I2] + HistBuffer[1][I2-1]) / 2.0;
        HistBuffer[2][I] = (HistBuffer[2][I2] + HistBuffer[2][I2-1]) / 2.0;
        HistBuffer[3][I] = (HistBuffer[3][I2] + HistBuffer[3][I2-1]) / 2.0;
        HistBuffer[4][I] = (HistBuffer[4][I2] + HistBuffer[4][I2-1]) / 2.0;
    }
    NCnt = NCnt / 2;
    return TRUE;
}

void HISTORYWINDOW_C::setRangeY (float Min, float Max)
{
    MaxVal   = Max;
    MinVal   = Min;
}

BOOL HISTORYWINDOW_C::addHistory  (float F1, float F2, float F3, float F4, float F5)
{
    if (!hWnd)
        return TRUE;

    if (NCnt>=MaxCnt)
        if (!compactVals())
            return FALSE;

    TotalCnt++;

    if (TotalCnt%Filter!=0)
        return TRUE;

    HistBuffer[0][NCnt] = F1;
    HistBuffer[1][NCnt] = F2;
    HistBuffer[2][NCnt] = F3;
    HistBuffer[3][NCnt] = F4;
    HistBuffer[4][NCnt] = F5;

    NCnt++;

    float MaxF = MAX (MAX (F1, F2), MAX (F3, MAX (F4, F5)));
    float MinF = MIN (MIN (F1, F2), MIN (F3, MIN (F4, F5)));

    MaxVal = MAX (MaxVal, MaxF);
    MinVal = MIN (MinVal, MinF);

    if (MaxVal==MinVal)
        ScaleH = 1.0;
    else
        ScaleH = 1.0 / fabs (MaxVal-MinVal);

    DWORD TC = GetTickCount();

    if (TC - ElapsedTicks > UpdateTicks) {

        ElapsedTicks = TC;
        repaint();
    }

    return TRUE;
}

BOOL HISTORYWINDOW_C::iconize()
{
    return ShowWindow (hWnd, SW_MINIMIZE);
}

BOOL HISTORYWINDOW_C::maximize()
{
    return ShowWindow (hWnd, SW_SHOWMAXIMIZED);
}

