/**************************************************************************
*                                                                         *
* Name   : IniFile.cpp                 ( Headerdatei:                 )   *
*                                                                         *
* Klasse(n) definiert : INIFILE_C                                         *
* Objekt(e) deklariert:                                                   *
*                                                                         *
* Abgeleitet von      :                                                   *
*                                                                         *
* Aufgabe:                                                                *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include "cTypes.hpp"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "cIniFile.hpp"


// --- lokale Konstanten:----------------------------------------------

// --- lokale Typen: --------------------------------------------------

// --- lokale Variablen: ----------------------------------------------


/*------------------------------------------------------------------------+
|                                                                         |
| lokale Funktionen                                                       |
|                                                                         |
+------------------------------------------------------------------------*/

/*------------------------------------------------------------------------+
|                                                                         |
| Methoden der Klasse INIFILE_C                                           |
|                                                                         |
+------------------------------------------------------------------------*/

INIFILE_C::INIFILE_C (char *aFName)             // Konstruktor
{
    strncpy ( FName, aFName, sizeof (FName) );

    if (!strchr (FName, '.'))
        strcat (FName, ".INI");

    setApplicationName ("_default_");
}

INIFILE_C::~INIFILE_C()            // Destruktor
{
}

void INIFILE_C::setApplicationName (char *aName)
{
    strncpy ( Application, aName, sizeof (Application) );
}

char* INIFILE_C::getString (char *aKeyName, char *aDefault)
{
    Status = INI_OK;

    GetPrivateProfileString (Application, aKeyName, "_NotFound_", Buffer, sizeof (Buffer), FName);
    if ( strcmp ("_NotFound_", Buffer) == 0 ) {
        Status = INI_NOTFOUND;
        return aDefault;
    }
    return Buffer;
}

int INIFILE_C::getInt (char *aKeyName, int aDefault)
// GetPrivateProfileInt liefert nur positive, deswegen:
{
    return (int) getLong (aKeyName, (int)aDefault);
}

BOOL INIFILE_C::getBool   (char *aKeyName, BOOL aDefault)
{
    getString (aKeyName, "");
    if (Status) return aDefault;

    if (stricmp (Buffer, "1") == 0) return TRUE;
    if (stricmp (Buffer, "JA") == 0) return TRUE;
    if (stricmp (Buffer, "YES") == 0) return TRUE;
    if (stricmp (Buffer, "TRUE") == 0) return TRUE;
    if (stricmp (Buffer, "ON") == 0) return TRUE;

    if (stricmp (Buffer, "0") == 0) return FALSE;
    if (stricmp (Buffer, "NEIN") == 0) return FALSE;
    if (stricmp (Buffer, "NO") == 0) return FALSE;
    if (stricmp (Buffer, "FALSE") == 0) return FALSE;
    if (stricmp (Buffer, "OFF") == 0) return FALSE;

    Status = INI_TYPEERROR;
    return aDefault;
}

long INIFILE_C::getLong   (char *aKeyName, long aDefault)
{
    getString (aKeyName, "");
    if (Status) return aDefault;
    return atol (Buffer);
}
float INIFILE_C::getFloat (char *aKeyName, float aDefault)
{
    getString (aKeyName, "");
    if (Status) return aDefault;
    return atof (Buffer);
}

double INIFILE_C::getDouble (char *aKeyName, double aDefault)
{
    getString (aKeyName, "");
    if (Status) return aDefault;
    return _atold (Buffer);
}

POINT INIFILE_C::getPoint  (char *aKeyName, POINT  aDefault)
{
    POINT P;

    getString (aKeyName, "");
    if (Status) return aDefault;
    if ( sscanf (Buffer, "%i %i", &P.x, &P.y)!=2 ) {
        Status = INI_TYPEERROR;
        return aDefault;
    }
    return P;
}

RECT INIFILE_C::getRect   (char *aKeyName, RECT   aDefault)
{
    RECT R;

    getString (aKeyName, "");
    if (Status) return aDefault;
    if ( sscanf (Buffer, "%i %i %i %i", &R.left, &R.top, R.right, R.bottom)!=4 ) {
        Status = INI_TYPEERROR;
        return aDefault;
    }
    return R;
}

BOOL INIFILE_C::writeString (char *aKeyName, char* aValue)
{
    Status = INI_OK;

    if ( strlen (aValue)==0 ) {
        Status = INI_TYPEERROR;
        return FALSE;
    }
    if (WritePrivateProfileString (Application, aKeyName, aValue, FName) ) {
        Status = INI_NOTFOUND;
        return FALSE;
    }
    return TRUE;
}

BOOL INIFILE_C::writeInt (char *aKeyName, int aValue)
{
    return writeString (aKeyName, itoa (aValue, Buffer, 10) );
}

BOOL INIFILE_C::writeBoolYN (char *aKeyName, BOOL aValue)
{
    return writeString (aKeyName, aValue ? "Yes" : "No");
}

BOOL INIFILE_C::writeBoolJN (char *aKeyName, BOOL aValue)
{
    return writeString (aKeyName, aValue ? "Ja" : "Nein");
}

BOOL INIFILE_C::writeBoolOO (char *aKeyName, BOOL aValue)
{
    return writeString (aKeyName, aValue ? "ON" : "OFF");
}

BOOL INIFILE_C::writeBoolTF (char *aKeyName, BOOL aValue)
{
    return writeString (aKeyName, aValue ? "TRUE" : "FALSE");
}

BOOL INIFILE_C::writeLong   (char *aKeyName, long aValue)
{
    return writeString (aKeyName, ltoa (aValue, Buffer, 10) );
}

BOOL INIFILE_C::writeFloat (char *aKeyName, float aValue, int aDecimals)
{
    return writeString (aKeyName, gcvt (aValue, aDecimals, Buffer) );
}

BOOL INIFILE_C::writeDouble (char *aKeyName, double aValue, int aDecimals)
{
    return writeString (aKeyName, gcvt (aValue, aDecimals, Buffer) );
}

BOOL INIFILE_C::writePoint  (char *aKeyName, POINT *aValue)
{
    sprintf (Buffer, "%i %i", aValue->x, aValue->y);
    return writeString (aKeyName, Buffer);
}

BOOL INIFILE_C::writeRect   (char *aKeyName, RECT *aValue)
{
    sprintf (Buffer, "%i %i %i %i", aValue->left, aValue->top,
                                    aValue->right, aValue->bottom);
    return writeString (aKeyName, Buffer);
}

BOOL INIFILE_C::deleteKey (char *aKeyName)
{
    Status = INI_OK;
    if (WritePrivateProfileString (Application, aKeyName, NULL, FName) ) return TRUE;
    Status = INI_NOTFOUND;
    return FALSE;
}

