/**************************************************************************
*                                                                         *
* Name   : cMath.cpp                   ( Headerdatei:                 )   *
*                                                                         *
* Klasse(n) definiert : -                                                 *
*                                                                         *
* Aufgabe:                                                                *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include "cTypes.hpp"

#include <math.h>
#include <mem.h>

#include "cGlobal.hpp"   // nur fr Warnings
#include "cMath.hpp"

// --- globale Konstanten:---------------------------------------------

MATRIX44_T Identity44 = { { 1.0, 0.0, 0.0, 0.0 },
                          { 0.0, 1.0, 0.0, 0.0 },
                          { 0.0, 0.0, 1.0, 0.0 },
                          { 0.0, 0.0, 0.0, 1.0 } };

MATRIX44_T Zero44     = { { 0.0, 0.0, 0.0, 0.0 },
                          { 0.0, 0.0, 0.0, 0.0 },
                          { 0.0, 0.0, 0.0, 0.0 },
                          { 0.0, 0.0, 0.0, 0.0 } };
Ppoint3    Zero3      = { 0.0, 0.0, 0.0 };

Pfloat Pi   = 3.141592653589793238;
Pfloat Sqr2 = sqrt (2.0);
Pfloat Sqr3 = sqrt (3.0);
Pfloat Ln2  = log (2.0);

/*------------------------------------------------------------------------+
|                                                                         |
| Funktionen                                                              |
|                                                                         |
+------------------------------------------------------------------------*/

void makeIdentity44 (MATRIX44_T M)
{
    memcpy (M, Identity44, sizeof (Identity44) );
}

void makeNull44 (MATRIX44_T M)
{
    memset (M, 0, sizeof (MATRIX44_T) );
}

void copyMatrix44 (MATRIX44_T Source, MATRIX44_T Dest)
{
    memcpy (Dest, Source, sizeof (MATRIX44_T) );
}

BOOL ptsEqual (Ppoint3 &P1, Ppoint3 &P2)
{
    return ( (P1.x==P2.x) && (P1.y==P2.y) && (P1.z==P2.z) );
}

BOOL isNull3 (Ppoint3 &P)
{
//  return ( memcmp ((void*)&P, &Zero3, sizeof (Ppoint3)) == 0 ); // funktioniert nicht fr -0
    return ( (P.x==0.0) && (P.y==0.0) && (P.z==0.0) );
}

Ppoint3 add3 (Ppoint3 &P1, Ppoint3 &P2)
{
    Ppoint3 R;

    R.x = P1.x + P2.x; R.y = P1.y + P2.y; R.z = P1.z + P2.z;

    return R;
}

Ppoint3 sub3 (Ppoint3 &P1, Ppoint3 &P2)
{
    Ppoint3 R;

    R.x = P1.x - P2.x; R.y = P1.y - P2.y; R.z = P1.z - P2.z;

    return R;
}

Pfloat abs3 (Ppoint3 &P)
{
    Pfloat R = P.x*P.x + P.y*P.y +P.z*P.z;

    return (R==0.0) ? 0.0 : sqrt (R);
}

Ppoint3 crossProduct (Ppoint3 &P1, Ppoint3 &P2)
// vgl. Mortenson s. 7
{
    Ppoint3 R;

    R.x = P1.y * P2.z - P1.z * P2.y;
    R.y = P1.z * P2.x - P1.x * P2.z;
    R.z = P1.x * P2.y - P1.y * P2.x;

    return R;
}

void multMatrix44 (MATRIX44_T Result, MATRIX44_T A, MATRIX44_T B)
{
    makeNull44 (Result);
    for (int I=0; I<4; I++)
        for (int J=0; J<4; J++)
            for (int K=0; K<4; K++)
                Result [I][J] += A[I][K]*B[K][J];
}

void transformPt44 (VECTOR4_T Result, MATRIX44_T M, VECTOR4_T Pt)
{
    for (int I=0; I<4; I++) {
        Result [I] = 0.0;
        for (int J=0; J<4; J++)
            Result [I] += M [I][J] * Pt [J];
    }
}

void transformXY44 (Ppoint3 &Result, MATRIX44_T M, int X, int Y)

// Eingabe: X, Y (Annahme Z == 0, W == 1)
// Ausgabe: Normalisierter Vektor (X', Y', Z') mit W = 1

{
    Pfloat fX = (Pfloat)X;
    Pfloat fY = (Pfloat)Y;

    Result.x = M[0][0] * fX + M[0][1] * fY + M[0][3];
    Result.y = M[1][0] * fX + M[1][1] * fY + M[1][3];
    Result.z = M[2][0] * fX + M[2][1] * fY + M[2][3];

    if ( (M[3][0]!=0) || (M[3][1]!=0) || M[3][3]!=1 ) {    // ggf. Normalisieren
        Pfloat Q = M[3][0] * fX + M[3][1] * fY + M[3][3];
        Result.x /= Q;
        Result.y /= Q;
        Result.z /= Q;
    }
}

Pfloat sqrDistance3 (Ppoint3 &P1, Ppoint3 &P2)
{
    Pfloat dX = P1.x - P2.x;
    Pfloat dY = P1.y - P2.y;
    Pfloat dZ = P1.z - P2.z;
    return dX*dX + dY*dY + dZ*dZ;
}

Pfloat distance3 (Ppoint3 &P1, Ppoint3 &P2)

// vgl GEMS, s. 432 (schnelle Aproximation)

{
    Pfloat dX = P1.x - P2.x;
    Pfloat dY = P1.y - P2.y;
    Pfloat dZ = P1.z - P2.z;
    Pfloat XYZ2 = dX*dX + dY*dY + dZ*dZ;
    if (XYZ2>0.0)
        return sqrt (XYZ2);
    else
        return 0.0;
}

Pfloat distancePtLn3 (Ppoint3 &Q, Ppoint3 &P1, Ppoint3 &P2)
// krzester Abstand zwischen P und Linie durch P1,P2
// vgl. Mortenson S.222, van Dam S.1100
{
    Pfloat DX = P2.x -  P1.x;
    Pfloat DY = P2.y -  P1.y;
    Pfloat DZ = P2.z -  P1.z;
    Pfloat XYZ2 = DX*DX + DY*DY + DZ*DZ;

    if (XYZ2==0.0) return 0.0;         // Kollinear !

    Pfloat U = ( DX*(Q.x-P1.x) + DY*(Q.y-P1.y) + DZ*(Q.z-P1.z) ) / XYZ2;

    Ppoint3 P;                         // Punkt auf der Geraden mit krzestem Abstand

    P.x = P1.x + U*DX;
    P.y = P1.y + U*DY;
    P.z = P1.z + U*DZ;

    return abs3 ( sub3 (P, Q) );
}

Pfloat distancePtPlane3 (Ppoint3 &P,  Ppoint3 &P1, Ppoint3 &P2, Ppoint3 &P3)
// kleinster Abstand zwischen P und der Flche durch das Dreieck P1,P2,P3
// vgl. Mortenson s.224
{
    Ppoint3 K = sub3 (P2, P1);
    Ppoint3 L = sub3 (P3, P1);
    Ppoint3 N = crossProduct (K, L);   // Flchennormale (kein Einheitsvektor)

    if (isNull3 (N)) return distancePtLn3 (P, P1, P2);

    Pfloat  A = abs3 (N);              // Betrag (N), Grundflche P1P2P3

    // Determinanten fr Cramersche Regel:

    Ppoint3 Q  = sub3 (P1, P);
    Pfloat  D0 = K.x*L.y*N.z + K.y*L.z*N.x + K.z*L.x*N.y -
                 K.x*L.z*N.y - K.y*L.x*N.z - K.z*L.y*N.x;
    Pfloat  D3 = K.x*L.y*Q.z + K.y*L.z*Q.x + K.z*L.x*Q.y -
                 K.x*L.z*Q.y - K.y*L.x*Q.z - K.z*L.y*Q.x;

    // Hhe von P ber der Flche:

    if (D0==0.0) {
        Global.warning ("distancePtPlane (...): entartetes Dreieck.");
        return 0.0;                   // keine Lsung (entartetes Dreieck ?)
        }
    else
        return fabs (D3 / D0 * A);
}

Pfloat distancePtTri3 (Ppoint3 &P,  Ppoint3 &P1, Ppoint3 &P2, Ppoint3 &P3)

// Abstand zwischen P und dem Dreieck P1,P2,P3. Wenn P ber dem Dreieck liegt, ist das
// Ergebnis das Gleiche wie bei distancePtPlane3. Sonst wird der Abstand zur nchsten Edge
// bzw. Vertex zurckgegeben.

{
    Ppoint3 K = sub3 (P2, P1);
    Ppoint3 L = sub3 (P3, P1);
    Ppoint3 N = crossProduct (K, L);   // Flchennormale (kein Einheitsvektor)

    if (isNull3 (N)) return distancePtLn3 (P, P1, P2);

    // Determinanten fr Cramersche Regel:

    Ppoint3 Q  = sub3 (P, P1);

    Pfloat  D0 = K.x*L.y*N.z + K.y*L.z*N.x + K.z*L.x*N.y -
                 K.x*L.z*N.y - K.y*L.x*N.z - K.z*L.y*N.x;
    Pfloat  D1 = Q.x*L.y*N.z + Q.y*L.z*N.x + Q.z*L.x*N.y -
                 Q.x*L.z*N.y - Q.y*L.x*N.z - Q.z*L.y*N.x;
    Pfloat  D2 = K.x*Q.y*N.z + K.y*Q.z*N.x + K.z*Q.x*N.y -
                 K.x*Q.z*N.y - K.y*Q.x*N.z - K.z*Q.y*N.x;

    if (D0==0.0) {
        Global.warning ("distancePtTri3 (...): entartetes Dreieck.");
        return 10000.0;
    }

    Pfloat A = D1 / D0;
    Pfloat B = D2 / D0;

    // P' = A*K + B*L. Innerhalb Dreieck ?:

    Pfloat Dist = -1.0;

    if (A>=0.0) {
        if (B>=0.0) {
            if ((A+B)<=1.0) {
                Pfloat  D3 = K.x*L.y*Q.z + K.y*L.z*Q.x + K.z*L.x*Q.y -
                             K.x*L.z*Q.y - K.y*L.x*Q.z - K.z*L.y*Q.x;
                Dist = fabs (D3/D0 * abs3 (N));           // P' liegt im Dreieck !
            }
            else            Dist = distancePtLn3 (P, P2, P3);
            }
        else {
            if ((A+B)<=1.0) Dist = distancePtLn3 (P, P1, P2);
            else            Dist = distance3 (P, P2);
        }
        }
    else {
        if (B>=0.0) {
            if ((A+B)<=1.0) Dist = distancePtLn3 (P, P1, P3);
            else            Dist = distance3 (P, P3);
            }
        else
            Dist = distance3 (P, P1);
    }

    return Dist;
}

Pfloat areaTri3 (Ppoint3 &P1, Ppoint3 &P2, Ppoint3 &P3)
// vgl. Mortenson s.63
{
    Ppoint3 R = crossProduct (sub3 (P2, P1), sub3 (P3, P1));

    return (isNull3 (R)) ? 0.0 : (Pfloat)0.5 * sqrt (R.x*R.x + R.y*R.y + R.z*R.z);
}

Pfloat volumeTetra (Ppoint3 &P1, Ppoint3 &P2, Ppoint3 &P3, Ppoint3 &P4)
{
    return distancePtPlane3 (P1, P2, P3, P4) * areaTri3 (P2, P3, P4) / 3.0;
}

