/**************************************************************************
*                                                                         *
* Name   : cOptions.cpp                ( Headerdatei:                 )   *
*                                                                         *
* Klasse(n) definiert : OPTIONS_C                                         *
* Objekt(e) deklariert: -                                                 *
* Abgeleitet von      : -                                                 *
*                                                                         *
* Aufgabe: Modale Dialogbox zur Einstellung der Systemparameter           *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include "cTypes.hpp"

#include <dir.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cGlobal.hpp"
#include "cGlue.hpp"
#include "cOptions.hpp"
#include "cStatWin.hpp"
#include "cTools.hpp"
#include "cIniFile.hpp"

#include "\src\lib\crtio.h"


// --- public Konstanten:----------------------------------------------

char *ISTypeStr[] = {
    "Sphere",
    "Cylinder"
};

char *MapMethodTypeStr[] = {
    "Centroid",
    "ISN"
};

char *SimulationModeStr[] = {
    "<NONE>",
    "Plane",
    "Plane2",
    "Sphere",
    "Cone",
    "Cylinder",
    "Box"
};


// --- lokale Typen: --------------------------------------------------

// --- lokale Variablen: ----------------------------------------------


/*------------------------------------------------------------------------+
|                                                                         |
| lokale Funktionen                                                       |
|                                                                         |
+------------------------------------------------------------------------*/

/*------------------------------------------------------------------------+
|                                                                         |
| Dialog msg handler                                                      |
|                                                                         |
+------------------------------------------------------------------------*/


BOOL FAR PASCAL _export convReducDlgProc (HWND hDlg, WORD Message, WORD wParam, LONG lParam)
{
    OPTIONS_C *P;

    if (Message==WM_INITDIALOG) {
        Global.GlueP->addGlue (hDlg, (void*)lParam);
        P = (OPTIONS_C*) lParam;
    } else
        P = (OPTIONS_C*) Global.GlueP->getThisFromWnd (hDlg);

    if (!P) return FALSE;

    switch (Message) {

        case WM_INITDIALOG:
            P->updateConvReducDlg (hDlg);
            break;
        case WM_DESTROY:
            Global.GlueP->delGlue (hDlg);
            break;

        case WM_COMMAND:

            switch ( wParam ) {

                case OPTS_PB_INFILE:
                    P->newFName (hDlg, wParam);
                    break;

                case OPTS_CB_FILTER:
                    P->Filtering = IsDlgButtonChecked (hDlg, wParam);
                    break;
                case OPTS_PB_FILTER:
                    P->execFilterDlg(hDlg);
                    P->updateConvReducDlg (hDlg);
                    break;

                case OPTS_CB_REDUCPL:
                    P->ReducePL = IsDlgButtonChecked (hDlg, wParam);
                    break;
                case OPTS_PB_REDUCPL:
                    P->execReducePLDlg(hDlg);
                    P->updateConvReducDlg (hDlg);
                    break;

                case OPTS_CB_REDUCTRI:
                    P->ReduceTri = IsDlgButtonChecked (hDlg, wParam);
                    break;
                case OPTS_PB_REDUCTRI:
                    P->execReduceTriDlg(hDlg);
                    P->updateConvReducDlg (hDlg);
                    break;

                case OPTS_CB_TEXTURE:
                    P->TextureMapping = IsDlgButtonChecked (hDlg, wParam);
                    break;
                case OPTS_PB_TEXTURE:
                    P->execTextureDlg(hDlg);
                    P->updateConvReducDlg (hDlg);
                    break;

                case OPTS_CB_GEOMETRY:
                    P->WriteGeometry = IsDlgButtonChecked (hDlg, wParam);
                    break;
                case OPTS_PB_GEOMETRY:
                    P->execOutputDlg(hDlg);
                    P->updateConvReducDlg (hDlg);
                    break;

                case OPTS_CB_CLOSEOBJ:
                    P->CloseObject = IsDlgButtonChecked (hDlg, wParam);
                    break;

                case OPTS_ED_HASHSIZE:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newLongEdit (hDlg, wParam, P->VertHashTabSize, 0L, 1000000L);
                    break;

                case IDOK:
                    EndDialog (hDlg, 1);         // return TRUE
                    break;
                case IDCANCEL:
                    EndDialog (hDlg, 0);         // return FALSE
                    break;
                default:
                    return FALSE;
            }
            break;
    default:
        return FALSE;
    }
    return TRUE;
}

BOOL FAR PASCAL _export filterDlgProc (HWND hDlg, WORD Message, WORD wParam, LONG lParam)
{
    OPTIONS_C *P;

    if (Message==WM_INITDIALOG) {
        Global.GlueP->addGlue (hDlg, (void*)lParam);
        P = (OPTIONS_C*) lParam;
    } else
        P = (OPTIONS_C*) Global.GlueP->getThisFromWnd (hDlg);

    if (!P) return FALSE;

    switch (Message) {

        case WM_INITDIALOG:
            P->updateFilterDlg (hDlg);
            break;
        case WM_DESTROY:
            Global.GlueP->delGlue (hDlg);
            break;

        case WM_COMMAND:

            switch ( wParam ) {

                case OPTS_ED_FACTR:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->FactR, 4, 2, 0.0, 9.99);
                    break;
                case OPTS_ED_FACTG:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->FactG, 4, 2, 0.0, 9.99);
                    break;
                case OPTS_ED_FACTB:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->FactB, 4, 2, 0.0, 9.99);
                    break;

                case OPTS_ED_MINY:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->ClipMinY, 5, 1, -10.0, 100.0);
                    break;
                case OPTS_ED_MAXY:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->ClipMaxY, 5, 1, -10.0, 100.0);
                    break;

                case IDOK:
                    EndDialog (hDlg, TRUE);
                    break;
                case IDCANCEL:
                    EndDialog (hDlg, FALSE);
                    break;
                default:
                    return FALSE;
            }
            break;
    default:
        return FALSE;
    }
    return TRUE;
}

BOOL FAR PASCAL _export reducePLDlgProc (HWND hDlg, WORD Message, WORD wParam, LONG lParam)
{
    OPTIONS_C *P;

    if (Message==WM_INITDIALOG) {
        Global.GlueP->addGlue (hDlg, (void*)lParam);
        P = (OPTIONS_C*) lParam;
    } else
        P = (OPTIONS_C*) Global.GlueP->getThisFromWnd (hDlg);

    if (!P) return FALSE;

    switch (Message) {

        case WM_INITDIALOG:
            P->updateReducePLDlg (hDlg);
            break;
        case WM_DESTROY:
            Global.GlueP->delGlue (hDlg);
            break;

        case WM_COMMAND:

            switch ( wParam ) {

                case OPTS_ED_MAXERRORPL:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->MaxErrorPL, 4, 2, 0.0, 9.99);
                    break;
                case OPTS_ED_MESHSIZEPL:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->MeshSizePL, 4, 1, 0.0, 99.9);
                    break;

                case IDOK:
                    EndDialog (hDlg, TRUE);
                    break;
                case IDCANCEL:
                    EndDialog (hDlg, FALSE);
                    break;
                default:
                    return FALSE;
            }
            break;
    default:
        return FALSE;
    }
    return TRUE;
}

BOOL FAR PASCAL _export reduceTriDlgProc (HWND hDlg, WORD Message, WORD wParam, LONG lParam)
{
    OPTIONS_C *P;

    if (Message==WM_INITDIALOG) {
        Global.GlueP->addGlue (hDlg, (void*)lParam);
        P = (OPTIONS_C*) lParam;
    } else
        P = (OPTIONS_C*) Global.GlueP->getThisFromWnd (hDlg);

    if (!P) return FALSE;

    switch (Message) {

        case WM_INITDIALOG:
            P->updateReduceTriDlg (hDlg);
            break;
        case WM_DESTROY:
            Global.GlueP->delGlue (hDlg);
            break;

        case WM_COMMAND:

            switch ( wParam ) {

                case OPTS_ED_MAXERRORTRI:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->MaxErrorTri, 4, 2, 0.0, 9.99);
                    break;
                case OPTS_ED_MESHSIZETRI:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->MeshSizeTri, 5, 2, 0.0, 99.99);
                    break;
                case OPTS_ED_PROPORTION:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->WorstProportion, 6, 2, 0.0, 999.99);
                    break;
                case OPTS_CB_KEEPEDGE:
                    P->KeepEdgePoints = IsDlgButtonChecked (hDlg, wParam);
                    break;

                case OPTS_ED_COMPACTTHR:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newLongEdit (hDlg, wParam, P->CompactThresh, 100L, 5000L);
                    break;
                case OPTS_ED_STOPTHR:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newLongEdit (hDlg, wParam, P->StopThresh, 100L, 1000000L);
                    break;
                case OPTS_ED_HITRATEIVAL:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newLongEdit (hDlg, wParam, P->HitRateInterval, 10L, 10000L);
                    break;
                case OPTS_CB_AVOIDDOUBLES:
                    P->AvoidDoubles = IsDlgButtonChecked (hDlg, wParam);
                    break;

                case OPTS_ED_TIMELIMIT:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->TimeLimitHours, 7, 2, 0.0, 9999.99);
                    break;

                case OPTS_ED_PROPFACT:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->ProportionFactor, 7, 2, -999.99, 999.99);
                    break;
                case OPTS_ED_ERRORFACT:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->ErrorFactor, 7, 2, -999.99, 999.99);
                    break;
                case OPTS_ED_VERTFACT:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->VertexFactor, 7, 2, -999.99, 999.99);
                    break;
                case OPTS_ED_AREAFACT:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->AreaFactor, 7, 2, -999.99, 999.99);
                    break;
                case OPTS_ED_SAVESTART:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->AutoSaveStartTime, 7, 2, 0.0, 9999.99);
                    break;
                case OPTS_ED_SAVEPERCENT:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->AutoSavePercentage, 4, 1, 0.5, 99.9);
                    break;

                case OPTS_CB_ANNEALING:
                    P->Annealing = IsDlgButtonChecked (hDlg, wParam);
                    break;
                case OPTS_ED_ANNEALFACT:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->AnnealFactor, 5, 2, 1.00, 99.00);
                    break;

                case IDOK:
                    EndDialog (hDlg, TRUE);
                    break;
                case IDCANCEL:
                    EndDialog (hDlg, FALSE);
                    break;
                default:
                    return FALSE;
            }
            break;
    default:
        return FALSE;
    }
    return TRUE;
}

BOOL FAR PASCAL _export textureDlgProc (HWND hDlg, WORD Message, WORD wParam, LONG lParam)
{
    OPTIONS_C *P;

    if (Message==WM_INITDIALOG) {
        Global.GlueP->addGlue (hDlg, (void*)lParam);
        P = (OPTIONS_C*) lParam;
    } else
        P = (OPTIONS_C*) Global.GlueP->getThisFromWnd (hDlg);

    if (!P) return FALSE;

    switch (Message) {

        case WM_INITDIALOG:
            P->updateTextureDlg (hDlg);
            break;
        case WM_DESTROY:
            Global.GlueP->delGlue (hDlg);
            break;

        case WM_COMMAND:

            switch ( wParam ) {

                case OPTS_PB_TEXFILE:
                    P->newFName (hDlg, wParam);
                    P->updateTextureDlg (hDlg);
                    break;

                case OPTS_ED_H:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->H, 5, 1, 0.0, 100.0);
                    break;
                case OPTS_ED_U:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newIntEdit (hDlg, wParam, P->U, 1, 9999);
                    break;
                case OPTS_ED_V:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newIntEdit (hDlg, wParam, P->V, 1, 9999);
                    break;

                case OPTS_RB_TEXBMP: P->TexFileType = TEX_BMP; break;
                case OPTS_RB_TEXRLE: P->TexFileType = TEX_RLE; break;

                case OPTS_RB_SPHERE:   P->Surface = SURF_SPHERE;   break;
                case OPTS_RB_CYLINDER: P->Surface = SURF_CYLINDER; break;

                case OPTS_RB_CENTROID: P->MapMethod = MAP_CENTROID; break;
                case OPTS_RB_ISN:      P->MapMethod = MAP_ISN;      break;

                case OPTS_CB_INTERPOL:
                    P->Interpolate = IsDlgButtonChecked (hDlg, wParam);
                    break;

                case IDOK:
                    EndDialog (hDlg, 1);         // return TRUE
                    break;
                case IDCANCEL:
                    EndDialog (hDlg, 0);         // return FALSE
                    break;
                default:
                    return FALSE;
            }
            break;
    default:
        return FALSE;
    }
    return TRUE;
}

BOOL FAR PASCAL _export outputDlgProc (HWND hDlg, WORD Message, WORD wParam, LONG lParam)
{
    OPTIONS_C *P;

    if (Message==WM_INITDIALOG) {
        Global.GlueP->addGlue (hDlg, (void*)lParam);
        P = (OPTIONS_C*) lParam;
    } else
        P = (OPTIONS_C*) Global.GlueP->getThisFromWnd (hDlg);

    if (!P) return FALSE;

    switch (Message) {

        case WM_INITDIALOG:
            P->updateOutputDlg (hDlg);
            break;
        case WM_DESTROY:
            Global.GlueP->delGlue (hDlg);
            break;

        case WM_COMMAND:

            switch ( wParam ) {

                case OPTS_PB_GEOMFILE:
                    P->newFName (hDlg, wParam);
                    P->updateOutputDlg (hDlg);
                    break;

                case OPTS_RB_DATANONE: P->GeomFileType = GEOM_DATANONE; break;
                case OPTS_RB_DATARGB:  P->GeomFileType = GEOM_DATARGB;  break;
                case OPTS_RB_DATAUV:   P->GeomFileType = GEOM_DATAUV;   break;

                case OPTS_RB_FORMATZOF:
                    P->GeomFileFormat = GEOM_ZOF;
                    P->updateOutputDlg (hDlg);
                    break;
                case OPTS_RB_FORMATDXF:
                    P->GeomFileFormat = GEOM_DXF;
                    P->updateOutputDlg (hDlg);
                    break;

                case OPTS_CB_GEOMSMALL:
                    P->GeomSmall = IsDlgButtonChecked (hDlg, wParam);
                    break;
                case OPTS_CB_GEOMCOMPR:
                    P->GeomCompress = IsDlgButtonChecked (hDlg, wParam);
                    break;

                case OPTS_CB_SPLITSET:
                    P->SplitTriset = IsDlgButtonChecked (hDlg, wParam);
                    break;
                case OPTS_ED_VERTWEIGHT:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->VertexWeight, 5, 1, 0.0, 100.0);
                    break;
                case OPTS_ED_FACEWEIGHT:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->FaceWeight, 5, 1, 0.0, 100.0);
                    break;
                case OPTS_ED_MAXWEIGHT:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newFloatEdit (hDlg, wParam, P->MaxWeight, 8, 1, 0.0, 100000.0);
                    break;

                case IDOK:
                    EndDialog (hDlg, 1);         // return TRUE
                    break;
                case IDCANCEL:
                    EndDialog (hDlg, 0);         // return FALSE
                    break;
                default:
                    return FALSE;
            }
            break;
    default:
        return FALSE;
    }
    return TRUE;
}

BOOL FAR PASCAL _export debugDlgProc (HWND hDlg, WORD Message, WORD wParam, LONG lParam)
{
    OPTIONS_C *P;

    if (Message==WM_INITDIALOG) {
        Global.GlueP->addGlue (hDlg, (void*)lParam);
        P = (OPTIONS_C*) lParam;
    } else
        P = (OPTIONS_C*) Global.GlueP->getThisFromWnd (hDlg);

    if (!P) return FALSE;

    switch (Message) {

        case WM_INITDIALOG:
            P->updateDebugDlg (hDlg);
            break;
        case WM_DESTROY:
            Global.GlueP->delGlue (hDlg);
            break;

        case WM_COMMAND:

            switch ( wParam ) {

                case DBG_CB_ENABINFOS:
                    P->EnableInfos = IsDlgButtonChecked (hDlg, wParam);
                    break;
                case DBG_CB_MARKFACES:
                    P->MarkFaces = IsDlgButtonChecked (hDlg, wParam);
                    break;
                case DBG_CB_CHKDOUBLES:
                    P->CheckDoubleEdges = IsDlgButtonChecked (hDlg, wParam);
                    break;

                case DBG_ED_UPDSTATSIVAL:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newIntEdit (hDlg, wParam, P->UpdateStatusInterval, 1, 2000);
                    break;
                case DBG_ED_UPDHISTIVAL:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newIntEdit (hDlg, wParam, P->UpdateHistoryInterval, 1, 2000);
                    break;
                case DBG_ED_LOGHISTIVAL:
                    if (HIWORD (lParam)==EN_KILLFOCUS)
                        newIntEdit (hDlg, wParam, P->LogHistoryInterval, 1, 2000);
                    break;

                case IDOK:
                    EndDialog (hDlg, TRUE);
                    break;
                case IDCANCEL:
                    EndDialog (hDlg, FALSE);
                    break;
                default:
                    return FALSE;
            }
            break;
    default:
        return FALSE;
    }
    return TRUE;
}

/*------------------------------------------------------------------------+
|                                                                         |
| Protected Methoden                                                      |
|                                                                         |
+------------------------------------------------------------------------*/

BOOL OPTIONS_C::makeFNames()
{
    char Drive [3];
    char Dir   [66];
    char Name  [9];
    char Ext   [5];

    SimulationMode = SIM_NONE;

    fnsplit (InFile, Drive, Dir, Name, Ext);

    if (strlen (Name)==0) return FALSE;

    if ( ! FileBox.fileExist (InFile) ) {

        if (stricmp (Name, "plane")==0)         SimulationMode = SIM_PLANE;
        else if (stricmp (Name, "plane2")==0)   SimulationMode = SIM_PLANE2;
        else if (stricmp (Name, "sphere")==0)   SimulationMode = SIM_SPHERE;
        else if (stricmp (Name, "cone")==0)     SimulationMode = SIM_CONE;
        else if (stricmp (Name, "cylinder")==0) SimulationMode = SIM_CYLINDER;
        else if (stricmp (Name, "box")==0)      SimulationMode = SIM_BOX;

        if (SimulationMode!=SIM_NONE) {

            strcpy (InFile,   "Simulation: ");
            strcat (InFile,   SimulationModeStr [SimulationMode]);
            strcpy (TexFile,  "TEMP.BMP");
            strcpy (GeomFile, "TEMP.ZOF");
            return TRUE;
        }

        Global.warning ("File '%s' not found.", InFile);
        return FALSE;
    }

    fnmerge (GeomFile, Drive, Dir, Name, ".zof");
    fnmerge (TexFile, Drive, Dir, Name, ".bmp");

    return TRUE;
}

/*------------------------------------------------------------------------+
|                                                                         |
| Message Server                                                          |
|                                                                         |
+------------------------------------------------------------------------*/

void OPTIONS_C::newFName (HWND hDlg, WORD wParam)
{
    static DWORD InfileFilterSelect   = 1L;
    static DWORD TexfileFilterSelect  = 1L;
    static DWORD GeomfileFilterSelect = 1L;

    char *szInFilter[]   = { "Spex3D binary",  "*.cef",
                             "ARENA ascii",    "*.zof\0" };

    char *szTexFilter[]  = { "24 bit DIB BGR", "*.bmp\0" };
    char *szGeomFilter[] = { "ARENA ascii",    "*.zof",
                             "AutoCAD ascii",  "*.dxf\0" };

Retry:

    switch (wParam) {

        case OPTS_PB_INFILE:
            do {
                // nicht openExistingFileDlg, damit 'sphere' etc erkannt werden:

                if (SimulationMode!=SIM_NONE)
                    InFile[0] = 0;

                if ( ! FileBox.openFileDlg (hDlg, " Open Input File ",
                                            InFile, sizeof (InFile),
                                            szInFilter, InfileFilterSelect,
                                            NULL) )
                    break;
            } while ( ! makeFNames() );
            break;

        case OPTS_PB_GEOMFILE:
            FileBox.saveFileDlg (hDlg, " Save Geometry to : ",
                                 GeomFile, sizeof (GeomFile),
                                 szGeomFilter, GeomfileFilterSelect, NULL);
            break;

        case OPTS_PB_TEXFILE:
            FileBox.saveFileDlg (hDlg, " Save Texture to : ",
                                 TexFile, sizeof (TexFile),
                                 szTexFilter, TexfileFilterSelect, "bmp");
            break;
    }

    updateConvReducDlg (hDlg);

    if ( (wParam==OPTS_PB_INFILE) && (FileBox.fileExist (GeomFile)) ) {

        int Ans;
        int  MsgBoxType = MB_APPLMODAL | MB_YESNO | MB_ICONQUESTION | MB_DEFBUTTON2;
        Ans = MessageBox (NULL, "Overwrite geometry file ?", "Confirm", MsgBoxType);

        if (Ans==IDNO) {

            wParam = OPTS_PB_GEOMFILE;
            goto Retry;
        }
    }
}

BOOL OPTIONS_C::updateConvReducDlg (HWND hDlg)
{
    char Name[15], Ext[5], TempStr[50];

    // 'do:' - Marker updaten:

    CheckDlgButton   (hDlg, OPTS_CB_FILTER,   Filtering);
    CheckDlgButton   (hDlg, OPTS_CB_REDUCPL,  ReducePL);
    CheckDlgButton   (hDlg, OPTS_CB_REDUCTRI, ReduceTri);
    CheckDlgButton   (hDlg, OPTS_CB_TEXTURE,  TextureMapping);
    CheckDlgButton   (hDlg, OPTS_CB_GEOMETRY, WriteGeometry);

    // Infozeilen updaten:

    SetDlgItemText  (hDlg, OPTS_ST_INFILE, (strlen (InFile)==0) ?
                     "<NONE>" : InFile);
    vSetDlgItemText (hDlg, OPTS_ST_FILTER,
                     "R*%3.1f, G*%3.1f, B*%3.1f, %3.1f < Y < %3.1f",
                     FactR, FactG, FactB, ClipMinY, ClipMaxY);
    vSetDlgItemText (hDlg, OPTS_ST_REDUCPL,
                     "Error: %.2f, Mesh size: %.1f", MaxErrorPL, MeshSizePL);
    vSetDlgItemText (hDlg, OPTS_ST_REDUCTRI,
                     "Error: %.2f, Mesh size: %.1f, %s",
                     MaxErrorTri, MeshSizeTri,
                     Annealing ? "Anneal" : "Direct"
                    );


    fnsplit (TexFile, NULL, NULL, Name, Ext);
    strcpy (TempStr, strcat (Name,Ext));

    vSetDlgItemText (hDlg, OPTS_ST_TEXTURE,
                     "%s, %s/%s, H=%.1f",
                     (strlen (TempStr)==0) ? "<NONE>" : TempStr,
                     ISTypeStr[Surface],
                     MapMethodTypeStr[MapMethod],
                     H
                    );

    fnsplit (GeomFile, NULL, NULL, Name, Ext);
    strcpy (TempStr, strcat (Name,Ext));

    vSetDlgItemText (hDlg, OPTS_ST_GEOMETRY,
                     "%s, %s, %s",
                     (strlen (TempStr)==0) ? "<NONE>" : TempStr,
                     (GeomFileType==GEOM_DATANONE) ? "XYZ" :
                         ( (GeomFileType==GEOM_DATARGB) ? "XYZ/RGB" : "XYZ/UV" ),
                     SplitTriset ? "Split" : "NoSplit"
                    );

    CheckDlgButton (hDlg, OPTS_CB_CLOSEOBJ, CloseObject);

    setDlgItemLong (hDlg, OPTS_ED_HASHSIZE, VertHashTabSize);


    return TRUE;
}

BOOL OPTIONS_C::updateFilterDlg (HWND hDlg)
{
    vSetDlgItemText (hDlg, OPTS_ED_FACTR, "%5.2f", FactR);
    vSetDlgItemText (hDlg, OPTS_ED_FACTG, "%5.2f", FactG);
    vSetDlgItemText (hDlg, OPTS_ED_FACTB, "%5.2f", FactB);
    vSetDlgItemText (hDlg, OPTS_ED_MINY,  "%6.2f", ClipMinY);
    vSetDlgItemText (hDlg, OPTS_ED_MAXY,  "%6.2f", ClipMaxY);

    return TRUE;
}

BOOL OPTIONS_C::updateReducePLDlg (HWND hDlg)
{
    vSetDlgItemText (hDlg, OPTS_ED_MAXERRORPL, "%5.2f", MaxErrorPL);
    vSetDlgItemText (hDlg, OPTS_ED_MESHSIZEPL, "%5.2f", MeshSizePL);

    return TRUE;
}

BOOL OPTIONS_C::updateReduceTriDlg (HWND hDlg)
{
    vSetDlgItemText (hDlg, OPTS_ED_MAXERRORTRI, "%5.2f", MaxErrorTri);
    vSetDlgItemText (hDlg, OPTS_ED_MESHSIZETRI, "%5.2f", MeshSizeTri);
    vSetDlgItemText (hDlg, OPTS_ED_PROPORTION,  "%5.2f", WorstProportion);
    CheckDlgButton  (hDlg, OPTS_CB_KEEPEDGE, KeepEdgePoints);

    setDlgItemLong  (hDlg, OPTS_ED_COMPACTTHR, CompactThresh);
    setDlgItemLong  (hDlg, OPTS_ED_STOPTHR, StopThresh);
    setDlgItemLong  (hDlg, OPTS_ED_HITRATEIVAL, HitRateInterval);
    CheckDlgButton  (hDlg, OPTS_CB_AVOIDDOUBLES, AvoidDoubles);

    vSetDlgItemText (hDlg, OPTS_ED_VERTFACT,  "%.2f", VertexFactor);
    vSetDlgItemText (hDlg, OPTS_ED_ERRORFACT,  "%.2f", ErrorFactor);
    vSetDlgItemText (hDlg, OPTS_ED_PROPFACT,  "%.2f", ProportionFactor);
    vSetDlgItemText (hDlg, OPTS_ED_AREAFACT,  "%.2f", AreaFactor);
    vSetDlgItemText (hDlg, OPTS_ED_SAVESTART, "%5.2f", AutoSaveStartTime);
    vSetDlgItemText (hDlg, OPTS_ED_SAVEPERCENT, "%5.2f", AutoSavePercentage);

    CheckDlgButton  (hDlg, OPTS_CB_ANNEALING, Annealing);
    vSetDlgItemText (hDlg, OPTS_ED_ANNEALFACT,  "%5.2f", AnnealFactor);
    vSetDlgItemText (hDlg, OPTS_ED_TIMELIMIT,  "%7.2f", TimeLimitHours);

    return TRUE;
}

BOOL OPTIONS_C::updateTextureDlg (HWND hDlg)
{
    int Temp;

    SetDlgItemText (hDlg, OPTS_ST_TEXFILE, (strlen (TexFile)==0) ?
                    "<NONE>" : TexFile);

    switch (Surface) {
        case SURF_SPHERE:   Temp = OPTS_RB_SPHERE; break;
        case SURF_CYLINDER:
        default :           Temp = OPTS_RB_CYLINDER; break;
    }
    CheckRadioButton (hDlg, OPTS_RB_CYLINDER, OPTS_RB_SPHERE, Temp);

    switch (MapMethod) {
        case MAP_CENTROID: Temp = OPTS_RB_CENTROID; break;
        case MAP_ISN:
        default :          Temp = OPTS_RB_ISN; break;
    }
    CheckRadioButton (hDlg, OPTS_RB_CENTROID, OPTS_RB_ISN, Temp);

/*  switch (TexFileType) {
        case TEX_BMP: Temp = OPTS_RB_TEXBMP; break;
        case TEX_RLE:
        default :     Temp = OPTS_RB_TEXRLE; break;
    }
    CheckRadioButton (hDlg, OPTS_RB_TEXBMP, OPTS_RB_TEXRLE, Temp);
*/

    CheckDlgButton   (hDlg, OPTS_CB_INTERPOL, Interpolate);

    vSetDlgItemText (hDlg, OPTS_ED_H, "%5.1f", H);
    SetDlgItemInt  (hDlg, OPTS_ED_U, U, FALSE);
    SetDlgItemInt  (hDlg, OPTS_ED_V, V, FALSE);

    return TRUE;
}

BOOL OPTIONS_C::updateOutputDlg (HWND hDlg)
{
    int Temp;

    SetDlgItemText (hDlg, OPTS_ST_GEOMFILE, (strlen (GeomFile)==0) ?
                    "<NONE>" : GeomFile);

    if (strstr (strupr(GeomFile), ".DXF")) {

        GeomFileType = GEOM_DATANONE;
        GeomFileFormat = GEOM_DXF;
        }
    else
        if (strstr (strupr(GeomFile), ".ZOF")) {

            GeomFileFormat = GEOM_ZOF;
        }

    switch (GeomFileType) {
        case GEOM_DATANONE: Temp = OPTS_RB_DATANONE; break;
        case GEOM_DATARGB:  Temp = OPTS_RB_DATARGB; break;
        case GEOM_DATAUV:
        default :           Temp = OPTS_RB_DATAUV; break;
    }
    CheckRadioButton (hDlg, OPTS_RB_DATANONE, OPTS_RB_DATAUV, Temp);

    switch (GeomFileFormat) {
        case GEOM_DXF: Temp = OPTS_RB_FORMATDXF; break;
        case GEOM_ZOF:
        default :      Temp = OPTS_RB_FORMATZOF; break;
    }
    CheckRadioButton (hDlg, OPTS_RB_FORMATDXF, OPTS_RB_FORMATZOF, Temp);

    if (GeomFileFormat==GEOM_DXF) {

        EnableWindow (GetDlgItem (hDlg, OPTS_RB_DATARGB), FALSE);
        EnableWindow (GetDlgItem (hDlg, OPTS_RB_DATAUV), FALSE);
        }
    else {

        EnableWindow (GetDlgItem (hDlg, OPTS_RB_DATARGB), TRUE);
        EnableWindow (GetDlgItem (hDlg, OPTS_RB_DATAUV), TRUE);
    }

    CheckDlgButton   (hDlg, OPTS_CB_GEOMSMALL, GeomSmall);
    CheckDlgButton   (hDlg, OPTS_CB_GEOMCOMPR, GeomCompress);
    CheckDlgButton   (hDlg, OPTS_CB_SPLITSET,  SplitTriset);

    vSetDlgItemText  (hDlg, OPTS_ED_VERTWEIGHT, "%7.2f", VertexWeight);
    vSetDlgItemText  (hDlg, OPTS_ED_FACEWEIGHT, "%7.2f", FaceWeight);
    vSetDlgItemText  (hDlg, OPTS_ED_MAXWEIGHT,  "%7.2f", MaxWeight);

    return TRUE;
}

BOOL OPTIONS_C::updateDebugDlg (HWND hDlg)
{
    CheckDlgButton (hDlg, DBG_CB_ENABINFOS, EnableInfos);
    CheckDlgButton (hDlg, DBG_CB_MARKFACES, MarkFaces);
    CheckDlgButton (hDlg, DBG_CB_CHKDOUBLES, CheckDoubleEdges);

    SetDlgItemInt (hDlg, DBG_ED_UPDSTATSIVAL, UpdateStatusInterval,  TRUE);
    SetDlgItemInt (hDlg, DBG_ED_UPDHISTIVAL,  UpdateHistoryInterval, TRUE);
    SetDlgItemInt (hDlg, DBG_ED_LOGHISTIVAL,  LogHistoryInterval,   TRUE);

    return TRUE;
}

/*------------------------------------------------------------------------+
|                                                                         |
| Methoden der Klasse OPTIONS_C                                           |
|                                                                         |
+------------------------------------------------------------------------*/

OPTIONS_C::OPTIONS_C()             // Konstruktor
{
    IniFileP = new INIFILE_C ("Comp3D");

    if (!IniFileP) {
        Global.fatal ("Can't open .INI file");
        return;
    }

    initOptions();

    if (CrtioIconic)
        ShowWindow (getcrtiowindow(), SW_SHOWMINNOACTIVE);
}

OPTIONS_C::~OPTIONS_C()            // Destruktor
{
    saveOptions();

    if (IniFileP) delete IniFileP;

    Global.OptionsP = NULL;
}

BOOL OPTIONS_C::initOptions()
{
    IniFileP->setApplicationName ("MainOptions");

    // Input:

    strcpy (InFile, "");

    // Filterung:

    Filtering           = IniFileP->getBool ("Filtering", FALSE);

    FactR               = IniFileP->getFloat ("FactR", 1.0);
    FactG               = IniFileP->getFloat ("FactG", 1.0);
    FactB               = IniFileP->getFloat ("FactB", 1.0);
    ClipMinY            = IniFileP->getFloat ("ClipMinY", 0.0);
    ClipMaxY            = IniFileP->getFloat ("ClipMaxY", 100.0);

    // Reduce Polyline:

    ReducePL            = IniFileP->getBool ("ReducePL", FALSE);
    MaxErrorPL          = IniFileP->getFloat ("MaxErrorPL", 0.05);
    MeshSizePL          = IniFileP->getFloat ("MeshSizePL", 50.0);

    // Reduce Triset:

    ReduceTri           = IniFileP->getBool ("ReduceTri", TRUE);
    MaxErrorTri         = IniFileP->getFloat ("MaxErrorTri", 0.05);
    MeshSizeTri         = IniFileP->getFloat ("MeshSizeTri", 50.0);
    WorstProportion     = IniFileP->getFloat ("WorstProportion", 20.0);
    KeepEdgePoints      = IniFileP->getBool  ("KeepEdgePoints", FALSE);

    CompactThresh       = IniFileP->getLong  ("CompactThresh", 1000L);
    StopThresh          = IniFileP->getLong  ("StopThresh", 2000L);
    HitRateInterval     = IniFileP->getLong  ("HitRateInterval", 500L);
    AvoidDoubles        = IniFileP->getBool  ("AvoidDoubles", FALSE);

    Annealing           = IniFileP->getBool  ("Annealing", TRUE);
    AnnealFactor        = IniFileP->getFloat ("AnnealFactor", 90.0);
    TimeLimitHours      = IniFileP->getFloat ("TimeLimitHours", 0.0);
    VertexFactor        = IniFileP->getFloat ("VertexFactor", 1.0);
    ErrorFactor         = IniFileP->getFloat ("ErrorFactor", 0.0);
    ProportionFactor    = IniFileP->getFloat ("ProportionFactor", 1.0);
    AreaFactor          = IniFileP->getFloat ("AreaFactor", 0.0);
    AutoSaveStartTime   = IniFileP->getFloat ("AutoSaveStartTime", 1.0);
    AutoSavePercentage  = IniFileP->getFloat ("AutoSavePercentage", 1.0);

    // Texturemapping:

    TextureMapping      = IniFileP->getBool  ("TextureMapping", TRUE);
    strcpy (TexFile, "TEMP.BMP");
    Surface             = (ISTYPE_E)
                              IniFileP->getInt   ("Surface", SURF_CYLINDER);
    MapMethod           = (MAPMETHOD_E)
                              IniFileP->getInt   ("MapMethod", MAP_ISN);
    TexFileType         = (TEXFILETYPE_E)
                              IniFileP->getInt   ("TexFileType", TEX_BMP);
    Interpolate         = IniFileP->getBool  ("Interpolate", TRUE);
    H                   = IniFileP->getFloat ("H", 50.0);
    U                   = IniFileP->getInt   ("U", 360);
    V                   = IniFileP->getInt   ("V", 400);

    // Geometry Output:

    WriteGeometry       = IniFileP->getBool  ("WriteGeometry", TRUE);
    strcpy (GeomFile, "TEMP.ZOF");
    GeomFileType        = (GEOMFILETYPE_E)
                              IniFileP->getInt   ("GeomFileType", GEOM_DATARGB);
    GeomFileFormat      = (GEOMFILEFORMAT_E)
                              IniFileP->getInt   ("GeomFileFormat", GEOM_ZOF);
    GeomSmall           = IniFileP->getBool  ("GeomSmall", TRUE);
    GeomCompress        = IniFileP->getBool  ("GeomCompress", FALSE);
    SplitTriset         = IniFileP->getBool  ("SplitTriset", TRUE);
    VertexWeight        = IniFileP->getFloat ("VertexWeight", 24.0);
    FaceWeight          = IniFileP->getFloat ("FaceWeight", 12.0);
    MaxWeight           = IniFileP->getFloat ("MaxWeight", 60000.0);

    // Debugging:

    HistoryGraph        = IniFileP->getBool  ("HistoryGraph", TRUE);
    HistoryIconic       = IniFileP->getBool  ("HistoryIconic", FALSE);
    LogCefRead          = IniFileP->getBool  ("LogCefRead", FALSE);
    LogSettings         = IniFileP->getBool  ("LogSettings", TRUE);
    LogHistory          = IniFileP->getBool  ("LogHistory", TRUE);
    EnableInfos         = IniFileP->getBool  ("EnableInfos", TRUE);
    MarkFaces           = IniFileP->getBool  ("MarkFaces", TRUE);
    CheckDoubleEdges       = IniFileP->getBool  ("CheckDoubleEdges", TRUE);
    UpdateStatusInterval= IniFileP->getInt   ("UpdateStatusInterval", 20);
    UpdateHistoryInterval=IniFileP->getInt   ("UpdateHistoryInterval", 100);
    LogHistoryInterval  = IniFileP->getInt   ("LogHistoryInterval", 300);
    CrtioIconic         = IniFileP->getBool  ("CrtioIconic", FALSE);
    StatusIconic        = IniFileP->getBool  ("StatusIconic", FALSE);

    // Sonstige:

    SimulationMode      = SIM_NONE;
    VertHashTabSize     = IniFileP->getLong  ("VertHashTabSize", 10000L);
    CloseObject         = IniFileP->getBool  ("CloseObject", FALSE);
    StopThresh          = IniFileP->getLong  ("StopThresh", 2000L);

    return TRUE;
}

BOOL OPTIONS_C::saveOptions()
{
    IniFileP->setApplicationName ("MainOptions");

    // Filterung:

    IniFileP->writeBoolYN  ("Filtering", Filtering);
    IniFileP->writeFloat   ("FactR", FactR);
    IniFileP->writeFloat   ("FactG", FactG);
    IniFileP->writeFloat   ("FactB", FactB);
    IniFileP->writeFloat   ("ClipMinY", ClipMinY);
    IniFileP->writeFloat   ("ClipMaxY", ClipMaxY);

    // Reduce Polyline:

    IniFileP->writeBoolYN  ("ReducePL", ReducePL);
    IniFileP->writeFloat   ("MaxErrorPL", MaxErrorPL);
    IniFileP->writeFloat   ("MeshSizePL", MeshSizePL);

    // Reduce Triset:

    IniFileP->writeBoolYN  ("ReduceTri", ReduceTri);
    IniFileP->writeFloat   ("MaxErrorTri", MaxErrorTri);
    IniFileP->writeFloat   ("MeshSizeTri", MeshSizeTri);
    IniFileP->writeFloat   ("WorstProportion", WorstProportion);
    IniFileP->writeBoolYN  ("KeepEdgePoints", KeepEdgePoints);

    IniFileP->writeLong    ("CompactThresh", CompactThresh);
    IniFileP->writeLong    ("StopThresh", StopThresh);
    IniFileP->writeLong    ("HitRateInterval", HitRateInterval);
    IniFileP->writeBoolYN  ("AvoidDoubles", AvoidDoubles);
    IniFileP->writeBoolYN  ("Annealing", Annealing);
    IniFileP->writeFloat   ("AnnealFactor", AnnealFactor);
    IniFileP->writeFloat   ("TimeLimitHours", TimeLimitHours);
    IniFileP->writeFloat   ("ProportionFactor", ProportionFactor);
    IniFileP->writeFloat   ("ErrorFactor", ErrorFactor);
    IniFileP->writeFloat   ("VertexFactor", VertexFactor);
    IniFileP->writeFloat   ("AreaFactor", AreaFactor);
    IniFileP->writeFloat   ("AutoSaveStartTime", AutoSaveStartTime);
    IniFileP->writeFloat   ("AutoSavePercentage", AutoSavePercentage);

    // Texturemapping:

    IniFileP->writeBoolYN  ("TextureMapping", TextureMapping);
    IniFileP->writeInt     ("Surface", Surface);
    IniFileP->writeInt     ("MapMethod", MapMethod);
    IniFileP->writeInt     ("TexFileType", TexFileType);
    IniFileP->writeBoolYN  ("Interpolate", Interpolate);
    IniFileP->writeFloat   ("H", H);
    IniFileP->writeInt     ("U", U);
    IniFileP->writeInt     ("V", V);

    // Geometry Output:

    IniFileP->writeBoolYN  ("WriteGeometry", WriteGeometry);
    IniFileP->writeInt     ("GeomFileType", GeomFileType);
    IniFileP->writeInt     ("GeomFileFormat", GeomFileFormat);
    IniFileP->writeBoolYN  ("GeomSmall", GeomSmall);
    IniFileP->writeBoolYN  ("GeomCompress", GeomCompress);
    IniFileP->writeBoolYN  ("SplitTriset", SplitTriset);
    IniFileP->writeFloat   ("VertexWeight", VertexWeight);
    IniFileP->writeFloat   ("FaceWeight", FaceWeight);
    IniFileP->writeFloat   ("MaxWeight", MaxWeight);

    // Debugging:

    IniFileP->writeBoolYN  ("HistoryGraph", HistoryGraph);
    IniFileP->writeBoolYN  ("HistoryIconic", HistoryIconic);
    IniFileP->writeBoolYN  ("LogCefRead", LogCefRead);
    IniFileP->writeBoolYN  ("LogSettings", LogSettings);
    IniFileP->writeBoolYN  ("LogHistory", LogHistory);
    IniFileP->writeBoolOO  ("EnableInfos", EnableInfos);
    IniFileP->writeBoolYN  ("MarkFaces", MarkFaces);
    IniFileP->writeBoolYN  ("CheckDoubleEdges", CheckDoubleEdges);
    IniFileP->writeInt     ("UpdateStatusInterval", UpdateStatusInterval);
    IniFileP->writeInt     ("UpdateHistoryInterval", UpdateHistoryInterval);
    IniFileP->writeInt     ("LogHistoryInterval", LogHistoryInterval);
    IniFileP->writeBoolYN  ("CrtioIconic", CrtioIconic);
    IniFileP->writeBoolYN  ("StatusIconic", StatusIconic);

    // Sonstige:

    IniFileP->writeLong    ("VertHashTabSize", VertHashTabSize);
    IniFileP->writeBoolYN  ("CloseObject", CloseObject);
    IniFileP->writeLong    ("StopThresh", StopThresh);

    return TRUE;
}

void OPTIONS_C::logSettings()
{
    if ( !LogSettings ) return;

    Global.logMsg ("Input\t\t: %s\n", InFile);

    Global.logMsg ("\nFiltering\t: %s\n", Filtering ? "TRUE" : "FALSE");
    if (Filtering) {
        Global.logMsg ("R*%f, G*%f, B*%f, %f <= Y <= %f",
                       FactR, FactG, FactB, ClipMinY, ClipMaxY);
    }

    Global.logMsg ("\nTexture mapping\t: %s\n",
                   TextureMapping ? "TRUE" : "FALSE");
    if (TextureMapping) {
        Global.logMsg ("Texture\t\t: %s\n", TexFile);
        Global.logMsg ("Mapping\t\t: %s, %s\n", ISTypeStr [Surface],
                                          MapMethodTypeStr [MapMethod]);
        Global.logMsg ("Interpolate\t: %s\n",  Interpolate ? "TRUE" : "FALSE");
        Global.logMsg ("U x V\t\t: %i x %i\n",  U, V);
    }

    Global.logMsg ("\nReducePl\t: %s\n",  ReducePL  ? "TRUE" : "FALSE");
    if (ReducePL) {
        Global.logMsg ("MaxError\t: %f\n", (float)MaxErrorPL);
        Global.logMsg ("MeshSize\t: %f\n", (float)MeshSizePL);
    }

    Global.logMsg ("\nReduceTri\t: %s\n", ReduceTri ? "TRUE" : "FALSE");
    if (ReduceTri) {
        Global.logMsg ("MaxError\t: %f\n", (float)MaxErrorTri);
        Global.logMsg ("MeshSize\t: %f\n", (float)MeshSizeTri);
        Global.logMsg ("Worst proportn\t: %f\n", (float)WorstProportion);
        Global.logMsg ("KeepEdgePoints\t: %s\n", KeepEdgePoints ? "TRUE" : "FALSE");

        Global.logMsg ("Compact thresh\t: %li\n", CompactThresh);
        Global.logMsg ("Stop thresh\t: %li\n", StopThresh);
        Global.logMsg ("HitRateIval\t: %li\n", HitRateInterval);

        Global.logMsg ("\nAvoidDoubles\t: %s\n", AvoidDoubles ? "TRUE" : "FALSE");
        Global.logMsg ("\nAnnealing\t: %s\n", Annealing ? "TRUE" : "FALSE");
        if (Annealing) {
            Global.logMsg ("Cooling step\t: %f\%\n", (float)AnnealFactor);
            Global.logMsg ("Time limit\t: %f hrs\n", (float)TimeLimitHours);
            Global.logMsg ("VertexFactor\t: %f\n", (float)VertexFactor);
            Global.logMsg ("ErrorFactor\t: %f\n", (float)ErrorFactor);
            Global.logMsg ("Proport.Factor\t: %f\n", (float)ProportionFactor);
            Global.logMsg ("AreaFactor\t: %f\n", (float)AreaFactor);
        }
    }

    Global.logMsg ("\nGeometry\t: %s\n", GeomFile);
}

int OPTIONS_C::execConvReducDlg (HWND hCaller)
{
    HANDLE hI = Global.getInstance();

    int Ret = DialogBoxParam ( hI, "DLG_CONVREDUC", hCaller,
                               MakeProcInstance ((FARPROC)convReducDlgProc, hI),
                               (LONG)this );

    // Statusanzeige updaten:

    Global.StatWinP->TexGeom   = ISTypeStr [Surface];
    Global.StatWinP->TexMethod = MapMethodTypeStr [MapMethod];

    char Name[15], Ext[5];

    fnsplit (InFile, NULL, NULL, Name, Ext);
    strcpy (Global.StatWinP->InName, strcat (Name,Ext));

    fnsplit (GeomFile, NULL, NULL, Name, Ext);
    strcpy (Global.StatWinP->OutName, strcat (Name,Ext));

    fnsplit (TexFile, NULL, NULL, Name, Ext);
    strcpy (Global.StatWinP->TexName, strcat (Name,Ext));

    // Settings in INI-Datei speichern:

    saveOptions();

    return Ret;
}

int OPTIONS_C::execFilterDlg (HWND hCaller)
{
    HANDLE hI = Global.getInstance();

    int Ret = DialogBoxParam ( hI, "DLG_FILTER", hCaller,
                               MakeProcInstance ((FARPROC)filterDlgProc, hI),
                               (LONG)this );

    return Ret;
}

int OPTIONS_C::execReducePLDlg (HWND hCaller)
{
    HANDLE hI = Global.getInstance();

    int Ret = DialogBoxParam ( hI, "DLG_REDUCEPL", hCaller,
                               MakeProcInstance ((FARPROC)reducePLDlgProc, hI),
                               (LONG)this );

    return Ret;
}

int OPTIONS_C::execReduceTriDlg (HWND hCaller)
{
    HANDLE hI = Global.getInstance();

    int Ret = DialogBoxParam ( hI, "DLG_REDUCETRI", hCaller,
                               MakeProcInstance ((FARPROC)reduceTriDlgProc, hI),
                               (LONG)this );

    return Ret;
}

int OPTIONS_C::execTextureDlg (HWND hCaller)
{
    HANDLE hI = Global.getInstance();

    int Ret = DialogBoxParam ( hI, "DLG_TEXTURE", hCaller,
                               MakeProcInstance ((FARPROC)textureDlgProc, hI),
                               (LONG)this );

    return Ret;
}

int OPTIONS_C::execOutputDlg (HWND hCaller)
{
    HANDLE hI = Global.getInstance();

    int Ret = DialogBoxParam ( hI, "DLG_OUTPUT", hCaller,
                               MakeProcInstance ((FARPROC)outputDlgProc, hI),
                               (LONG)this );

    return Ret;
}

int OPTIONS_C::execDebugDlg (HWND hCaller)
{
    HANDLE hI = Global.getInstance();

    int Ret = DialogBoxParam ( hI, "DLG_DEBUG", hCaller,
                               MakeProcInstance ((FARPROC)debugDlgProc, hI),
                               (LONG)this );

    return Ret;
}

