/**************************************************************************
*                                                                         *
* Name   : cPolyLin                    ( Headerdatei:                 )   *
*                                                                         *
* Klasse(n) definiert : NODE_C, NODE3D_C,                                 *
*                       PLLIST_C, DYNARRAY_C,                             *
*                       PLARRAYSET_C                                      *
* Objekt(e) deklariert: -                                                 *
* Abgeleitet von      : -                                                 *
*                                                                         *
* Aufgabe:                                                                *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include "cTypes.hpp"

#include <stdlib.h>
#include <mem.h>

#include "cGlobal.hpp"
#include "cTriset.hpp"
#include "cPolyLin.hpp"

// --- lokale Konstanten:----------------------------------------------

// --- lokale Typen: --------------------------------------------------

// --- lokale Variablen: ----------------------------------------------


/*------------------------------------------------------------------------+
|                                                                         |
| Methoden der Klasse DYNARRAY_C                                          |
|                                                                         |
+------------------------------------------------------------------------*/

static long InstCnt = 0L;

template <class T> DYNARRAY_C<T>::DYNARRAY_C (int aMax, int aGrowBy)     // Konstruktor
{
    InstCnt ++;

    MaxN    = aMax;
    GrowBy  = aGrowBy;
    N       = 0;
    Val = (T*) calloc (aMax, sizeof (T));
    if (!Val) aMax = 0;
}

template <class T> DYNARRAY_C<T>::~DYNARRAY_C()                          // Destruktor
{
    InstCnt --;
    if (Val) free (Val);
}

template <class T> BOOL DYNARRAY_C<T>::append (T &anElement)
{
    if (N>=MaxN)
        if ( ! newSize (MaxN+GrowBy) ) return FALSE;
    Val[N++] = anElement;
    return TRUE;
}

template <class T> BOOL DYNARRAY_C<T>::insertFirst (T &Value)
{
    if (N>=MaxN)
        if ( ! newSize (MaxN+GrowBy) ) return FALSE;

    memmove (&Val[1], &Val[0], N*sizeof(long));

    Val[0] = Value;
    N++;

    return TRUE;
}

template <class T> BOOL DYNARRAY_C<T>::deleteLast()
{
    if (N==0)
        return FALSE;

    N--;

    memset (&Val[N], 0, sizeof (T));

    return TRUE;
}

template <class T> BOOL DYNARRAY_C<T>::deleteBetween (int A, int B)
{
    if ( (A>=MaxN) || (B>=MaxN) ) return FALSE;

    if (B<A) return FALSE;

    memmove (&Val[A+1], &Val[B], (N-B)*sizeof (T) );

    N = N - (B-A-1);

    return TRUE;
}

template <class T> BOOL DYNARRAY_C<T>::copyFrom (DYNARRAY_C<T> *SourceP)
{
    clear();

    if (newSize (SourceP->nElem()))
        memcpy (Val, SourceP->Val, SourceP->nElem()*sizeof (T));
    else
        return FALSE;

    N = SourceP->nElem();
    return TRUE;
}

template <class T> void DYNARRAY_C<T>::clear()
{
    N = 0;
    memset (Val, 0, MaxN*sizeof(T));
}

template <class T> BOOL DYNARRAY_C<T>::adjustSize()
{
    return newSize (N);
}

template <class T> BOOL DYNARRAY_C<T>::newSize (int aVertexCnt)
{
    unsigned NewSize = (unsigned)(aVertexCnt * sizeof (*Val));

    Val = (T*) realloc (Val, NewSize);
    MaxN = aVertexCnt;

    return (Val!=NULL);
}

// Bei dump() mu differenziert werden:

void DYNARRAY_C<long>::dump()
{
    Global.debugMsg ("DYNARRAY_C: this=%p, N=%i, Max=%i, GrowBy=%i", this, N, MaxN, GrowBy);

}

void DYNARRAY_C<Pptco3>::dump()
{
    Global.debugMsg ("DYNARRAY_C: this=%p, N=%i, Max=%i, GrowBy=%i", this, N, MaxN, GrowBy);

    for (int I=0; I<N; I++) {
        Pptco3 P = Val[I];
        Global.debugMsg ("#%4i: (%10.5lf, %10.5lf, %10.5lf): R:%6.3lf G:%6.3lf B:%6.3lf\n",
                          P.point.x, P.point.y, P.point.z,
                          P.colour.x, P.colour.y, P.colour.z );
    }
}

void DYNARRAY_C<FACE_S>::dump()
{
    Global.debugMsg ("DYNARRAY_C<FACE_S>: this=%p, N=%i, Max=%i, GrowBy=%i", this, N, MaxN, GrowBy);

    for (int I=0; I<N; I++) {

        FACE_S F = Val[I];

        Global.debugMsg ("\t#%4i: %li, %li, %li (HV: %i)\n",
                         F.Index[0], F.Index[1], F.Index[2], F.N);
    }
}

/*------------------------------------------------------------------------+
|                                                                         |
| Fehler im Linker: jede Instanz des Templates mu einmal benutzt werden!!|
|                                                                         |
+------------------------------------------------------------------------*/

DYNARRAY_C<Pptco3> DummyPA(1,1);
DYNARRAY_C<long>   DummyLA(1,1);
DYNARRAY_C<FACE_S> DummyFA(1,1);


/*------------------------------------------------------------------------+
|                                                                         |
| Methoden der Klasse POLYGON_C                                           |
|                                                                         |
+------------------------------------------------------------------------*/

POLYGON_C::POLYGON_C (int aMax)     // Konstruktor
{
    N = Max = High = 0;

    VP = (long*) malloc (aMax * sizeof (long));
    if (!VP) Global.warning ("POLYGON_C: out of mem");

    Max = aMax;
}

POLYGON_C::~POLYGON_C()             // Destruktor
{
    if (VP) free (VP);
    Max = N = High = 0;
}

void POLYGON_C::add (long Value)
{
    if (High>=Max) return;
    VP[High++] = Value + 1;            // +1 !!!
    N++;
}

void POLYGON_C::insertFirst (long Value)
{
    if (High>=Max) return;

    memmove (&VP[1], &VP[0], High*sizeof(long));
    High++;

    VP[0] = Value + 1;            // +1 !!!
    N++;
}

void POLYGON_C::clear()
{
    High = N = 0;
}

int POLYGON_C::del (int I)
{
    if ((I>=N) || (I<0)) return -1;

    for (int X=0; X<High; X++) {
        if (VP[X]>0) I--;
        if (I<0) break;
    }
    if (I>=0) return -1;

    VP[X] *= -1;
    N--;
    return X;
}

void POLYGON_C::undel (int Tag)
{
    if ((Tag>=High) || (Tag<0)) return;
    if (VP[Tag]>0) return;
    VP[Tag] *= -1;
    N++;
}

long POLYGON_C::value (int I)           // liefert das I'te gltige Element
{
    if ((I>=N) || (I<0)) return -1;

    for (int X=0; X<High; X++) {
        if (VP[X]>0) I--;
        if (I<0) return VP[X] - 1;     // -1 !!!
    }
    return -1;
}

long POLYGON_C::prec (int I)
{
    return (I==0) ? value (N-1) : value (I-1);
}

long POLYGON_C::succ (int I)
{
    return (I==(N-1)) ? value (0) : value (I+1);
}

