/**************************************************************************
*                                                                         *
* Name   : cReduce.cpp                 ( Headerdatei:                 )   *
*                                                                         *
* Klasse(n) definiert : REDUCE_C                                          *
* Objekt(e) deklariert: -                                                 *
* Abgeleitet von      : -                                                 *
*                                                                         *
* Aufgabe:                                                                *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include "cTypes.hpp"

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <values.h>

#include "cGlobal.hpp"
#include "cReduce.hpp"
#include "cHistWin.hpp"
#include "cMath.hpp"
#include "cOptions.hpp"
#include "cStatWin.hpp"
#include "cTriset.hpp"
#include "cPolyLin.hpp"


// --- lokale Konstanten:----------------------------------------------

static const int     VERTTHRESH     = 7;           // max Verts, bei denen exhaustive search gemacht wird.
static const Pfloat  INFINITYFLOAT  = 10000.0;

// --- lokale Typen: --------------------------------------------------

// --- lokale Variablen: ----------------------------------------------


/*------------------------------------------------------------------------+
|                                                                         |
| lokale Funktionen                                                       |
|                                                                         |
+------------------------------------------------------------------------*/

static Pfloat volumeTetraIdx (long P1, long P2, long P3, long P4)
{
    TRISET_C  *TP = Global.TrisetP;

    return volumeTetra (*TP->getVertPtP (P1), *TP->getVertPtP (P2),
                        *TP->getVertPtP (P3), *TP->getVertPtP (P4) );
}

static Pfloat areaTri3Idx (long P1, long P2, long P3)
{
    TRISET_C  *TP = Global.TrisetP;

    Pfloat A = areaTri3 (*TP->getVertPtP (P1), *TP->getVertPtP (P2), *TP->getVertPtP (P3));

    return A;
}

static Pfloat distancePtTri3Idx (long P, long P1, long P2, long P3)
{
    TRISET_C  *TP = Global.TrisetP;

    Pfloat D = distancePtTri3 (*TP->getVertPtP (P), *TP->getVertPtP (P1),
                               *TP->getVertPtP (P2), *TP->getVertPtP (P3) );
    return D;
}

static Pfloat areaTriset (FACEARRAY_C *FAP)
{
    Pfloat Area = 0.0;

    int    N = FAP->nElem();

    for (int I=0; I<N; I++) {

        long *VP = FAP->Val[I].Index;

        Area += areaTri3Idx (VP[0], VP[1], VP[2]);
    }
    return Area;
}


/*------------------------------------------------------------------------+
|                                                                         |
| Methoden der Klasse REDUCE_C                                            |
|                                                                         |
+------------------------------------------------------------------------*/

REDUCE_C::REDUCE_C()             // Konstruktor
    :PgIdx      (10,5),
     OldSurfIdx (10,5),
     BestSurf   (10,5),
     CurSurf    (10,5),
     PG         (MAXPGSIZE),
     TempIdxList(10,5)
{

    PlKickCnt         = 0;
}

REDUCE_C::~REDUCE_C()            // Destruktor
{
    Global.ReduceP = NULL;
}

void REDUCE_C::init()
{
    MaxErrorPL      = Global.OptionsP->MaxErrorPL;
    MaxMeshSizePL   = Global.OptionsP->MeshSizePL;
    MaxErrorTri     = Global.OptionsP->MaxErrorTri;
    MaxMeshSizeTri  = Global.OptionsP->MeshSizeTri;
    WorstProportion = Global.OptionsP->WorstProportion;
}

/*------------------------------------------------------------------------+
|                                                                         |
| Reduktion einer Polyline                                                |
|                                                                         |
+------------------------------------------------------------------------*/

Pfloat REDUCE_C::greatestDistancePtLn (Pptco3 *V, int A, int B)

// berechnet den grten Abstand zwischen der Linie V[A], V[B] und den
// dazwischenliegenden Punkten des Polygonzugs.

{
    Pfloat D;
    Pfloat MaxD = 0.0;

    for (int I=A+1; I<B; I++) {
        D = distancePtLn3 (V[I].point, V[A].point, V[B].point);
        if (D>MaxD)
            MaxD = D;
    }
    return MaxD;
}

float REDUCE_C::proportion (Ppoint3 *V1P, Ppoint3 *V2P, Ppoint3 *V3P)

// Liefert ein Ma fr die Gleichmigkeit eines Dreiecks.
// 0 <= proportion <= 1;   0: gleichseitig, 1: kollinear

{
    float H;

    static const float twoOverSqrt3 = 2.0/sqrt(3.0);

    float    D12  = distance3 (*V1P, *V2P);
    float    D13  = distance3 (*V1P, *V3P);
    float    D23  = distance3 (*V2P, *V3P);

    float    Dmax = MAX (D12, MAX (D13, D23));

    // H: Hhe ber der lngsten Seite:

    if (Dmax==D12)      { H = distancePtLn3 (*V3P, *V1P, *V2P); }
    else if (Dmax==D13) { H = distancePtLn3 (*V2P, *V1P, *V3P); }
    else                { H = distancePtLn3 (*V1P, *V2P, *V3P); }

    if (Dmax==0.0)
        return 1.0;               // Dreieck ist ein Punkt

    return (1 - twoOverSqrt3 * H / Dmax);
}

BOOL REDUCE_C::isTriValid (long I1, long I2, long I3)

// berprft ein Dreieck auf Zulssigkeit. Ein Dreieck wird zurckgewiesen,
// wenn es zu schmal ist (Verhltnis von grter Seitenlnge zur Hhe ber
// dieser Seite) ODER die grte Seitenlnge grer ist, als die maximal
// erlaubte Maschengre. Ausserdem wird das Dreieck optional verworfen, wenn
// es bereits vorhanden ist.
// Aus Performance-Grnden wird proportion() nicht aufgerufen, sondern berechnet.

{
    Pfloat    H;

    TRISET_C  *TP  = Global.TrisetP;

    Ppoint3   V1   = *TP->getVertPtP (I1);
    Ppoint3   V2   = *TP->getVertPtP (I2);
    Ppoint3   V3   = *TP->getVertPtP (I3);

    Pfloat    D12  = distance3 (V1,V2);
    Pfloat    D13  = distance3 (V1,V3);
    Pfloat    D23  = distance3 (V2,V3);

    Pfloat    Dmax = MAX (D12, MAX (D13, D23));

    if (Dmax > MaxMeshSizeTri) return FALSE;

    // H: Hhe ber der lngsten Seite:

    if (Dmax==D12)      { H = distancePtLn3 (V3, V1, V2); }
    else if (Dmax==D13) { H = distancePtLn3 (V2, V1, V3); }
    else                { H = distancePtLn3 (V1, V2, V3); }

    if (H==0.0) return FALSE;          // Dreieck ist eine Linie

    if ( (Dmax/H) > WorstProportion )
        return FALSE;

    return TRUE;
}

int REDUCE_C::reducePL (VERTEXARRAY_C &PLA)
{
    Pfloat MaxD;
//  Pfloat MaxUsedD;

    int    Start     = 0;              // Indices laufen von 0..PLA.nVerts()-1 !
    Pptco3 *PL       = PLA.Val;
    Pfloat MaxError2 = MaxErrorPL * 2.0;
    int    Kicked    = 0;

    if (!Global.OptionsP->ReducePL) return 0;    // keine Reduktion gewnscht

    if (PLA.nElem()<3) return 0;                // nichts zu tun

    Global.StatWinP->status ("reducing polyline ...");

    do {
        int GoodI = -1;

        for (int I=Start+2; I<PLA.nElem(); I++) {

            if ( distance3 ( PL [Start].point, PL [I].point ) > MaxMeshSizePL )
                break;

            MaxD = greatestDistancePtLn (PL, Start, I);

            if (MaxD>MaxError2)
                break;

            if (MaxD<=MaxErrorPL) {
                GoodI = I;
//              MaxUsedD = MaxD;
            }
        }
        if (GoodI==-1)
            Start++;
        else {
            Kicked += (GoodI - Start - 1);
            PLA.deleteBetween (Start, GoodI);    // nElem() und Indices ndern sich !!
            Start = Start + 1;                   // Start zeigt auf den Punkt der GoodI war
        }

    } while (Start < (PLA.nElem()-2));

    PlKickCnt += Kicked;
    Global.StatWinP->ReducVerts1 = PlKickCnt;

    Global.StatWinP->status ("");
    return Kicked;
}

/*------------------------------------------------------------------------+
|                                                                         |
| Hilfsfunktionen fr Globale Reduktion                                   |
|                                                                         |
+------------------------------------------------------------------------*/

/*------------------------------------------------------------------------+
|                                                                         |
| Reduktion eines Polygons Version I                                      |
|                                                                         |
+------------------------------------------------------------------------*/

void REDUCE_C::distributeHiddenVerts (long         Center,
                                      INDEXARRAY_C *OldSurfIdxP,
                                      FACEARRAY_C  *NewSurfP,
                                      Pfloat       &Error,
                                      Pfloat       &AvgError)

// Verteilt die Hidden Vertizes in *OldSurfIdxP auf die jeweils nchsten Faces
// in *NewSurfP. Der maximale und der durchschnittliche Fehler wird ermittelt.
//   Center >= 0       : Center wird ebenfalls einem Face in *NewSurfP zugeordnet.
//   Center <  0       : Wenn -Center ein Hidden Vertex in *OldSurfIdxP ist, wird
//                       er NICHT verteilt.
//   Center = NOINDEX  : Center wird ignoriert
// Da die neue Tesselierung nicht benutzt wird, wenn Error > MaxError ist,
// kann die Schleife beim ersten Auftreten dieser Bedingung abgebrochen werden.
// In diesem Fall sind *NewSurfP und AvgError nicht gltig!

{
    Pfloat NearestError;
    int    NearestFace;
    int    I;

    TRISET_C *TP    = Global.TrisetP;

    int      Nold   = OldSurfIdxP->nElem();
    int      Nnew   = NewSurfP->nElem();
    Pfloat   Err2Sum = 0.0;
    Pfloat   MaxErr = 0.0;
    int      Cnt    = 0;

    // Da *NewSurfP evtl. schon in vorigen erfolglosen Versuchen benutzt
    // wurd, mssen ggf. vorhandene Hidden-Vertizes gelscht werden:

    for (I=0; I<Nnew; I++) {

        _clearHV ( &(NewSurfP->Val[I]) );
    }

    // Center positiv ? Dann wird er Hidden-Vertex des nchstgelegenen Dreiecks der
    // neuen Tesselierung:

    NearestError = INFINITYFLOAT;

    if ( (Center != NOINDEX) && (Center>=0L) ) {

        for (I=0; I<Nnew; I++) {

            long  *CurFaceIdxP = NewSurfP->Val[I].Index;

            Pfloat E = distancePtTri3Idx (Center, CurFaceIdxP[0], CurFaceIdxP[1],
                                                  CurFaceIdxP[2]);

            if (E<NearestError) {

                NearestError = E;
                NearestFace  = I;
            }
        }
        if ( ! _addHV (&(NewSurfP->Val[NearestFace]), Center) ) {

            Global.fatal ("distributeHiddenVerts: out of mem");
            return;
        }

        if (NearestError>MaxErr) MaxErr = NearestError;

        Err2Sum += (NearestError*NearestError);
        Cnt++;
    }

    // Falls Center negativ ist, soll -Center ignoriert werden, sonst passiert nichts:

    if (Center!=NOINDEX) {
        if (Center<0)
            Center = -Center;
        else
            Center = NOINDEX;
    }

    // Jeder Hidden-Vertex der alten Dreiecke wird einem der neuen Dreiecke
    // zugeordnet:

    for (int F=0; F<Nold; F++) {       // fr jedes alte Dreieck

        FACE_S       *OldFP  = TP->getFaceP (OldSurfIdxP->Val[F]);
        long         *OldHVP = OldFP->HVP;
        int          Vmax    = OldFP->N;

        // fr jeden Hidden Vertex im alten Dreieck:

        for (int V=0; V<Vmax; V++) {

            long HV = OldHVP[V];

            NearestError = INFINITYFLOAT;

            // ggf Center ignorieren

            if ( HV==Center ) {
                continue;
            }

            // nchstgelegenes neues Dreieck suchen:

            for (I=0; I<Nnew; I++) {

                long  *CurFaceIdxP = NewSurfP->Val[I].Index;

                Pfloat E = distancePtTri3Idx (HV, CurFaceIdxP[0],
                                              CurFaceIdxP[1], CurFaceIdxP[2]);

                if (E<NearestError) {

                    NearestError = E;
                    NearestFace  = I;
                }
            }

            if ( ! _addHV (&(NewSurfP->Val[NearestFace]), HV) ) {

                Global.fatal ("distributeHiddenVerts: out of mem");
                return;
            }

            if (NearestError>MaxErr) MaxErr = NearestError;

            Err2Sum += (NearestError*NearestError);
            Cnt++;
        }
    }

    // Statistik:

    Error    = MaxErr;

    AvgError = (Err2Sum>0.0) ? sqrt(Err2Sum) / Cnt : 0.0;
}

/*------------------------------------------------------------------------+
|                                                                         |
| Reduktion des Trisets, minimales Volumen zwischen den Flchen           |
|                                                                         |
+------------------------------------------------------------------------*/

int REDUCE_C::addFaceToCurSurf (int    I,
                                BOOL   &Invalid,
                                Pfloat &Volume)

// Polygonpunkt I wird abgetrennt, das zugehrige Dreieck I-1,I,I+1 wird an
// CurSurf angehngt. OneFaceInvalid und das Volumen werden berechnet.
// Rckgabewert: Tag des gelschten Punktes.
// Nach dem Aufruf mu die nderung mit
//     PG.undel(Tag)
//     CurSurf.deleteLast()
// rckgngig gemacht werden.

{
    FACE_S TempFace;

    long P1 = PG.prec  (I);
    long P2 = PG.value (I);
    long P3 = PG.succ  (I);

    // Flag setzen: ist das abgetrennte Dreieck zulssig ?

    Invalid = !isTriValid (P1, P2, P3);

    // Vertex I vom Polygon abtrennen, entstehendes Dreieck wird an CurSurf
    // angehngt

    TempFace.Index[0] = P1;
    TempFace.Index[1] = P2;
    TempFace.Index[2] = P3;
    TempFace.N        = 0;
    TempFace.MaxN     = 0;
    TempFace.HVP      = NULL;

    CurSurf.append (TempFace);

    // Flcheninkrement berechnen:

    Volume = volumeTetraIdx (CurVert, P1, P2, P3);

    return PG.del (I);
}

void REDUCE_C::tesselateBest (int N, Pfloat AddVolume, BOOL OneFaceInvalid)

{
    FACE_S TempFace;

    Pfloat Volume1, Volume2, Volume3;
    BOOL   Invalid1, Invalid2, Invalid3;
    long   P1, P2, P3;
    int    Tag1, Tag2, Tag3;
    Pfloat Error, AvgError, Area;

    if (N<3) {                         // da ist was faul !
        Global.debugInfo ("tesselateBest: Center Vertex: %li, N=%i", CurVert, N);
        return;
    }

    if (N==3) {

        // Dieses Dreieck ist das letzte des Patches.
        // Die Tesselierung wird verworfen, wenn:
        // 1. es bereits eine Lsung gibt, die ein geringeres Fehlervolumen
        //    hat.
        // 2. Mindestens ein Dreieck zu schmal oder zu gro ist.
        // 3. es bereits eine Lsung mit gleichem Fehlervolumen aber kleinerer
        //    Flche gibt
        // 4. Mindestens ein Punkt (hidden vertex oder Center) weiter als
        //    MaxError von allen neuen Dreiecken entfernt ist.
        // Anderenfalls wird diese Tesselierung nach BestSurf kopiert.

        P1 = PG.value (0);
        P2 = PG.value (1);
        P3 = PG.value (2);

        Volume1 = AddVolume + volumeTetraIdx (CurVert, P1, P2, P3);

        if (Volume1 > BestVolume) {

            return;
        }

        if (Volume1 < BestVolume) {

            BestVolume    = Volume1;
            BestMinArea   = INFINITYFLOAT;
            BestSurfValid = FALSE;
        }

        TempFace.Index[0] = P1;
        TempFace.Index[1] = P2;
        TempFace.Index[2] = P3;
        TempFace.N        = 0;
        TempFace.MaxN     = 0;
        TempFace.HVP      = NULL;

        CurSurf.append (TempFace);

        Area = areaTriset (&CurSurf);

        // wenn die flchenkleinste Tesselierung ungltige Dreiecke enthlt,
        // darf NICHT eine Lsg mit grerer Flche gewhlt werden

        if ( OneFaceInvalid || (! isTriValid (P1, P2, P3) ) ) {

            if (Area<BestMinArea) {

                BestMinArea   = Area;
                BestSurfValid = FALSE;
            }
            goto REJECT;
        }

        distributeHiddenVerts (CurVert, &OldSurfIdx, &CurSurf, Error, AvgError);

        if ( Error > MaxErrorTri ) {

            goto REJECT;
        }

        if ( Area > BestMinArea ) {

            goto REJECT;
        }

        if ( (Area==BestMinArea) && BestSurfValid ) {

            goto REJECT;
        }

        BestMinArea = Area;
        BestSurf.copyFrom (&CurSurf);
        BestSurfValid = TRUE;

    REJECT:

        CurSurf.deleteLast();

        return;

    }

    // N>3: Fr Ecke 0 und 1: Abtrennen und Rekursion

    Tag1 = addFaceToCurSurf (0, Invalid1, Volume1);
    tesselateBest (N-1, AddVolume + Volume1, (OneFaceInvalid || Invalid1) );
    PG.undel (Tag1); CurSurf.deleteLast();

    Tag1 = addFaceToCurSurf (1, Invalid1, Volume1);
    tesselateBest (N-1, AddVolume + Volume1, (OneFaceInvalid || Invalid1) );
    PG.undel (Tag1); CurSurf.deleteLast();


    // N==4: Alle Mglichkeiten schon abgedeckt:

    if (N==4) {

        return;
    }

    // N==5: 1 Mglichkeit fehlt noch:

    if (N==5) {

        Tag1 = addFaceToCurSurf (2, Invalid1, Volume1);
        Tag2 = addFaceToCurSurf (1, Invalid2, Volume2);

        tesselateBest (N-2, AddVolume + Volume1 + Volume2,
                           ( OneFaceInvalid || Invalid1 || Invalid2 ) );
        PG.undel (Tag2);
        PG.undel (Tag1);
        CurSurf.deleteLast();
        CurSurf.deleteLast();

        return;
    }

    // N==6: 4 Mglichkeiten fehlen noch:

    if (N==6) {

        // wie N==5:

        Tag1 = addFaceToCurSurf (2, Invalid1, Volume1);
        Tag2 = addFaceToCurSurf (1, Invalid2, Volume2);

        tesselateBest (N-2, AddVolume + Volume1 + Volume2,
                           ( OneFaceInvalid || Invalid1 || Invalid2 ) );
        PG.undel (Tag2);
        PG.undel (Tag1);
        CurSurf.deleteLast();
        CurSurf.deleteLast();

        // neu:

        Tag1 = addFaceToCurSurf (N-1, Invalid1, Volume1);
        Tag2 = addFaceToCurSurf (0, Invalid2, Volume2);

        tesselateBest (N-2, AddVolume + Volume1 + Volume2,
                           ( OneFaceInvalid || Invalid1 || Invalid2 ) );
        PG.undel (Tag2);
        PG.undel (Tag1);
        CurSurf.deleteLast();
        CurSurf.deleteLast();

        return;
    }

    // N==7: 14 Mglichkeiten fehlen noch:

    if (N==7) {

        Tag1 = addFaceToCurSurf (2, Invalid1, Volume1);
        Tag2 = addFaceToCurSurf (1, Invalid2, Volume2);

        tesselateBest (N-2, AddVolume + Volume1 + Volume2,
                           ( OneFaceInvalid || Invalid1 || Invalid2 ) );
        PG.undel (Tag2);
        PG.undel (Tag1);
        CurSurf.deleteLast();
        CurSurf.deleteLast();

        Tag1 = addFaceToCurSurf (N-1, Invalid1, Volume1);
        Tag2 = addFaceToCurSurf (0, Invalid2, Volume2);

        tesselateBest (N-2, AddVolume + Volume1 + Volume2,
                           ( OneFaceInvalid || Invalid1 || Invalid2 ) );
        PG.undel (Tag2);
        PG.undel (Tag1);
        CurSurf.deleteLast();
        CurSurf.deleteLast();

        // neu:

        Tag1 = addFaceToCurSurf (2, Invalid1, Volume1);
        Tag2 = addFaceToCurSurf (2, Invalid2, Volume2);
        Tag3 = addFaceToCurSurf (1, Invalid3, Volume3);

        tesselateBest (N-3, AddVolume + Volume1 + Volume2 + Volume3,
                           (OneFaceInvalid || Invalid1 || Invalid2 || Invalid3) );
        PG.undel (Tag3);
        PG.undel (Tag2);
        PG.undel (Tag1);
        CurSurf.deleteLast();
        CurSurf.deleteLast();
        CurSurf.deleteLast();

        Tag1 = addFaceToCurSurf (3, Invalid1, Volume1);
        Tag2 = addFaceToCurSurf (2, Invalid2, Volume2);
        Tag3 = addFaceToCurSurf (1, Invalid3, Volume3);

        tesselateBest (N-3, AddVolume + Volume1 + Volume2 + Volume3,
                           (OneFaceInvalid || Invalid1 || Invalid2 || Invalid3) );
        PG.undel (Tag3);
        PG.undel (Tag2);
        PG.undel (Tag1);
        CurSurf.deleteLast();
        CurSurf.deleteLast();
        CurSurf.deleteLast();

        return;
    }

    // N > 7: die harte Tour (sehr langsam, weil viele redundante Tesselierungen):
    // fr jeden Eckpunkt E:
    //     Dreieck E-1, E, E+1 abtrennen  ... Rekursion fr N-1

    for (int I=2; I<N; I++) {                          // Polygon um 1 reduzieren


        Tag1 = addFaceToCurSurf (I, Invalid1, Volume1);

        // Rekursiver Aufruf:

        tesselateBest (N-1, AddVolume + Volume1, OneFaceInvalid || Invalid1);

        // Dreieck von CurSurf abtrennen, Vertex I entlschen:

        CurSurf.deleteLast();

        PG.undel (Tag1);

    } // for
}



BOOL REDUCE_C::hasBestSurfDoubleEdges()
{
    // Hier soll die Entstehung entarteter Sonderflle verhindert werden:
    // in einem 'gesunden' Triset, mit Faces gleichmssiger Orientierung,
    // wird jede Edge von genau zwei Faces geteilt. D.h jede Edge kommt
    // in genau zwei Faces vor und zwar mit UMGEKEHRTER Orientierung.
    // Eine Ausnahme bilden die am Rand liegenden Edges, die nur einmal
    // vorkommen.
    // Das Auftreten von zwei Edges gleicher Orientierung ist also immer
    // zu vermeiden !!
    // Die neuen Faces befinden sich in <BestSurface>, die alten in <OldSurfIdx>.
    // Die alten Faces sind noch nicht gelscht, Diese Edges drfen nicht
    // bercksichtigt werden!

    TRISET_C *TP = Global.TrisetP;

    int NewFaceCnt = BestSurf.nElem();
    int OldFaceCnt = OldSurfIdx.nElem();

    TempIdxList.clear();

    for (int I=0; I<OldFaceCnt; I++)
        TempIdxList.append (OldSurfIdx.Val[I]);

    for (int NF=0; NF<NewFaceCnt; NF++) {

        long *FIP = (long*)&BestSurf.Val[NF].Index;

        if ( TP->existDoubleEdgeExcept (FIP[0], FIP[1], FIP[2], &TempIdxList) )
            return TRUE;

    }

    return FALSE;
}



BOOL REDUCE_C::findBestTesselation()

// Tesseliert das Polygon PgIdx. Fr N <= VERTTHRESH wird die optimale Lsung
// gefunden (erschpfende Suche). Wenn N zu gro ist, wird das Polygon zunchst
// durch einen einfacheren Algorithmus verkleinert.
//
// Die Lsung erfllt folgende Kriterien:
//
//   1.) Der maximale Abstand zwischen den bereits gelschten Punkten und den
//       neuen Flchen ist kleiner gleich MaxError.
//   2.) Die maximale Seitenlnge der neuen Flchen ist kleiner MaxMeshSizeTri
//   3.) Keine der neuen Flchen ist zu schmal (Verhltnis von Seitenlnge zur
//       Hhe)
//   4.) Unter allen Mglichkeiten, die 1.) bis 3.) erfllen, wird diejenige
//       gewhlt, die
//       1. die geringste Volumendifferenz zur vorigen Flche hat, und bei
//          Gleichheit
//       2. die geringste Flche bildet
//
// Falls eine Lsung gefunden wurde, ist sie in BestSurf gespeichert.
// Return value: TRUE, wenn eine Lsung gefunden wurde, sonst oder im Fehlerfall: FALSE;

{
    int    I;

    int    N = PgIdx.nElem();

    if (N<3) {

        // Global.debugInfo ("findBestTesselation: illegal polygon size (%i)", N);
        return FALSE;
    }

    if (N>VERTTHRESH)
        return FALSE;

    Pfloat AddVolume      = 0.0;
    BOOL   OneFaceInvalid = FALSE;

    CurSurf.clear();

    // Polygon nach PG kopieren:

    PG.clear();
    for (I=0; I<N; I++)
        PG.add (PgIdx.Val[I]);

    // erschpfende Suche nach der besten Lsung:

    BestVolume    = INFINITYFLOAT;
    BestMinArea   = INFINITYFLOAT;
    BestSurfValid = FALSE;

    tesselateBest (PG.N, AddVolume, OneFaceInvalid);

    return BestSurfValid;
}



BOOL REDUCE_C::tesselateValid (int N)

{
    FACE_S TempFace;
    long   P1, P2, P3;
    Pfloat Error, AvgError;

    if (N<3) {                         // da ist was faul !
        Global.debugInfo ("tesselateValid: Center Vertex: %li, N=%i", CurVert, N);
        return FALSE;
    }

    if (N==3) {

        // Dieses Dreieck ist das letzte des Patches. Ist es gltig, wird es zum
        // Surface addiert.
        // Die Tesselierung wird verworfen, wenn Restriktionen verletzt werden.
        // Anderenfalls wird diese Tesselierung nach BestSurf kopiert.

        P1 = PG.value (0);
        P2 = PG.value (1);
        P3 = PG.value (2);

        if ( ! isTriValid (P1, P2, P3) )
            return FALSE;

        TempFace.Index[0] = P1;
        TempFace.Index[1] = P2;
        TempFace.Index[2] = P3;
        TempFace.N        = 0;
        TempFace.MaxN     = 0;
        TempFace.HVP      = NULL;

        CurSurf.append (TempFace);

        distributeHiddenVerts (CurVert, &OldSurfIdx, &CurSurf, Error, AvgError);

        if ( Error > MaxErrorTri )
            return FALSE;

        BestSurf.copyFrom (&CurSurf);
        BestSurfValid = TRUE;

        return TRUE;
    }

    // N > 3: Polygon ist komplexer als ein Dreieck:
    // einen zufllig gewhlten Eckpunkt abtrennen, und Rekursion

    int I = random (N);

    P1 = PG.prec  (I);
    P2 = PG.value (I);
    P3 = PG.succ  (I);

    // ist das abgetrennte Dreieck zulssig ?

    if ( ! isTriValid (P1, P2, P3) )
        return FALSE;

    // Vertex I vom Polygon abtrennen, entstehendes Dreieck wird an CurSurf
    // angehngt

    TempFace.Index[0] = P1;
    TempFace.Index[1] = P2;
    TempFace.Index[2] = P3;
    TempFace.N        = 0;
    TempFace.MaxN     = 0;
    TempFace.HVP      = NULL;

    CurSurf.append (TempFace);

    PG.del (I);

    // Rekursiver Aufruf:

    return tesselateValid (N-1);
}

BOOL REDUCE_C::findValidTesselation()

// Generiert eine Tesselierung des Polygons in PgIdx, und prft, ob die Restriktionen
// eingehalten werden.
// Diese Prozedur arbeitet wesentlich simpler als findBestTesselation(), und wird beim
// Annealing genutzt.
// Falls eine Lsung gefunden wurde, ist sie in BestSurf gespeichert.
// Return value: TRUE, wenn eine Lsung gefunden wurde, sonst oder im Fehlerfall: FALSE;

{
    int    I;

    int    N = PgIdx.nElem();

    if (N<3)
        return FALSE;

    if (N>VERTTHRESH)
        return FALSE;

    CurSurf.clear();

    // Polygon nach PG kopieren:

    PG.clear();
    for (I=0; I<N; I++)
        PG.add (PgIdx.Val[I]);

    // eine gltige Tesselierung generieren:

    BestSurfValid = FALSE;

    tesselateValid (PG.N);

    return BestSurfValid;
}

/*------------------------------------------------------------------------+
|                                                                         |
| Fr Simulated Annealing                                                 |
|                                                                         |
+------------------------------------------------------------------------*/

BOOL REDUCE_C::findUndelTesselation (long VertIdx)

// Generiert Tesselierung, die entsteht, wenn VertIdx entlscht wird.
// Wenn Randbedingungen (Maschengre, Proportion, Fehler) verletzt werden,
// wird FALSE zurckgegeben.
// Sonst ist das alte Patch in OldSurfIdx und das neue in BestSurf gespeichert.
// Danach kann deltaEnergy() und ggf. replacePatch(..) aufgerufen werden.
// Auch wenn ein Entlschen erlaubt ist, kann nicht gesagt werden, ob es wirklich
// durchgefhrt wird (hngt vom Metropoliskriterium ab). Wenn replacePatch(..)
// aufgerufen wird mu auerdem der Vaterfaceindex auf NOINDEX gesetzt werden.
{
    int          I, J;
    int          NewN;
    long         TempL;
    FACE_S       TempF;
    FACE_S       *TempFP;
    long         *FIP, *FIP2;
    Pfloat       Error, AvgError;

    // Initialisierungen:

    TRISET_C *TP      = Global.TrisetP;
    VERTEX_T *VertP   = TP->getVertP (VertIdx);

    if (VertP->FI == 0) {

        Global.debugInfo ("findUndelTesselation (%li): Vertex has no father", VertIdx);
        return FALSE;
    }

    FACE_S   *FFP      = TP->getFaceP (VertP->FI);    // Vater Face von <VertIdx>
    long     *FFIP     = FFP->Index;                  // Eckpunkte von FFP

    OldSurfIdx.clear();
    BestSurf.clear();
    BestSurfValid = FALSE;

    // Indizes der alten Faces ermitteln. Dazu gehren das Vaterface und zwei Nachbarn
    // Nachbarfaces teilen mit dem Vater Face eine Edge, allerdings in umgekehrter
    // Orientierung:

    OldSurfIdx.append (VertP->FI);
    TempL = TP->findOrientatedEdge (FFIP[0], FFIP[2], 0L);
    if (TempL!=NOINDEX) OldSurfIdx.append (TempL);
    TempL = TP->findOrientatedEdge (FFIP[1], FFIP[0], 0L);
    if (TempL!=NOINDEX) OldSurfIdx.append (TempL);

    if (OldSurfIdx.nElem() != 3) {

//      Global.debugMsg ("findUndelTesselation: Vertex %li hat nur %i Nachbarfaces\n",
//                       VertIdx, OldSurfIdx.nElem() );
        return FALSE;
    }

    // Neue Tesselierung generieren (Randbedingungen mssen eingehalten werden)
    // VertIdx wird mit allen Edges der Nachbarfaces, die NICHT an das Vaterface
    // grenzen, zu einem Face verbunden.

    TempF.N = TempF.MaxN = 0;          // Temporr fr Appendfunktion
    TempF.HVP = 0;
    TempF.Index[2] = VertIdx;

    // Das 1. Old Face hat die Edge 0,2 mit dem Vater Face gemein.

    TempFP  = TP->getFaceP (OldSurfIdx.Val[1]);   // 1.Old Face
    FIP = TempFP->Index;                          // Eckpunkte von 1. Old Face

    if ( (FIP[0]!=FFIP[0]) && (FIP[0]!=FFIP[2]) ) {

        TempF.Index[0]=FIP[2]; TempF.Index[1]=FIP[0];
        BestSurf.append (TempF);
        TempF.Index[0]=FIP[0]; TempF.Index[1]=FIP[1];
        BestSurf.append (TempF);
    }
    if ( (FIP[1]!=FFIP[0]) && (FIP[1]!=FFIP[2]) ) {

        TempF.Index[0]=FIP[0]; TempF.Index[1]=FIP[1];
        BestSurf.append (TempF);
        TempF.Index[0]=FIP[1]; TempF.Index[1]=FIP[2];
        BestSurf.append (TempF);
    }
    if ( (FIP[2]!=FFIP[0]) && (FIP[2]!=FFIP[2]) ) {

        TempF.Index[0]=FIP[1]; TempF.Index[1]=FIP[2];
        BestSurf.append (TempF);
        TempF.Index[0]=FIP[2]; TempF.Index[1]=FIP[0];
        BestSurf.append (TempF);
    }

    // Das 2. Old Face hat die Edge 1,0 mit dem Vater Face gemein.

    TempFP  = TP->getFaceP (OldSurfIdx.Val[2]);
    FIP = TempFP->Index;

    if ( (FIP[0]!=FFIP[1]) && (FIP[0]!=FFIP[0]) ) {

        TempF.Index[0]=FIP[2]; TempF.Index[1]=FIP[0];
        BestSurf.append (TempF);
        TempF.Index[0]=FIP[0]; TempF.Index[1]=FIP[1];
        BestSurf.append (TempF);
    }
    if ( (FIP[1]!=FFIP[1]) && (FIP[1]!=FFIP[0]) ) {

        TempF.Index[0]=FIP[0]; TempF.Index[1]=FIP[1];
        BestSurf.append (TempF);
        TempF.Index[0]=FIP[1]; TempF.Index[1]=FIP[2];
        BestSurf.append (TempF);
    }
    if ( (FIP[2]!=FFIP[1]) && (FIP[2]!=FFIP[0]) ) {

        TempF.Index[0]=FIP[1]; TempF.Index[1]=FIP[2];
        BestSurf.append (TempF);
        TempF.Index[0]=FIP[2]; TempF.Index[1]=FIP[0];
        BestSurf.append (TempF);
    }

    // Es fehlt noch ein Face zwischen VertIdx und der nicht benutzten Edge
    // des Vater Face:

    TempF.Index[0]=FFIP[1]; TempF.Index[1]=FFIP[2];
    BestSurf.append (TempF);

    NewN = BestSurf.nElem();

    if (NewN != 5) {

        Global.debugMsg ("undelBig: %i neue Faces erzeugt\n", BestSurf.nElem() );
        return FALSE;
    }

    // Abbruch, wenn doppelte Faces entstanden sind (Sonderfall):

    for (I=0; I<NewN-1; I++) {

        FIP = BestSurf.Val[I].Index;

        for (J=I+1; J<NewN; J++) {

            FIP2 = BestSurf.Val[J].Index;

            if ( ( (FIP[0]==FIP2[0]) || (FIP[0]==FIP2[1]) || (FIP[0]==FIP2[2]) ) &&
                 ( (FIP[1]==FIP2[0]) || (FIP[1]==FIP2[1]) || (FIP[1]==FIP2[2]) ) &&
                 ( (FIP[2]==FIP2[0]) || (FIP[2]==FIP2[1]) || (FIP[2]==FIP2[2]) ) )
                return FALSE;
        }
    }

    // Abbruch, wenn Randbedingungen verletzt wrden:

    for (I=0; I<NewN; I++) {

        FIP = BestSurf.Val[I].Index;

        if ( ! isTriValid (FIP[0], FIP[1], FIP[2]) )
            return FALSE;
    }

    // Hidden Vertices neu zuordnen (VertIdx negativ angeben -> wird ignoriert)

    distributeHiddenVerts (-VertIdx, &OldSurfIdx, &BestSurf, Error, AvgError);

    if (Error > Global.OptionsP->MaxErrorTri) {
        BestSurf.clear();
        return FALSE;
    }

    BestSurfValid = TRUE;
    return TRUE;
}



float REDUCE_C::energy()
{
    double    VertexWeight     = 0.0;
    double    ErrorWeight      = 0.0;
    double    ProportionWeight = 0.0;
    double    AreaWeight       = 0.0;

    TRISET_C  *TP          = Global.TrisetP;
    OPTIONS_C *OP          = Global.OptionsP;

    float     PropFactor   = OP->ProportionFactor;
    float     ErrorFactor  = OP->ErrorFactor;
    float     VertexFactor = OP->VertexFactor;
    float     AreaFactor   = OP->AreaFactor;
    BOOL      ConsiderVerts = (VertexFactor != 0.0);
    BOOL      ConsiderError = (ErrorFactor != 0.0);
    BOOL      ConsiderProp  = (PropFactor != 0.0);
    BOOL      ConsiderArea  = (AreaFactor != 0.0);

    long      NF           = TP->FaceCnt;
    float     E            = 0.0;


    for (long F=0; F<NF; F++) {

        FACE_S  *FP  = TP->getFaceP (F);
        long    *FIP = FP->Index;
        int     NHV  = FP->N;
        Ppoint3 V1   = *TP->getVertPtP (FIP[0]);
        Ppoint3 V2   = *TP->getVertPtP (FIP[1]);
        Ppoint3 V3   = *TP->getVertPtP (FIP[2]);

        if (!_isFaceDeleted (FIP)) {

           if (ConsiderError) {

               for (int I=0; I<NHV; I++) {

                   double D = distancePtTri3 (*TP->getVertPtP(FP->HVP[I]), V1, V2, V3);
                   ErrorWeight += (D * D);
               }
           }

           if (ConsiderProp) {

               ProportionWeight += proportion (&V1, &V2, &V3);
           }

           if (ConsiderArea) {

               AreaWeight += areaTri3 (V1, V2, V3);
           }
        }
    }

    if (ConsiderVerts)
        VertexWeight = (float)TP->VisibleVertexCnt;

    E = VertexWeight*VertexFactor   + ErrorWeight*ErrorFactor +
        ProportionWeight*PropFactor + AreaWeight*AreaFactor;

    Global.debugMsg ("VW:%f, EW:%f, PW:%f, AW:%f\n",
                     VertexWeight*VertexFactor,
                     ErrorWeight*ErrorFactor,
                     ProportionWeight*PropFactor,
                     AreaWeight*AreaFactor);

    return E;
}



float REDUCE_C::deltaEnergy()

// Berechnet wird die Energiedifferenz zwischen einer alten und einer neuen
// Tesselierungsvariante. Die Patches mssen in OldSurfIdx und BestSurf abgelegt
// sein.
// Die Energie setzt sich aus 4 Anteilen zusammen (vgl. Doku)
{
    int       I, F;
    float     E;
    FACE_S    *FP;
    long      *FIP;
    int       NHV;
    Ppoint3   *V1P, *V2P, *V3P;

    double    VertexWeight     = 0.0;
    double    ErrorWeight      = 0.0;
    double    ProportionWeight = 0.0;
    double    AreaWeight       = 0.0;

    TRISET_C  *TP          = Global.TrisetP;
    OPTIONS_C *OP          = Global.OptionsP;

    double    PropFactor   = OP->ProportionFactor;
    double    ErrorFactor  = OP->ErrorFactor;
    double    VertexFactor = OP->VertexFactor;
    double    AreaFactor   = OP->AreaFactor;
    BOOL      ConsiderVerts = (VertexFactor != 0.0);
    BOOL      ConsiderError = (ErrorFactor != 0.0);
    BOOL      ConsiderProp  = (PropFactor != 0.0);
    BOOL      ConsiderArea  = (AreaFactor != 0.0);

    double    OldError2    = 0.0;      // Summe der Fehlerquardate
    double    NewError2    = 0.0;
    double    OldProp      = 0.0;      // Summe der Proportionswerte
    double    NewProp      = 0.0;
    double    OldArea      = 0.0;      // Flche der Faces
    double    NewArea      = 0.0;

    // Der Energieanteil der von der Anzahl der Vertizes abhngt kann sich nur um
    // +- 1 ndern:

    if (ConsiderVerts)
        VertexWeight = (CurVertVisible ? -1.0 : 1.0 );

    // Energieanteile fr das alte Patch ermitteln:

    for (F=0; F<OldSurfIdx.nElem(); F++) {

        FP   = TP->getFaceP (OldSurfIdx.Val[F]);
        FIP  = FP->Index;
        NHV  = FP->N;
        V1P  = TP->getVertPtP (FIP[0]);
        V2P  = TP->getVertPtP (FIP[1]);
        V3P  = TP->getVertPtP (FIP[2]);

        if (ConsiderError) {

            for (I=0; I<NHV; I++) {

                double D = distancePtTri3 (*TP->getVertPtP(FP->HVP[I]), *V1P, *V2P, *V3P);
                OldError2 += (D * D);
            }
        }
        if (ConsiderProp)
            OldProp += proportion (V1P, V2P, V3P);
        if (ConsiderArea)
            OldArea += areaTri3 (*V1P, *V2P, *V3P);

    }

    // Energieanteile fr das neue Patch ermitteln:

    for (F=0; F<BestSurf.nElem(); F++) {

        FP   = &BestSurf.Val[F];
        FIP  = FP->Index;
        NHV  = FP->N;
        V1P  = TP->getVertPtP (FIP[0]);
        V2P  = TP->getVertPtP (FIP[1]);
        V3P  = TP->getVertPtP (FIP[2]);

        if (ConsiderError) {

            for (I=0; I<NHV; I++) {

                double D = distancePtTri3 (*TP->getVertPtP(FP->HVP[I]), *V1P, *V2P, *V3P);
                NewError2 += (D * D);
            }
        }
        if (ConsiderProp)
            NewProp += proportion (V1P, V2P, V3P);
        if (ConsiderArea)
            NewArea += areaTri3 (*V1P, *V2P, *V3P);
    }

    ErrorWeight      = (NewError2 - OldError2);
    ProportionWeight = (NewProp - OldProp);
    AreaWeight       = (NewArea - OldArea);

    E = (float) ( VertexFactor * VertexWeight +
                  ErrorFactor  * ErrorWeight +
                  PropFactor   * ProportionWeight +
                  AreaFactor   * AreaWeight
                );

    return E;
}



BOOL REDUCE_C::metropolis (float DeltaEnergy, float Temp)
{
    double Exponent;
    double Pundel;

    if (DeltaEnergy < 0.0) return TRUE;

    if (Temp > 0.0001) {

        Exponent = - DeltaEnergy / Temp;

        if (Exponent>FMINEXP && Exponent<FMAXEXP)
            Pundel = exp (Exponent);
        else
            Pundel = 0.0;
        }
    else
        return FALSE;

    double R = random(10000);
    double P = 10000.0*Pundel;

    return ( R < P );
}

/*------------------------------------------------------------------------+
|                                                                         |
| Reduktion eines Trisets auf direktem Weg (Punkt auswhlen, lschen      |
| wenn mglich, ...)                                                      |
|                                                                         |
+------------------------------------------------------------------------*/

void REDUCE_C::reduceDirect()
{
    Global.StatWinP->status ("reducing triset (direct) ...");

    TRISET_C        *TP        = Global.TrisetP;
    OPTIONS_C       *OP        = Global.OptionsP;
    STATUSWINDOW_C  *SP        = Global.StatWinP;
    HISTORYWINDOW_C *HistWinP  = SP->HistWinP;

    HistWinP->setRangeY (0.0, 100.0);

    BOOL        DerJuengsteTag    = FALSE;
    long        InitialVertCnt    = TP->VertexCnt;

    float       DeltaT            = OP->HitRateInterval;   // Zeitraum fr HitRate

    long        LastSuccess       = 0L;          // Wert von Processed beim
                                                 // letzten  del., undel.
    long        LastFaceCnt       = TP->FaceCnt; // Faces beim letzten Kompaktieren
    long        StartTime         = 0L;          // Startzeit in Ticks
    long        DeletedVerts      = 0L;          // Anzahl der gelschten Punkte
    int         AutoSaveCnt       = 0;
    float       AutoSaveStartTime = OP->AutoSaveStartTime*3600;
    float       DelRate           = 0.0;
    long        DelCnt            = 0L;          // Erfolgreiche Lschungen
    long        DelRejectCnt      = 0L;          // Gescheiterte Lschversuche

    // CurVert = CurVert +- 0..Walkrange

    int         WalkRange         = (int) (TP->VertexCnt / 8);

    Processed      = 0L;                         // Anzahl der betrachteten Punkte
    CurVert        = 0L;

    SavedEnergy    = (float)InitialVertCnt;
    StartTime      = Global.getTimeMark();

    // +---------------------------------------------------------+
    // | Reduktionsschleife bis Stopkriterium erfllt            |
    // +---------------------------------------------------------+

    do {

        // +---------------------------------------------------------+
        // | Statusanzeige, Logdateien, Statusvariablen updaten      |
        // +---------------------------------------------------------+

        if ( (Processed % OP->UpdateStatusInterval) == 0 ) {

            SP->status ("reducing triset (direct)... %li,  "
                        "del: %li(reject:%li), E:%.3f",
                        Processed,
                        DelCnt, DelRejectCnt, TP->VisibleVertexCnt
                       );
            SP->Percentage     = 100 * DelRate;
            SP->ReducProcessed = Processed;
            SP->Temperature    = 0;
            SP->updateDlg();
        }
/*      else
            Global.yield();                // den Tasten eine Chance ...
*/
        // Statistik in Logdatei schreiben:

        if (OP->HistoryGraph) {

            if (Processed % OP->UpdateHistoryInterval == 0) {

                HistWinP->addHistory (100.0 * DeletedVerts / InitialVertCnt,
                                      100.0 * DelRate,
                                      0, 0, 0
                                     );
            }
        }

        if (OP->LogHistory) {

            if (Processed % OP->LogHistoryInterval == 0) {

                Global.statsOut ("%5.2f, %5.2f, %6.2f, %5.2\n",
                                 100.0 * DeletedVerts / InitialVertCnt,
                                 100.0 * DelRate,
                                 0, 0
                                );
            }
        }

        Processed++;

        if (DelRate<0.001)
            DelRate   = 0.0;
        else
            DelRate = (DeltaT-1.0) * DelRate / DeltaT;

        // +---------------------------------------------------------+
        // | Punkt auswhlen                                         |
        // +---------------------------------------------------------+

        CurVert  = CurVert + random (2*WalkRange+1) - WalkRange / 2;

        CurVert  = labs (CurVert % TP->VertexCnt);

        CurVertP = TP->getVertP (CurVert);

        CurVertVisible = (CurVertP->FI==0L);

        // +---------------------------------------------------------+
        // | Wenn der Punkt noch nicht gelscht war:                 |
        // |   versuchen zu lschen                                  |
        // +---------------------------------------------------------+

        if ( CurVertVisible ) {

            // Nachbarn des Punktes suchen                             |

            CurVertValid = TP->getNeighbours (CurVert, &PgIdx, &OldSurfIdx, PgClosed);

            // Randpunkte fhren zu einem Polygonzug, der nur ber sie selbst
            // geschlossen ist (PgClosed=FALSE). Sie drfen nur gelscht werden,
            // wenn die neue Randlinie keine groe Entfernung zu ihnen hat, und
            // die Option KeepEdgePoints nicht gesetzt ist

            if ( OP->KeepEdgePoints && ! PgClosed )
                CurVertValid = FALSE;

            if (CurVertValid && ! PgClosed) {

                 if ( distancePtLn3 ( *TP->getVertPtP (CurVert),
                                     *TP->getVertPtP ( PgIdx.Val[0] ),
                                     *TP->getVertPtP ( PgIdx.Val[PgIdx.nElem()-1] )
                                   ) > MaxErrorTri )
                    CurVertValid = FALSE;
            }

            if (CurVertValid) {

                // lschen, wenn eine der mglichen neuen Tesselierungen die
                // Randbedingungen einhlt:

                if ( findBestTesselation() ) {

                    TP->replacePatch (&OldSurfIdx, &BestSurf);

                    DeletedVerts++;
                    CurEnergy = (float)TP->VisibleVertexCnt;

                    SP->ReducVerts2++;
                    SP->ReducFaces += 2;

                    LastSuccess = Processed;
                    SP->ReducLastSuccess = LastSuccess;

                    DelRate += 1.0 / DeltaT;             // Erfolgsbilanz schnen

                    // +---------------------------------------------------------+
                    // | ggf. Backup vom Triset machen                           |
                    // +---------------------------------------------------------+

                    if ( (Global.elapsed(StartTime) > AutoSaveStartTime) &&
                         (100.0*(SavedEnergy-TP->VisibleVertexCnt)/TP->VisibleVertexCnt
                            >= OP->AutoSavePercentage ) ) {

                        AutoSaveCnt++;

                        Global.debugMsg ("Autosave #%i: elapsed:%s\n",
                                         AutoSaveCnt,
                                         Global.elapsedStrHMS(Global.elapsed(StartTime))
                                        );
                        Global.logMsg   ("Autosave #%i: elapsed:%s\n",
                                         AutoSaveCnt,
                                         Global.elapsedStrHMS(Global.elapsed(StartTime))
                                        );
                        TP->backup();
                        SavedEnergy = (float)TP->VisibleVertexCnt;
                    }

                    // +---------------------------------------------------------+
                    // | ggf. Triset Kompaktieren                                |
                    // +---------------------------------------------------------+

                    if ( (TP->FaceCnt-LastFaceCnt) > OP->CompactThresh ) {

                        TP->compactFaces();

                        LastFaceCnt = TP->FaceCnt;
                        WalkRange   = (int) (TP->VertexCnt / 8)+1;
                    }
                    DelCnt++;

                    } // findBestTesselation()
                else
                    DelRejectCnt++;

            } // vertex valid

        } // CurVertVisible

        // +---------------------------------------------------------+
        // | Abbruchkriterium                                        |
        // +---------------------------------------------------------+

        if ( ( (Processed - LastSuccess) > OP->StopThresh) ||
             ( Global.CancelRequest ) ) {

            DerJuengsteTag = TRUE;
        }

    } while ( ! DerJuengsteTag );

    // +---------------------------------------------------------+
    // | Ende der Reduktion                                      |
    // +---------------------------------------------------------+

    TP->compactFaces();
    TP->integrityCheck();

    SP->ReducProcessed = Processed;
    SP->ReducLastSuccess = LastSuccess;
    SP->updateDlg();

    SP->status ("");
}


/*------------------------------------------------------------------------+
|                                                                         |
| Reduktion eines Trisets durch Simulated Annealing                       |
|                                                                         |
+------------------------------------------------------------------------*/

void REDUCE_C::reduceAnneal()
{
    Global.StatWinP->status ("reducing triset (annealing) ...");

    TRISET_C        *TP        = Global.TrisetP;
    OPTIONS_C       *OP        = Global.OptionsP;
    STATUSWINDOW_C  *SP        = Global.StatWinP;
    HISTORYWINDOW_C *HistWinP  = SP->HistWinP;

    HistWinP->setRangeY (0.0, 100.0);

    BOOL        DerJuengsteTag    = FALSE;
    long        InitialVertCnt    = TP->VertexCnt;

    float       DeltaT            = OP->HitRateInterval;   // Zeitraum fr HitRate
    long        MinWaitCnt        = 2 * DeltaT;  // min. Zeit, in der T konst. bleibt
    float       InitTemperature;                 // Temperatur am Anfang
    float       FinalTemperature;                // Temperatur am Ende
    float       EpsilonEqui       = 0.01;        // epsilon fr Systemgleichgewicht
    float       AnnealFactor      = OP->AnnealFactor / 100.0;
    long        TimeLimitTicks;
    long        LastCoolTicks;
    BOOL        AutoSchedule      = (OP->TimeLimitHours>0.0);

    float       Temperature;
    float       InitialEnergy     = energy();    // Zur History-Skalierung
    float       DeltaEnergy;
    int         AutoSaveCnt       = 0;
    float       AutoSaveStartTime = OP->AutoSaveStartTime*3600.0;
    long        WaitCnt           = 0;           // Processed seit letzter Abkhlung
    long        LastSuccess       = 0L;          // Wert von Processed beim
                                                 // letzten  del., undel.
    long        LastFaceCnt       = TP->FaceCnt; // Faces beim letzten Kompaktieren
    long        StartTime;                       // Startzeit in Ticks
    long        DeletedVerts      = 0L;          // Anzahl der gelschten Punkte
    float       DelRate           = 0.0;
    long        DelCnt            = 0L;          // Erfolgreiche Lschungen
    long        DelRejectCnt      = 0L;          // Gescheiterte Lschversuche
    float       UndelRate         = 0.0;
    long        UndelCnt          = 0L;          // Erfolgreiche Entlschungen
    long        UndelRejectCnt    = 0L;          // Gescheiterte Entlschversuche
    long        MaxDeleted        = 0L;          // max. erreichte Anzahl von
                                                 // gelschten Punkten


    // Bestimmung des Annealing-Schedule:

    StartTime        = Global.getTimeMark();

    InitTemperature  = -InitialEnergy / ( InitialVertCnt * log(0.9) );
    FinalTemperature = 0.1;
    Temperature      = InitTemperature;
    Global.debugMsg ("E: %f, N: %li -> Tstart = %f (Tend=%f)\n",
                     InitialEnergy, InitialVertCnt, InitTemperature, FinalTemperature);
    Global.logMsg   ("E: %f, N: %li -> Tstart = %f (Tend=%f)\n",
                     InitialEnergy, InitialVertCnt, InitTemperature, FinalTemperature);
    LastCoolTicks    = StartTime;
    TimeLimitTicks   = (long)(OP->TimeLimitHours * 1000 * 3600) + StartTime;

    int WalkRange    = (int) (TP->VertexCnt / 8);

    Processed        = 0L;                         // Anzahl der betrachteten Punkte
    CurVert          = 0L;

    CurEnergy        = InitialEnergy;
    SavedEnergy      = MAXFLOAT;

    if (InitialEnergy<=1.0) InitialEnergy = 10;

    // +---------------------------------------------------------+
    // | Reduktionsschleife bis Stopkriterium erfllt            |
    // +---------------------------------------------------------+

    do {

/*********************
    
        for (long i=0; i<TP->VertexCnt; i++) {

            if (TP->getVertP(i)->FI==0L) continue;
            FACE_S *FP = TP->getFaceP (TP->getVertP(i)->FI);
            long   *FIP= FP->Index;
            if (distancePtTri3Idx (i, FIP[0], FIP[1], FIP[2])>0.000001) {
                Global.beep();
                Global.debugMsg ("dist V->Vater = %f\n",distancePtTri3Idx (i, FIP[0], FIP[1], FIP[2]));
                TP->dumpVertexInfo (i);
                TP->dumpFaceInfo (TP->getVertP(i)->FI);
                TP->markFace (TP->getVertP(i)->FI, 255, 255, 255);
                return;
            }
        }
**********************/

        // +---------------------------------------------------------+
        // | Annealingparameter setzen                               |
        // +---------------------------------------------------------+

        if ( WaitCnt > MinWaitCnt ) {

            // System im Gleichgewicht, dann Temperatur absenken:

            if ( fabs ( UndelRate - DelRate ) < EpsilonEqui ) {

                if ( AutoSchedule && (Temperature<InitTemperature) ) {

                    long LastEquiTicks = Global.getTimeMark() - LastCoolTicks;
                    long TicksLeft     = TimeLimitTicks - Global.getTimeMark();
                    int  N             = int (TicksLeft / LastEquiTicks);

                    if ( (N>0) && (Temperature>(FinalTemperature+0.01)) ) {

                        float C = pow ( (FinalTemperature/Temperature), 1.0/N);
                        Temperature *= C;
//                      Global.debugMsg ("Cooling down: N=%i, C=%f\n", N, C);
                    } else {
                        Temperature = 0.0;
                    }

                } else {

                    Temperature *= AnnealFactor;
                }
                WaitCnt = 0;
                LastCoolTicks = Global.getTimeMark();
            }
        }

        // +---------------------------------------------------------+
        // | Statusanzeige, Logdateien, Statusvariablen updaten      |
        // +---------------------------------------------------------+

        if ( (Processed % OP->UpdateStatusInterval) == 0 ) {

            // Abschtzung fr P(undel)
            float Cold = 0.0;
            if (Temperature > 0.0001) {
                float Exponent = - (CurEnergy / TP->VisibleVertexCnt) / Temperature;
                if (Exponent>FMINEXP && Exponent<FMAXEXP) Cold = exp (Exponent);
            }

            SP->status ("reducing triset (annealing) ... %li,  "
                        "undel:%li (reject:%li, rate: %.1f) "
                        "del: %li(reject:%li), E=%f, "
                        "Cold=%.3f",
                        Processed,
                        UndelCnt, UndelRejectCnt, UndelRate * 100.0,
                        DelCnt, DelRejectCnt, CurEnergy,
                        Cold
                       );

            SP->Percentage     = 100.0 * DelRate;
            SP->ReducProcessed = Processed;
            SP->Temperature    = Temperature;
            SP->updateDlg();
            }
/*      else
            Global.yield();                // den Tasten eine Chance ...
*/
        // Statistik in Logdatei schreiben:

        if (OP->HistoryGraph) {

            if (Processed % OP->UpdateHistoryInterval == 0) {

                HistWinP->addHistory (100.0 * DeletedVerts / InitialVertCnt,
                                      100.0 * DelRate,
                                      100.0 * Temperature / InitTemperature,
                                      100.0 * CurEnergy / InitialEnergy,
                                      0
                                     );
            }
        }

        if (OP->LogHistory) {

            if (Processed % OP->LogHistoryInterval == 0) {

                Global.statsOut ("%5.2f, %5.2f, %6.2f, %6.2f\n",
                                 100.0 * DeletedVerts / InitialVertCnt,
                                 100.0 * DelRate,
                                 100.0 * Temperature / InitTemperature,
                                 100.0 * CurEnergy / InitialEnergy
                                );
            }
        }

        if ((Processed % 10000L) == 0) {

            float NewE = energy();
            Global.debugMsg ("Incremental E: %f, actual: %f\n", CurEnergy, NewE);
            CurEnergy = NewE;             // Rundungsfehler ausgleichen
        }

        Processed++;
        WaitCnt++;

        if (DelRate<0.001)
            DelRate   = 0.0;
        else
            DelRate = (DeltaT-1.0) * DelRate / DeltaT;

        if (UndelRate<0.001)
            UndelRate = 0.0;
        else
            UndelRate = (DeltaT-1.0) * UndelRate / DeltaT;

        // if (Processed == 1071L) Global.info ("BINGO: Processed");
   
        // +---------------------------------------------------------+
        // | ggf. Zustand abspeichern                                |
        // +---------------------------------------------------------+

        if ( ( Global.elapsed (StartTime) > AutoSaveStartTime ) &&
             ( (100.0*(SavedEnergy-CurEnergy) / CurEnergy) >= OP->AutoSavePercentage ) ) {

            AutoSaveCnt++;

            Global.debugMsg ("Autosave #%i: elapsed:%s, E:%.3f\n",
                             AutoSaveCnt,
                             Global.elapsedStrHMS(Global.elapsed(StartTime)),
                             CurEnergy
                            );
            Global.logMsg   ("Autosave #%i: elapsed:%s, E:%.3f\n",
                             AutoSaveCnt,
                             Global.elapsedStrHMS(Global.elapsed(StartTime)),
                             CurEnergy
                            );
            TP->backup();
            SavedEnergy = CurEnergy;
        }

        // +---------------------------------------------------------+
        // | Punkt auswhlen                                         |
        // +---------------------------------------------------------+

        CurVert  = CurVert + random (2*WalkRange+1) - WalkRange / 2;

        CurVert  = labs (CurVert % TP->VertexCnt);

        CurVertP = TP->getVertP (CurVert);

        CurVertVisible = (CurVertP->FI==0L);


        // +---------------------------------------------------------+
        // | Wenn der gewhlte Punkt sichtbar ist, wird geprft, ob  |
        // | eine Lschung mglich ist. Wenn ja, wird die beste      |
        // | Tesselierung generiert.                                 |
        // | Wenn der Punkt schon gelscht war, ist eine Entlschung |
        // | immer erlaubt.                                          |
        // | DeltaEnergy gibt die Energie der Zustandsnderung an.   |
        // +---------------------------------------------------------+

        if (CurVertVisible) {          // auf mgliche Lschung untersuchen

            CurVertValid = TP->getNeighbours (CurVert, &PgIdx, &OldSurfIdx, PgClosed);

            // Randpunkte fhren zu einem Polygonzug, der nur ber sie selbst
            // geschlossen ist (PgClosed=FALSE). Sie drfen nur gelscht werden,
            // wenn die neue Randlinie keine groe Entfernung zu ihnen hat, und
            // die Option KeepEdgePoints nicht gesetzt ist

            if ( OP->KeepEdgePoints && ! PgClosed )
                CurVertValid = FALSE;

            if (CurVertValid && ! PgClosed)
                if ( distancePtLn3 ( *TP->getVertPtP (CurVert),
                                     *TP->getVertPtP ( PgIdx.Val[0] ),
                                     *TP->getVertPtP ( PgIdx.Val[PgIdx.nElem()-1] )
                                   ) > MaxErrorTri )
                    CurVertValid = FALSE;

            if (CurVertValid) {

                CurVertValid = findBestTesselation();
                // CurVertValid = findValidTesselation();
            }

            if (CurVertValid)
                if (Global.OptionsP->AvoidDoubles)
                    CurVertValid = !hasBestSurfDoubleEdges();

            if (!CurVertValid)
                DelRejectCnt++;

        } else {                       // auf mgliche Entlschung untersuchen

            if ( findUndelTesselation (CurVert) ) {

                CurVertValid = TRUE;
            } else {
                UndelRejectCnt++;
                CurVertValid = FALSE;
            }
        }

        if ( CurVertValid ) {

            DeltaEnergy = deltaEnergy();
        }

        // +---------------------------------------------------------+
        // | Wenn es das Metropoliskriterium zult:                 |
        // |   gelschten Punkt: versuchen zu entlschen,            |
        // | oder                                                    |
        // |   sichtbaren Punkt: lschen                             |
        // +---------------------------------------------------------+

        if ( CurVertValid && metropolis (DeltaEnergy, Temperature) ) {

            if ( ! CurVertVisible ) {  // Punkt war schon gelscht

                // Alte Faces durch neue ersetzen

                TP->replacePatch (&OldSurfIdx, &BestSurf);

             /*********
                FACE_S *FP = TP->getFaceP (CurVertP->FI);
                long   *FIP= FP->Index;
                if (distancePtTri3Idx (CurVert, FIP[0], FIP[1], FIP[2])>0.000001) {
                    Global.beep();
                    Global.debugMsg ("undel: dist V->Vater = %f\n",distancePtTri3Idx (CurVert, FIP[0], FIP[1], FIP[2]));
                    TP->dumpVertexInfo (CurVert);
                    TP->dumpFaceInfo (CurVertP->FI);
                    TP->markFace (CurVertP->FI, 255, 255, 255);
                    return;
                }
             *******/

                // Vertex wird sichtbar, hat also kein Vater Face mehr:

                CurVertP->FI = 0L;

                SP->ReducVerts2--;
                SP->ReducFaces -= 2;

                CurEnergy += DeltaEnergy;

                UndelCnt++;
                DeletedVerts--;
                LastSuccess = Processed;
                UndelRate += 1.0 / DeltaT;

/*              if ( ! TP->integrityCheck() ) {
                    Global.info ("BINGO: double edges AFTER undel");
                    TP->getVertPtCoP(CurVert) ->colour.x = 1.0;
                    TP->getVertPtCoP(CurVert) ->colour.y = 0.0;
                    TP->getVertPtCoP(CurVert) ->colour.z = 0.0;
                    Global.CancelRequest = TRUE;
                }
*/
            } else {                     // CurVert ist nicht gelscht

                // lschen

                TP->replacePatch (&OldSurfIdx, &BestSurf);

             /*********
                FACE_S *FP = TP->getFaceP (CurVertP->FI);
                long   *FIP= FP->Index;
                if (distancePtTri3Idx (CurVert, FIP[0], FIP[1], FIP[2])>0.000001) {
                    Global.beep();
                    Global.debugMsg ("del: dist V->Vater = %f\n",distancePtTri3Idx (CurVert, FIP[0], FIP[1], FIP[2]));
                    TP->dumpVertexInfo (CurVert);
                    TP->dumpFaceInfo (CurVertP->FI);
                    TP->markFace (CurVertP->FI, 255, 255, 255);
                    return;
                }
             *******/

                CurEnergy += DeltaEnergy;
                DeletedVerts++;
                if (DeletedVerts>MaxDeleted) MaxDeleted = DeletedVerts;

                SP->ReducVerts2++;
                SP->ReducFaces += 2;

                LastSuccess = Processed;
                SP->ReducLastSuccess = LastSuccess;

                DelRate += 1.0 / DeltaT;             // Erfolgsbilanz schnen

                // ggf. FaceArray Komprimieren:

                if ( (TP->FaceCnt-LastFaceCnt) > OP->CompactThresh ) {

                    TP->compactFaces();

                    LastFaceCnt = TP->FaceCnt;
                    WalkRange   = (int) (TP->VertexCnt / 8)+1;
                }
                DelCnt++;

            } // CurVert nicht gelscht

        } // metropolis

        // +---------------------------------------------------------+
        // | Abbruchkriterium                                        |
        // +---------------------------------------------------------+

        if ( ( (Processed - LastSuccess) > OP->StopThresh) ||
             ( Global.CancelRequest ) ) {

            DerJuengsteTag = TRUE;
        }

    } while ( ! DerJuengsteTag );

    // +---------------------------------------------------------+
    // | Ende der Reduktion                                      |
    // +---------------------------------------------------------+


    Global.debugMsg ("Deleted verts:%li (MaxDel:%li)\n", DeletedVerts, MaxDeleted);
    Global.debugMsg ("reducing triset (annealing) ... %li,  "
                     "undel:%li (reject:%li, rate: %.1f) "
                     "del: %li(reject:%li), E=%f, ",
                     Processed,
                     UndelCnt, UndelRejectCnt, UndelRate * 100.0,
                     DelCnt, DelRejectCnt, CurEnergy
                    );
    energy();

    SP->ReducProcessed = Processed;
    SP->ReducLastSuccess = LastSuccess;
    SP->updateDlg();
    SP->status ("");

    TP->compactFaces();
    TP->integrityCheck();
}



void REDUCE_C::reduceTriset()
{
    SavedEnergy = 0.0;
    CurEnergy   = 0.0;

    if (!Global.OptionsP->ReduceTri) return;

    Global.StatWinP->enableBackup (TRUE);

    if ( Global.OptionsP->Annealing )

        reduceAnneal();
    else
        reduceDirect();

    Global.StatWinP->enableBackup (FALSE);
}
