/**************************************************************************
*                                                                         *
* Name   : cStatWin.cpp                ( Headerdatei:                 )   *
*                                                                         *
* Klasse(n) definiert : STATWINDOW_C                                      *
* Objekt(e) deklariert: -                                                 *
* Abgeleitet von      : -                                                 *
*                                                                         *
* Aufgabe:                                                                *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include "cTypes.hpp"

#include <stdarg.h>
#include <conio.h>
#include <dos.h>
#include <stdio.h>

#include "cStatWin.hpp"
#include "cGlue.hpp"
#include "cHistWin.hpp"
#include "cOptions.hpp"
#include "cTools.hpp"
#include "cTriset.hpp"


// --- lokale Konstanten:----------------------------------------------

// --- lokale Typen: --------------------------------------------------

// --- lokale Variablen: ----------------------------------------------


/*------------------------------------------------------------------------+
|                                                                         |
| lokale Funktionen                                                       |
|                                                                         |
+------------------------------------------------------------------------*/

/*------------------------------------------------------------------------+
|                                                                         |
| Dialog msg handler                                                      |
|                                                                         |
+------------------------------------------------------------------------*/

#pragma argsused
BOOL FAR PASCAL _export statWinDlgProc (HWND hDlg, WORD Message, WORD wParam, LONG lParam)
{
    STATUSWINDOW_C *P;

    if (Message==WM_INITDIALOG) {
        Global.GlueP->addGlue (hDlg, (void*)lParam);
        P = (STATUSWINDOW_C*) lParam;
    } else
        P = (STATUSWINDOW_C*) Global.GlueP->getThisFromWnd (hDlg);

    if (!P) return FALSE;              // not processed

    switch (Message) {

        case WM_INITDIALOG:
            SetClassWord (hDlg, GCW_HICON, P->getIcon());
            if (Global.OptionsP->StatusIconic)
                ShowWindow (hDlg, SW_SHOWMINNOACTIVE);
            P->updateDlg (hDlg);
            break;
        case WM_DESTROY:
            Global.OptionsP->StatusIconic = IsIconic (hDlg);
            Global.GlueP->delGlue (hDlg);
            break;

        case WM_COMMAND:
            switch (wParam) {

                case STAT_PB_BACKUP:
                    if (Global.TrisetP)
                        Global.TrisetP->backup();
                    break;

                case IDCANCEL:
                    HWND hStopButton = GetDlgItem (hDlg, IDCANCEL);
                    EnableWindow (hStopButton, !Global.CancelRequest);
                    P->cancelRequest();
                    break;
                default:
                    return FALSE;
            }
            break;

    default:
        return FALSE;
    }
    return TRUE;                       // Message processed
}


/*------------------------------------------------------------------------+
|                                                                         |
| Methoden der Klasse STATUSWINDOW_C                                      |
|                                                                         |
+------------------------------------------------------------------------*/

STATUSWINDOW_C::STATUSWINDOW_C()             // Konstruktor
{
    hDlg      = 0;
    hIcon     = LoadIcon (Global.getInstance(), "ICO_STATWIN");
    hBarBrush = CreateSolidBrush (0x00008000L);   // dunkelgrn

    HistWinP = new HISTORYWINDOW_C (Global.getMainWnd(), SW_SHOWNORMAL, 100 );
    if (Global.OptionsP->HistoryIconic) HistWinP->iconize();
    init();
}

STATUSWINDOW_C::~STATUSWINDOW_C()            // Destruktor
{
    closeDlg();
    Global.StatWinP = NULL;

    if (hBarBrush) DeleteObject (hBarBrush);
    if (hIcon)     DestroyIcon (hIcon);

    Global.OptionsP->HistoryIconic = IsIconic (HistWinP->getWnd());
    if (HistWinP)  delete HistWinP;
}

void STATUSWINDOW_C::init()
{
    HistWinP->clear();

    EnableBackup   = FALSE;

    UpdateTicks    = 1000;

    Status[0]      = '\0';
    Percentage     = 0.0;
    Temperature    = 0.0;

    InName[0]      = '\0';
    InVerts        = 0L;
    InPls          = 0L;

    TriVerts       = 0L;
    TriFaces       = 0L;
    TriVertsValid  = 0L;
    TriFacesValid  = 0L;

    TexName[0]     = '\0';
    TexFaces       = 0L;
    TexGeom        = NULL;
    TexMethod      = NULL;

    ReducVerts1    = 0L;
    ReducVerts2    = 0L;
    ReducFaces     = 0L;
    ReducProcessed = 0L;
    ReducLastSuccess = 0L;
    OutName[0]     = '\0';
    OutSets        = 0;

    timerReset();

    LastUpdateTicks = 0;
}

void STATUSWINDOW_C::updateDlg()
{
    if (hDlg) updateDlg (hDlg);
}

void STATUSWINDOW_C::updateDlg (HWND hDlg)
{
    struct dfree Free;

    // Maus und Tastatureingaben ermglichen

    Global.yield();

    // nur updaten, wenn seit dem letzten mal mind. UpdateTicks vergangen sind

    DWORD TC = GetTickCount();

    if ( (TC-LastUpdateTicks) < UpdateTicks )
        return;
    LastUpdateTicks = TC;

    // Wenn Hauptfenster ein Icon ist: nur Prozentangabe zeigen:

    if (IsIconic (Global.getMainWnd())) {

        char Str[20];
        sprintf (Str, "Comp3D: %5.1f%%", Percentage );
        SetWindowText (Global.getMainWnd(), Str);
        return;
        }
    else {

        SetWindowText (Global.getMainWnd(), "Comp3d");
        if (IsIconic (hDlg)) {

            char Str[20];
            sprintf (Str, "Status: %5.1f%%", Percentage );
            SetWindowText (hDlg, Str);
            return;
            }
    }

    SetWindowText (hDlg, "Status");

    long   Avail     = 0L;
    int    ReducRate = 0;

    if (TimerOn)
        Elapsed = Global.elapsed (StartTime);

    getdfree (0, &Free);
    if (Free.df_sclus != 0xFFFF)
        Avail = (long) Free.df_avail * (long) Free.df_bsec * (long) Free.df_sclus;

    unsigned DiskFree = (unsigned) (Avail / 1000);
    unsigned MemFree  = (unsigned) (GetFreeSpace (0) / 1000);
    unsigned MaxFree  = (unsigned) (GlobalCompact (0) / 1000);

    if (InVerts>0)
        ReducRate = 100.0 * TriVertsValid / InVerts;

    if (hDlg==0) return;

    SetDlgItemText (hDlg, STAT_ST_STATUS,     Status);
    SetDlgItemText (hDlg, STAT_ST_PERCENTAGE, floatToString (Percentage, 5, 1) );
    SetDlgItemText (hDlg, STAT_ST_ELAP,       Global.elapsedStrHMS (Elapsed) );
    SetDlgItemText (hDlg, STAT_ST_TEMP,       floatToString (Temperature, 6, 2) );
    setDlgItemLong (hDlg, STAT_ST_MEMFREE,    MemFree);
    setDlgItemLong (hDlg, STAT_ST_MAXFREE,    MaxFree);
    setDlgItemLong (hDlg, STAT_ST_DISKFREE,   DiskFree);

    SetDlgItemText (hDlg, STAT_ST_INNAME,     InName);
    setDlgItemLong (hDlg, STAT_ST_INVERTS,    InVerts);
    SetDlgItemInt  (hDlg, STAT_ST_INPLS,      InPls, TRUE);

    setDlgItemLong (hDlg, STAT_ST_TRIVERTS,   TriVerts);
    setDlgItemLong (hDlg, STAT_ST_TRIVERTVAL, TriVertsValid);
    setDlgItemLong (hDlg, STAT_ST_TRIFACES,   TriFaces);
    setDlgItemLong (hDlg, STAT_ST_TRIFACEVAL, TriFacesValid);

    SetDlgItemText (hDlg, STAT_ST_TEXNAME,    TexName);
    SetDlgItemText (hDlg, STAT_ST_TEXGEOM,    TexGeom);
    SetDlgItemText (hDlg, STAT_ST_TEXMETHOD,  TexMethod);
    setDlgItemLong (hDlg, STAT_ST_TEXFACES,   TexFaces);

    setDlgItemLong (hDlg, STAT_ST_REDUCVERTS1,ReducVerts1);
    setDlgItemLong (hDlg, STAT_ST_REDUCVERTS2,ReducVerts2);
    setDlgItemLong (hDlg, STAT_ST_REDUCFACES, ReducFaces);
    SetDlgItemInt  (hDlg, STAT_ST_REDUCRATE,  ReducRate, TRUE);

    SetDlgItemText (hDlg, STAT_ST_OUTNAME,    OutName);
    SetDlgItemInt  (hDlg, STAT_ST_OUTSETS,    OutSets, TRUE);

    HWND hStopButton = GetDlgItem (hDlg, IDCANCEL);
    EnableWindow (hStopButton, !Global.CancelRequest);

    HWND hBackupButton = GetDlgItem (hDlg, STAT_PB_BACKUP);
    EnableWindow (hBackupButton, EnableBackup);

    // Balken zeichnen:

    RECT Rect;

    HWND hBmWnd = GetDlgItem (hDlg, STAT_BM_BAR);
    HDC  hBmDC  = GetDC (hBmWnd);

    GetWindowRect (hBmWnd, &Rect);

    HBRUSH OldBrush = SelectObject (hBmDC, hBarBrush);

    int XVal =  Percentage / 100 * (Rect.right - Rect.left);
    Rectangle (hBmDC, 0, 0, XVal, Rect.bottom-Rect.top);

    HBRUSH hBackGnd = GetStockObject (LTGRAY_BRUSH);
    SelectObject (hBmDC, hBackGnd);

    Rectangle (hBmDC, XVal, 0, Rect.right-Rect.left, Rect.bottom-Rect.top);

    SelectObject (hBmDC, OldBrush);

    ReleaseDC (hBmWnd, hBmDC);
}

BOOL STATUSWINDOW_C::enableBackup (BOOL OnOff)
{
    if (!hDlg) return FALSE;

    BOOL B = EnableBackup;

    EnableBackup = OnOff;

    HWND hBackupButton = GetDlgItem (hDlg, STAT_PB_BACKUP);
    EnableWindow (hBackupButton, EnableBackup);

    return B;
}

BOOL STATUSWINDOW_C::openDlg()
{
    HANDLE hI = Global.getInstance();
    HWND   hW = Global.getMainWnd();

    hDlg = CreateDialogParam ( hI, "DLG_STATWIN", hW,
                               MakeProcInstance ((FARPROC)statWinDlgProc, hI),
                               (LONG)this );
    if (!hDlg) return FALSE;

    ShowWindow (hDlg, SW_SHOW);
    return TRUE;
}

BOOL STATUSWINDOW_C::closeDlg()
{
    BOOL Stat;

    if (!hDlg) return FALSE;

    if ( IsWindow (hDlg) )
        Stat = DestroyWindow (hDlg);
    hDlg = 0;
    return Stat;
}

void STATUSWINDOW_C::cancelRequest()
{
    if (Global.CancelRequest)

        Global.beep();
    else {

        int MsgBoxType = MB_APPLMODAL | MB_YESNO | MB_ICONQUESTION
                         | MB_DEFBUTTON2;
        if ( MessageBox (GetTopWindow (Global.getMainWnd()),
                         "Sure ?", "Cancel conversion", MsgBoxType) == IDYES ) {
            Global.CancelRequest=TRUE;
            status ("CANCEL !!!");
        }
    }
}

void STATUSWINDOW_C::status (char *Format, ... )
{
    if (!hDlg) return;

    va_list ArgP;

    va_start (ArgP, Format);

    vsprintf (Status, Format, ArgP);

    va_end (ArgP);

    SetDlgItemText (hDlg, STAT_ST_STATUS, Status);
}

void STATUSWINDOW_C::timerReset()
{
    TimerOn = TRUE;
    Elapsed = 0.0;
    StartTime = Global.getTimeMark();
}

HWND STATUSWINDOW_C::toTop()
{
    HWND LastTop = GetTopWindow (Global.getMainWnd() );
    if (hDlg) BringWindowToTop (hDlg);
    return LastTop;
}

BOOL STATUSWINDOW_C::iconize()
{
    if (!hDlg) return FALSE;

    return ShowWindow (hDlg, SW_MINIMIZE);
}

BOOL STATUSWINDOW_C::maximize()
{
    if (!hDlg) return FALSE;

    return ShowWindow (hDlg, SW_RESTORE);
}

void STATUSWINDOW_C::logStatus (BOOL FirstTime)
{
    if ( ! Global.OptionsP->LogSettings ) return;

    Global.logMsg ("\nElapsed\t\t: %s\n", Global.elapsedStrHMSH (Elapsed));

    if (FirstTime) {
        Global.logMsg ("Pfloat\t\t: %s\n",    (sizeof (Pfloat)==4) ? "float" : "double");
        Global.logMsg ("Verts read\t: %li\n", InVerts);
        Global.logMsg ("PL's read\t: %i\n", InPls);
        Global.logMsg ("Faces mapped\t: %li\n", TexFaces);
        }
    else {
        Global.logMsg ("Processed\t: %li (LastSuccess: %li)\n", ReducProcessed,
                       ReducLastSuccess);
        Global.logMsg ("Efficiency\t: %.1f%%\n", Percentage);
        Global.logMsg ("Temperature\t: %.1fC\n", Temperature);
    }

    Global.logMsg ("Triset\t\t: Verts: %li (total: %li), Faces: %li (entries: %li)\n",
                   TriVertsValid, TriVerts, TriFacesValid, TriFaces);

    float ReducRate = 0;

    if (InVerts!=0) ReducRate = 100.0 * TriVertsValid / InVerts;

    Global.logMsg ("Reduction\t: Rate: %5.1f%%, del. Verts: %li, del. faces: %li\n",
                   ReducRate, ReducVerts1 + ReducVerts2, ReducFaces);
}
