/**************************************************************************
*                                                                         *
* Name   : cTexture.cpp                ( Headerdatei:                 )   *
*                                                                         *
* Klasse(n) definiert : TEXTURE_C                                         *
* Objekt(e) deklariert: -                                                 *
* Abgeleitet von      :                                                   *
*                                                                         *
* Aufgabe:                                                                *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include "cTypes.hpp"

#include <conio.h>
#include <dir.h>
#include <math.h>
#include <mem.h>
#include <stdio.h>
#include <string.h>

#include "cGlobal.hpp"
#include "cTexture.hpp"
#include "cBitmap.hpp"
#include "cStatWin.hpp"

// --- lokale Konstanten:----------------------------------------------

#define EPSILONFLOAT 0.00001

// --- lokale Typen: --------------------------------------------------

// --- lokale Variablen: ----------------------------------------------


/*------------------------------------------------------------------------+
|                                                                         |
| lokale Funktionen                                                       |
|                                                                         |
+------------------------------------------------------------------------*/

/*------------------------------------------------------------------------+
|                                                                         |
| Class TEXTURE_C                                                         |
|                                                                         |
+------------------------------------------------------------------------*/

TEXTURE_C::TEXTURE_C ()
{
    BMP     = NULL;
    init();
}

TEXTURE_C::~TEXTURE_C ()
{
    if (BMP) delete BMP;
    Global.TextureP = NULL;
}

BOOL TEXTURE_C::init()
{
    OPTIONS_C *OP = Global.OptionsP;
    FaceCnt = 0;
    MissCnt = 0;
    Doubles = 0;

    disposeBM();

    if (!OP ) return FALSE;

    MaxU    = OP->U;
    MaxV    = OP->V;
    Surface = OP->Surface;
    Method  = OP->MapMethod;

    if (!OP->TextureMapping) return TRUE;

    Global.StatWinP->TexGeom   = ISTypeStr [Surface];
    Global.StatWinP->TexMethod = MapMethodTypeStr [Method];

    char Name[15], Ext[5];

    fnsplit (OP->TexFile, NULL, NULL, Name, Ext);
    strcpy (Global.StatWinP->TexName, strcat (Name,Ext));

    BMP = new BITMAP_C (TRUE, 3, 0, MaxU, MaxV);
    if (!BMP) return FALSE;

    BMP->clear();

    return TRUE;
}

BOOL TEXTURE_C::write()
{
    BOOL B;

    if (!Global.OptionsP->TextureMapping) return TRUE;

    if (!BMP) return FALSE;

    switch (Global.OptionsP->TexFileType) {
        case TEX_BMP:
            Global.StatWinP->status ("writing texture (Win 3.0 DIB: %i x %i x 24 bit) ...", MaxU, MaxV);
            B = BMP->writeBMP24 (Global.OptionsP->TexFile);
            break;
        case TEX_RLE:
        default:
            Global.warning ("Texture file format not supported");
            B = TRUE;
            break;
    }

    Global.StatWinP->status ("");
    Global.StatWinP->updateDlg();

    if (!B) Global.warning ("Error writing texture");

    return B;
}

BOOL TEXTURE_C::disposeBM()
{
    if (!BMP) return FALSE;
    delete BMP;
    BMP = NULL;
    return TRUE;
}

void TEXTURE_C::xyz2uv (Ppoint3 &Pt, int &U, int &V)
{

    Pfloat      H = Global.OptionsP->H;              // Hhe des Zylinders in cm

    switch (Surface) {
        case SURF_SPHERE:
            switch (Method) {
                case MAP_CENTROID:
                case MAP_ISN:                         // ist das selbe

                    if ((Pt.x!=0) || (Pt.z!=0))
                        U = ((Pfloat)MaxU / (2.0*M_PI)) * (atan2 (Pt.x, Pt.z) + M_PI);
                    else
                        U = 0;

                    if ((Pt.x!=0) || (Pt.y!=0) || (Pt.z!=0))
                        V = (Pfloat)MaxV / M_PI *
                              atan2 (Pt.y-H/2, sqrt (Pt.x*Pt.x+Pt.z*Pt.z)) + MaxV/2;
                    else
                        V = 0;
                    break;
            }
            break;

        case SURF_CYLINDER:

            switch (Method) {
                case MAP_CENTROID:
                    break;

                case MAP_ISN:
                    V = ((Pfloat)MaxV / H) * Pt.y;
                    if (fabs (Pt.z) > EPSILONFLOAT)
                        U = ((Pfloat)MaxU / (2.0*M_PI)) * (atan2 (Pt.x, Pt.z) + M_PI);
                    else
                        if (Pt.x>0)
                            U = 0.75 * MaxU;
                        else
                            U = 0.25 * MaxU;
                    break;
                default:;
            }
            break;
        default:;
    }
}

void TEXTURE_C::setPoint (int U, int V, Pcobundl &CF)
{
    RGB_COLOR_T CI;

    if ( (U>=0) && (U<MaxU) && (V>=0) && (V<MaxV) ) {
        CI.R = (BYTE)(CF.x * 255);
        CI.G = (BYTE)(CF.y * 255);
        CI.B = (BYTE)(CF.z * 255);
        BMP->putPixRGB (U, V, CI);
        }
    else
        MissCnt++;
}

void TEXTURE_C::interpolateLinear (int V, Pfloat U1, Pfloat U2, Pcobundl C1, Pcobundl &C2)
// Interpoliert und setzt Farben einer waagerechten Linie (C1:call by value, C2: call by ref.!)
{
    int         Start = (int)U1;
    int         End   = (int)U2;
    Pfloat      Q     = fabs (U2-U1);
    int         IncU  = (Start<End)>0 ? 1 : -1;

    if (Start==End) {
        setPoint (Start, V, C1);
        return;
    }

    Pfloat      IncR  = (C2.x-C1.x) / Q;
    Pfloat      IncG  = (C2.y-C1.y) / Q;
    Pfloat      IncB  = (C2.z-C1.z) / Q;

    while (Start!=End) {
        setPoint (Start, V, C1);
        Start += IncU;
        C1.x  += IncR;
        C1.y  += IncG;
        C1.z  += IncB;
    }
    setPoint (End, V, C2);
}

#define swap(a,b,t) (t)=(a);(a)=(b);(b)=(t);

/*-- paintTri2 -----------------------------------------------------------*
 |                                                                        |
 | vgl. Glassner s.361 ff                                                 |
 |                                                                        |
 | Hat noch Fehler !!!                                                    |
 |                                                                        |
 *------------------------------------------------------------------------*/

void TEXTURE_C::paintTri2 (int U1, int V1, Pcobundl C1,
                           int U2, int V2, Pcobundl C2,
                           int U3, int V3, Pcobundl C3)
{
    int         V, Temp;
    Pfloat      UA, UE,
                DU12, DU13, DU23,
                DRU, DGU, DBU;
    int         IncU;
    Pcobundl    TempC;

    // Sortieren: V1 >= V2 >= V3, bei waagerechten Linien kleinere U's zuerst:

    if (V1<V2)  { swap (V1, V2, Temp); swap (U1, U2, Temp); swap (C1, C2, TempC); }
    if (V1<V3)  { swap (V1, V3, Temp); swap (U1, U3, Temp); swap (C1, C3, TempC); }
    if (V2<V3)  { swap (V2, V3, Temp); swap (U2, U3, Temp); swap (C2, C3, TempC); }
    if (V1==V2)
        if (U1>U2) { swap (V1, V2, Temp); swap (U1, U2, Temp); swap (C1, C2, TempC); }
    if (V2==V3)
        if (U2>U3) { swap (V2, V3, Temp); swap (U2, U3, Temp); swap (C2, C3, TempC); }

    // Distanzen in V-Richtung:

    Pfloat Q13 = (Pfloat) (V1-V3);
    Pfloat Q23 = (Pfloat) (V2-V3);
    Pfloat Q12 = (Pfloat) (V1-V2);

    // Sonderfall: Entartetes Dreieck (waagerechte Linie)

    if (Q13 < 1.0) {
        interpolateLinear (V1, U1, U2, C1, C2);
        interpolateLinear (V1, U2, U3, C2, C3);
        return;                                  // early out
    }

    // Inkrements von U, R, G und B zwischen P1 und P3 fr einen Schritt in -V Richtung:

           DU13 = (Pfloat) (U3-U1)     / Q13;
    Pfloat DR13 = (Pfloat) (C3.x-C1.x) / Q13;
    Pfloat DG13 = (Pfloat) (C3.y-C1.y) / Q13;
    Pfloat DB13 = (Pfloat) (C3.z-C1.z) / Q13;

    // Inkrements von U, R, G und B auf der Hhe von P2 fr einen Schritt in +U Richtung:

    Pfloat L;

    if (Q12 < 1.0) {
        L    = fabs (U2 - U1);         // Sonderfall: P1,P2 waagerecht
        DU12 = (Pfloat)0;
        DU23 = (Pfloat) (U3-U2) / Q23;
        }
    else {
        DU12 = (Pfloat) (U2-U1) / Q12;

        if (Q23 < 1.0) {               // Sonderfall: P2,P3 waagerecht
            L    = fabs (U3-U2);
            DU23 = (Pfloat)0;
            }
        else {
            L    = fabs ( U2 - U1 - (Pfloat)(U1-U3) / (Q12 * Q13) );
            DU23 = (Pfloat) (U3-U2) / Q23;
        }
    }

    DRU  = (Pfloat) (C2.x - (C1.x + Q12 * DR13)) / L;
    DGU  = (Pfloat) (C2.y - (C1.y + Q12 * DG13)) / L;
    DBU  = (Pfloat) (C2.z - (C1.z + Q12 * DB13)) / L;

    IncU = (DU12>=DU13) ? 1 : -1;

    if (DU12==0) IncU = 1;

    // Startwerte:

    UA = (Pfloat) U1;

    if (Q12 < 1.0)
        UE = (Pfloat) U2;
    else
        UE = (Pfloat) U1 + 0.5 * DU12;

    Pcobundl CV = C1;

    for (V=V1; V>=V3; V--) {

        Pcobundl CU = CV;

        int N = (int) (fabs (UA - UE) + 0.5);
        int I, U;

        for (I=0, U=(int)UA; I<N; U+=IncU, I++) {
            setPoint (U, V, CU);
            CU.x += DRU; CU.y += DGU; CU.z += DBU;
        }
//      setPoint (UE, V, CU);
        CV.x += DR13; CV.y += DG13; CV.z += DB13;
        UA += DU13;
        if (V<V2) UE += DU23; else UE += DU12;
    }
}

/*-- paintTri ------------------------------------------------------------*
 |                                                                        |
 | Tests Ok                                                               |
 |                                                                        |
 *------------------------------------------------------------------------*/

void TEXTURE_C::paintTri (int U1, int V1, Pcobundl C1,
                          int U2, int V2, Pcobundl C2,
                          int U3, int V3, Pcobundl C3)
{
    int         V, Temp;
    Pcobundl    CA, CE;
    Pfloat      UA, UE;
    Pcobundl    TempC;

    // Sortieren (V1 >= V2 >= V3)

    if (V1<V2) { swap (V1, V2, Temp); swap (U1, U2, Temp); swap (C1, C2, TempC); }
    if (V1<V3) { swap (V1, V3, Temp); swap (U1, U3, Temp); swap (C1, C3, TempC); }
    if (V2<V3) { swap (V2, V3, Temp); swap (U2, U3, Temp); swap (C2, C3, TempC); }

    // Distanzen in V-Richtung:

    Pfloat Q12 = (Pfloat) abs(V2-V1);
    Pfloat Q13 = (Pfloat) abs(V3-V1);
    Pfloat Q23 = (Pfloat) abs(V3-V2);

    // Sonderfall: Entartetes Dreieck (waagerechte Linie)

    if (Q13 < 1.0) {
        interpolateLinear (V1, U1, U2, C1, C2);
        interpolateLinear (V1, U2, U3, C2, C3);
        return;
    }

    // Inkrements der U, R, G und B Werte berechnen:

    Pfloat DU13 = (Pfloat)(U3-U1)/Q13;
    Pfloat DR13 = (Pfloat)(C3.x-C1.x)/Q13;
    Pfloat DG13 = (Pfloat)(C3.y-C1.y)/Q13;
    Pfloat DB13 = (Pfloat)(C3.z-C1.z)/Q13;

    // Startwerte:

    UA = UE = (Pfloat)U1;
    CA = CE = C1;

    // erstes Dreieck (Zeilenweise zwischen V1 und V2)

    if (Q12 >= 1.0) {
        Pfloat DU12 = (Pfloat)(U2-U1)/Q12;
        Pfloat DR12 = (Pfloat)(C2.x-C1.x)/Q12;
        Pfloat DG12 = (Pfloat)(C2.y-C1.y)/Q12;
        Pfloat DB12 = (Pfloat)(C2.z-C1.z)/Q12;

        for (V=V1; V>=V2; V--) {
            interpolateLinear (V, UA, UE, CA, CE);
            UA += DU13; CA.x += DR13; CA.y += DG13; CA.z += DB13;
            UE += DU12; CE.x += DR12; CE.y += DG12; CE.z += DB12;
        }
    }

    // zweites Dreieck (Zeilenweise zwischen V2-1 und V3)

    if (Q23 >= 1.0) {
        Pfloat DU23 = (Pfloat)(U3-U2)/Q23;
        Pfloat DR23 = (Pfloat)(C3.x-C2.x)/Q23;
        Pfloat DG23 = (Pfloat)(C3.y-C2.y)/Q23;
        Pfloat DB23 = (Pfloat)(C3.z-C2.z)/Q23;

        UE = (Pfloat)U2 + DU23;
        CE.x = C2.x + DR23;
        CE.y = C2.y + DG23;
        CE.z = C2.z + DB23;

        for (V=V2-1; V>=V3; V--) {
            interpolateLinear (V, UA, UE, CA, CE);
            UA += DU13; CA.x += DR13; CA.y += DG13; CA.z += DB13;
            UE += DU23; CE.x += DR23; CE.y += DG23; CE.z += DB23;
        }
    }
}

void TEXTURE_C::mapTriangle (Pptco3 &P1, Pptco3 &P2, Pptco3 &P3)
// erzeugt ein Dreieck auf der Textur, wobei die Flchenfarben linear interpoliert werden:
{
    int         U1, V1, U2, V2, U3, V3;

    if (!Global.OptionsP->TextureMapping) return;

    FaceCnt++;
    Global.StatWinP->TexFaces = FaceCnt;

    // Texturkoordinaten (U,V) der Eckpunkte berechnen:

    xyz2uv (P1.point, U1, V1);
    xyz2uv (P2.point, U2, V2);
    xyz2uv (P3.point, U3, V3);

    // Abfangen von Dreiecken, die BitMap in U-Richtung berschreiten:

    int MaxU_2 = (MaxU + 1) / 2;

    if ( (abs (U2-U1) > MaxU_2) || (abs (U3-U1) > MaxU_2) || (abs (U3-U2) > MaxU_2) )
        return;

    if ( Global.OptionsP->Interpolate )
        paintTri (U1, V1, P1.colour, U2, V2, P2.colour, U3, V3, P3.colour);
    else {
        setPoint (U1, V1, P1.colour);
        setPoint (U2, V2, P2.colour);
        setPoint (U3, V3, P3.colour);
    }
}
