
/**************************************************************************
*                                                                         *
* Name   : cTools.cpp                  ( Headerdatei:                 )   *
*                                                                         *
* Klasse(n) definiert : -                                                 *
*                                                                         *
* Aufgabe: Sammlung von Hilfsroutinen                                     *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include "cTypes.hpp"

#include <mem.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cGlobal.hpp"
#include "cTools.hpp"


/*------------------------------------------------------------------------+
|                                                                         |
| lokale Funktionen                                                       |
|                                                                         |
+------------------------------------------------------------------------*/

/*------------------------------------------------------------------------+
|                                                                         |
| Debugging                                                               |
|                                                                         |
+------------------------------------------------------------------------*/

/***************      HeapWalking luft nicht unter Windows ?

int checkHeapNode (void *P)
{
    int Stat;

    Global.debugMsg ("Checking node %p: ", P);

    Stat = farheapchecknode (P);

    switch (Stat) {

       case _HEAPEMPTY:
            Global.debugMsg ("*** no heap ***\n");
            break;
       case _HEAPCORRUPT:
            Global.debugMsg ("*** heap corrupt ***\n");
            break;
       case _BADNODE:
            Global.debugMsg ("*** bad node ***\n");
            break;
       case _FREEENTRY:
            Global.debugMsg ("*** free entry ***\n");
            break;
       case _USEDENTRY:
            Global.debugMsg ("used entry\n");
            break;
       default:
            Global.debugMsg ("*** unknown status ***\n");
            break;
    }
    return Stat;
}


BOOL checkHeap()
{
    if( farheapcheck() == _HEAPCORRUPT ) {
        Global.debugMsg ("*** heap is corrupted ***\n");
        return FALSE;
    }

    Global.debugMsg ("heap checks ok\n");
    return TRUE;
}

BOOL walkHeap()
{
    BOOL Stat;

    Global.debugMsg ("walking heap: ");

    if ( farheapfillfree (-1) < 0 ) {

        Global.debugMsg ("*** heap corrupt ***\n");
        return FALSE;
    }

    Stat = farheapcheckfree (-1);

    if ( Stat< 0 )

        switch ( Stat ) {

            case _HEAPCORRUPT:
                Global.debugMsg ("*** heap corrupt ***\n");
                return FALSE;
            case _BADVALUE:
                Global.debugMsg ("*** bad value in free space ***\n");
                return FALSE;
            default:
                Global.debugMsg ("*** unknown error ***\n");
                return FALSE;
        }
        Global.debugMsg ("checks ok\n");

    //------------------------------------------------------

    struct heapinfo HI;

    HI.ptr = NULL;

    Global.debugMsg ("\tAddr\tSize\tStat\n");

    while( farheapwalk( &HI ) == _HEAPOK )

        Global.debugMsg ("\t%p\t%7u\t%s\n", HI.ptr, HI.size, HI.in_use ? "used" : "free" );

    return TRUE;
}

***********************/

/*------------------------------------------------------------------------+
|                                                                         |
| Files:                                                                  |
|                                                                         |
+------------------------------------------------------------------------*/

long fileSize (FILE *stream)
{
   long curpos, length;

   curpos = ftell(stream);
   fseek(stream, 0L, SEEK_END);
   length = ftell(stream);
   fseek(stream, curpos, SEEK_SET);
   return length;
}

BOOL fileExist (char *lpszFSpec)
{
    OFSTRUCT FileInfoRec;

    if (!lpszFSpec) return FALSE;

    return ( OpenFile (lpszFSpec, &FileInfoRec, OF_EXIST) != -1 );
}

/*------------------------------------------------------------------------+
|                                                                         |
| Strings                                                                 |
|                                                                         |
+------------------------------------------------------------------------*/

char* floatToString (float F, int Width, int Digits)
{
    static char FString [30];

    sprintf (FString, "%*.*f", Width, Digits, F);
    return FString;
}

BOOL stringToFloat (char *S, float *F)
{
   int Stat = sscanf (S, "%f", F);

   return (Stat==1);
}

char* stripZeros (double F)
{
    static char FloatStr [30];

    int P = sprintf (FloatStr, ">%lf", F);

    //  0.000 -> 0.0
//  while ( (FloatStr[--P]=='0') && (FloatStr[P-1]!='.') ) FloatStr[P]=0;
//  return FloatStr+1;

    //  0.000 -> 0
    while (FloatStr[--P]=='0') {

        FloatStr[P]=0;
    }
    if (FloatStr [P]=='.') FloatStr[P] = 0;

    return FloatStr+1;
}

/*------------------------------------------------------------------------+
|                                                                         |
| Dialogboxen                                                             |
|                                                                         |
+------------------------------------------------------------------------*/

void vSetDlgItemText (HWND hDlg, int IdDlgItem, char *Format, ...)
{
    char    Msg[255];
    va_list ArgP;

    va_start (ArgP, Format);
    if (strlen (Format)+1 < sizeof (Msg))
        vsprintf (Msg, Format, ArgP);
    va_end (ArgP);

    SetDlgItemText (hDlg, IdDlgItem, Msg);
}

void newFloatEdit (HWND hDlg, WORD W, float &F, int Width, int Digits,
                          float Min, float Max)
{
    char  TempStr [20];
    float TempF;

    GetDlgItemText (hDlg, W, TempStr, sizeof (TempStr)-1);

    if (stringToFloat (TempStr, &TempF)) F = TempF;
    if (F<Min) F=Min;
    else if (F>Max) F=Max;

    SetDlgItemText (hDlg, W, floatToString (F, Width, Digits));
}

void newIntEdit (HWND hDlg, WORD W, int &I, int Min, int Max)
{
    int   TempI;
    BOOL  Translated;

    TempI = GetDlgItemInt (hDlg, W, &Translated, TRUE);

    if (Translated) I = TempI;

    if (I<Min) I=Min; else if (I>Max) I=Max;

    SetDlgItemInt (hDlg, W, I, TRUE);
}

void newLongEdit (HWND hDlg, WORD W, long &L, long Min, long Max)
{
    char TempStr[33];

    GetDlgItemText (hDlg, W, TempStr, sizeof (TempStr)-1);

    L = atol (TempStr);
    if (L<Min) L=Min;
    else if (L>Max) L=Max;

    SetDlgItemText (hDlg, W, ltoa (L, TempStr, 10));
}

void setDlgItemLong (HWND hDlg, int IdDlgItem, long Value)
{
    vSetDlgItemText (hDlg, IdDlgItem, "%li", Value);
}

/*------------------------------------------------------------------------+
|                                                                         |
| Mens                                                                   |
|                                                                         |
+------------------------------------------------------------------------*/

#define ENABLEMENU   MF_BYCOMMAND|MF_ENABLED
#define GRAYMENU     MF_BYCOMMAND|MF_GRAYED

BOOL enableMenuItem (HMENU hMenu, int IdDlgItem, BOOL Enable)
{
    return EnableMenuItem (hMenu, IdDlgItem, Enable ? ENABLEMENU : GRAYMENU);
}

#define  CHECKMENU    MF_BYCOMMAND|MF_CHECKED
#define  UNCHECKMENU  MF_BYCOMMAND|MF_UNCHECKED

BOOL checkMenuItem (HMENU hMenu, int IdDlgItem, BOOL Check)
{
    return CheckMenuItem (hMenu, IdDlgItem, Check ? CHECKMENU : UNCHECKMENU);
}

BOOL toggleMenuItem (HMENU hMenu, int IdDlgItem, BOOL &Var)
{
    Var = !Var;
    return CheckMenuItem (hMenu, IdDlgItem, Var ? CHECKMENU : UNCHECKMENU);
}
