/**************************************************************************
*                                                                         *
* Name   : cTriset.cpp                 ( Headerdatei:                 )   *
*                                                                         *
* Klasse(n) definiert : TRISET_C                                          *
* Objekt(e) deklariert: -                                                 *
* Abgeleitet von      : -                                                 *
*                                                                         *
* Aufgabe:                                                                *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include "cTypes.hpp"

#include <dir.h>
#include <limits.h>
#include <mem.h>
#include <stdlib.h>
#include <string.h>

#include "cGlobal.hpp"
#include "cTriset.hpp"
#include "cTexture.hpp"
#include "cReduce.hpp"
#include "cMath.hpp"
#include "cPolyLin.hpp"
#include "cStatWin.hpp"
#include "cGeomOut.hpp"

extern "C" {

    // externe 32-bit Assembler Routine. Sucht ein Face, da 'Index' als ein Vertex hat
    BOOL scanFaceArray (FACE_S **Start, long *MaxCnt, long *Index, long Value);
}

// --- lokale Konstanten:----------------------------------------------

// --- lokale Typen: --------------------------------------------------

// --- lokale Variablen: ----------------------------------------------

/*------------------------------------------------------------------------+
|                                                                         |
| Globale Funktionen fr Zugriff auf FACE_S und den damit verknpften     |
| Hidden Vertex Listen. Knnen sowohl fr                                 |
|     TRISET_C.FaceArrayList[].HVP... als auch fr                        |
|     FACEARRAY_C.Val[].HVP                                               |
| benutzt werden.                                                         |
|                                                                         |
+------------------------------------------------------------------------*/


void _delFace (FACE_S *FP)
{
    long *FIP = FP->Index;

    FIP[0] = FIP[1] = FIP[2] = 0L;

    // Speicher fr hidden verts freigeben:

    _clearHV (FP);
}

BOOL _isFaceDeleted (long *VP)
{
    if (VP[0]==0.0) if (VP[1]==0.0) if (VP[2]==0.0) return TRUE;

    return FALSE;
}

BOOL _addHV (FACE_S *FP, long VertIdx)
{
    if (FP->N>=FP->MaxN)
        if ( ! _newHVSize (FP, FP->MaxN + IndexArrayIncrement) )
            return FALSE;

    FP->HVP [FP->N++] = VertIdx;

    return TRUE;
}

void _clearHV (FACE_S *FP)
{
    if (FP->HVP) {

        free (FP->HVP);
        FP->HVP = NULL;
    }
    FP->N = FP->MaxN = 0;
}

BOOL _adjustHVSize (FACE_S *FP)
{
    return _newHVSize (FP, FP->N);
}

BOOL _newHVSize (FACE_S *FP, int aVertexCnt)
{
    unsigned NewSize = (unsigned) (aVertexCnt * sizeof (*FP->HVP) );

    if (NewSize > 0xFF00) {
        Global.fatal ("_newHVSize: too many entries");
        return FALSE;
    }

    FP->HVP  = (long*) realloc (FP->HVP, NewSize);

    if (FP->HVP == NULL) {
        Global.fatal ("_newHVSize: out of mem");
        return FALSE;
    }

    FP->MaxN = aVertexCnt;

    return TRUE;
}

/*------------------------------------------------------------------------+
|                                                                         |
| lokale Funktionen                                                       |
|                                                                         |
+------------------------------------------------------------------------*/


/*------------------------------------------------------------------------+
|                                                                         |
| Methoden der Klasse FACEARRAY_C                                         |
|                                                                         |
+------------------------------------------------------------------------*/

FACEARRAY_C::FACEARRAY_C (int aMax, int aGrowBy)     // Konstruktor
:DYNARRAY_C<FACE_S> (aMax, aGrowBy)
{
}

FACEARRAY_C::~FACEARRAY_C()                          // Destruktor
{
}

BOOL FACEARRAY_C::newSize (int nFaces)
{
    if (!DYNARRAY_C<FACE_S>::newSize (nFaces))
        return FALSE;

    for (int I=N; I<MaxN; I++)
        _clearHV (&Val[I]);

    return TRUE;
}

BOOL FACEARRAY_C::copyFrom (FACEARRAY_C *SourceFAP)
{
    int I;

    // alte Faces und HV's lschen

    clear();

    int nSourceFaces = SourceFAP->nElem();

    for (I=0; I < nSourceFaces; I++) {

        int nSourceHV = SourceFAP->Val[I].N;

        append (SourceFAP->Val[I]);

        // Hidden Verts mssen extra kopiert werden:

        if (nSourceHV==0)
            continue;

        Val[I].HVP = (long*) calloc (nSourceHV, sizeof (long));

        if ( !Val[I].HVP ) {

            Global.fatal ("FACEARRAY_C::copyFrom: Out of mem");
            return FALSE;
        }
        memcpy (Val[I].HVP, SourceFAP->Val[I].HVP,
                nSourceHV * sizeof (long) );
    }
    return TRUE;
}

BOOL FACEARRAY_C::deleteLast()
{
    _clearHV (&Val[N-1]);

    return DYNARRAY_C<FACE_S>::deleteLast();
}

void FACEARRAY_C::clear()
{
    for (int I=0; I<N; I++) {

        _clearHV (&Val[I]);
    }
    DYNARRAY_C<FACE_S>::clear();
}


/*------------------------------------------------------------------------+
|                                                                         |
| Methoden der Klasse TRISET_C                                            |
|                                                                         |
+------------------------------------------------------------------------*/

TRISET_C::TRISET_C()             // Konstruktor
{
    Echo           = FALSE;
    Swapped        = FALSE;
    MapTriangles   = TRUE;
    StoreTriangles = TRUE;
    FaceCnt        = 0;
    VertexCnt      = 0;
    VertHashSize   = 0;
    StartSearch    = 0L;

    int I;
    for (I=0; I<MaxVertexArrays; I++) VertArrayList [I] = NULL;
    for (I=0; I<MaxFaceArrays; I++)   FaceArrayList [I] = NULL;
    for (I=0; I<MaxVertHashArrays; I++)
                                      VertHashArrayList [I] = NULL;

    // Sicherheitskontrolle: der Compiler mu hier
    // 'condition always false' bringen !

    if ( ( (long)FacesPerArray * sizeof (FACE_S) > 0xFFF0) ||
         ( (long)VertsPerArray * sizeof (VERTEX_T) > 0xFFF0) ||
         ( (long)VertHashEntriesPerArray * sizeof (long) > 0xFFF0) )

        Global.fatal ("TRISET_C: Arrays sind zu gro");
}

TRISET_C::~TRISET_C()            // Destruktor
{
    destroyVertHashTab();
    clear();

//  Global.TrisetP = NULL;   // ACHTUNG: 2.TRISET wird von ZofWrite temporr angelegt!!
}

void TRISET_C::init()
{
    destroyVertHashTab();
    clear();
}

BOOL TRISET_C::echo (BOOL OnOff)
{
    BOOL B = Echo;
    Echo = OnOff;
    return B;
}

void TRISET_C::updateStatus()
{
    if (!Echo) return;

    Global.StatWinP->TriVerts      = VertexCnt;
    Global.StatWinP->TriFaces      = FaceCnt;
    Global.StatWinP->TriVertsValid = VisibleVertexCnt;
    Global.StatWinP->TriFacesValid = ValidFaceCnt;
    Global.StatWinP->updateDlg();
}


/*------------------------------------------------------------------------+
|                                                                         |
| elemantare Methoden fr Zugriff auf BigArrays                           |
|                                                                         |
+------------------------------------------------------------------------*/


// BC 3.0: Inlines funktionieren nicht! Darum fr den Gebrauch
//         innerhalb des Moduls Makrodefinitionen:

#define MacroGetVertP(I)     ( &VertArrayList [int (I >> Log2VertsPerArray)] \
                                              [int (I & VertexOffsMask)] )
#define MacroGetVertPtP(I)   ( &VertArrayList [int (I >> Log2VertsPerArray)] \
                                              [int (I & VertexOffsMask)].PC.point)
#define MacroGetVertPtCoP(I) ( &VertArrayList [int (I >> Log2VertsPerArray)] \
                                              [int (I & VertexOffsMask)].PC)
#define MacroGetFaceP(I)     ( &FaceArrayList [int (I>>Log2FacesPerArray)] \
                                              [int (I & FaceOffsMask)])
#define MacroGetFaceIdxP(I)  ( &FaceArrayList [int (I >> Log2FacesPerArray)] \
                                              [int (I & FaceOffsMask)].Index[0])

VERTEX_T *TRISET_C::getVertP (long Index)
{
    return MacroGetVertP (Index);
}

Ppoint3 *TRISET_C::getVertPtP (long Index)
{
    return MacroGetVertPtP (Index);
}

Pptco3 *TRISET_C::getVertPtCoP (long Index)
{
    return MacroGetVertPtCoP (Index);
}

FACE_S *TRISET_C::getFaceP (long Index)
{
    return MacroGetFaceP (Index);
}

long *TRISET_C::getFaceIdxP (long Index)
{
    return MacroGetFaceIdxP (Index);
}

void TRISET_C::markFace (long Fidx, BYTE R, BYTE G, BYTE B)
{
    Pcobundl *CP;

    long *FIP = MacroGetFaceIdxP (Fidx);

    if ( !existFace (FIP[0], FIP[1], FIP[2]) ) return;

    CP = & (MacroGetVertPtCoP (FIP[0]) -> colour);
    CP->x = (float)R/255; CP->y = (float)G/255; CP->z = (float)B/255;

    CP = & (MacroGetVertPtCoP (FIP[1]) -> colour);
    CP->x = (float)R/255; CP->y = (float)G/255; CP->z = (float)B/255;

    CP = & (MacroGetVertPtCoP (FIP[2]) -> colour);
    CP->x = (float)R/255; CP->y = (float)G/255; CP->z = (float)B/255;

}

BOOL TRISET_C::isFaceDeleted (long Index)
{
    long *VP = MacroGetFaceIdxP (Index);

    if (VP[0]==0L) if (VP[1]==0L) if (VP[2]==0L) return TRUE;

    return FALSE;
}

BOOL TRISET_C::existFace (long I1, long I2, long I3)
{
    for (long F=0; F<FaceCnt; F++) {

        long *VP = MacroGetFaceIdxP (F);

        if ( ( (VP[0]==I1) || (VP[0]==I2) || (VP[0]==I3) ) &&
             ( (VP[1]==I1) || (VP[1]==I2) || (VP[1]==I3) ) &&
             ( (VP[2]==I1) || (VP[2]==I2) || (VP[2]==I3) ) )
            return TRUE;
    }

    return FALSE;
}

long TRISET_C::findOrientatedEdge (long VIdx1, long VIdx2, long FirstFaceIdx)

// Liefert den Index eines Faces, das eine Edge (VIdx1, VIdx2) hat.
//         NOINDEX, wenn nicht gefunden

{
    long     CurFaceIdx = FirstFaceIdx-1;
    int      CurBlock   = int (FirstFaceIdx >> Log2FacesPerArray);
    FACE_S  *CurFAP     = MacroGetFaceP (FirstFaceIdx);
    int      MaxBlock   = int (FaceCnt >> Log2FacesPerArray);

    long     RestLen    = ( CurBlock==MaxBlock ) ?
                            ((FaceCnt - FirstFaceIdx) & FaceOffsMask)
                            : FacesPerArray - (FirstFaceIdx & FaceOffsMask);

    BOOL Found = FALSE;

    do {
        // scanFaceArray (StartFaceP, Cnt, FoundIdx, VertIdx) sucht ab <StartFaceP>
        // die nchsten <Cnt> Faces nach dem Vertex <VertIdx> ab.
        // Wenn gefunden, ist <Cnt>        = Rest,
        //                    <FoundIdx>   = Index des gefundenen Faces,
        //                    <StartFaceP> zeigt auf den nchsten Wert.
        // Es werden zunchst alle Faces gesucht, in denen VIdx1 vorkommt. Diese
        // werden dann weiter untersucht.

        if ( scanFaceArray ( &CurFAP, &RestLen, &CurFaceIdx, VIdx1 ) ) {

                // Face <*(CurFAP-1)> enthlt Vertex <VIdx1>:

                long *FIP = (CurFAP-1)->Index;

                if ( FIP[0]==VIdx1 ) {
                    if (FIP[1]==VIdx2) Found = TRUE;
                } else
                    if ( FIP[1]==VIdx1 ) {
                        if (FIP[2]==VIdx2) Found = TRUE;
                    } else
                        if ( FIP[2]==VIdx1 ) {
                            if (FIP[0]==VIdx2) Found = TRUE;
                        } else
                            Global.debugInfo ("findOrientatedEdge: was faul");
                if (Found) break;

        } else {

            // nchsten 64k-Block dursuchen:

            if ( CurBlock == MaxBlock ) {

                break;                 // fertig
                }
            else {                     // nchstes Array

                CurBlock++;
                CurFAP   = FaceArrayList[CurBlock];
                RestLen  = ( CurBlock==MaxBlock ) ?
                            (FaceCnt & FaceOffsMask) : FacesPerArray ;
            }
        }

    } while (TRUE);

//***************************
// redundante Kontrolle:

/*
    long *FIP_;

    if (Found) {

        FIP_ = MacroGetFaceIdxP (CurFaceIdx);

        if ( ! ( ( (FIP_[0]==VIdx1) && (FIP_[1]==VIdx2) ) ||
                 ( (FIP_[1]==VIdx1) && (FIP_[2]==VIdx2) ) ||
                 ( (FIP_[2]==VIdx1) && (FIP_[0]==VIdx2) ) ) )
                Global.info ("findOrientatedEdge: sagt ja, ist aber falsch");
    } else {

        for (long X=FirstFaceIdx; X<FaceCnt; X++) {

            FIP_ = MacroGetFaceIdxP (X);

            if ( ( (FIP_[0]==VIdx1) && (FIP_[1]==VIdx2) ) ||
                 ( (FIP_[1]==VIdx1) && (FIP_[2]==VIdx2) ) ||
                 ( (FIP_[2]==VIdx1) && (FIP_[0]==VIdx2) ) )
                Global.info ("findOrientatedEdge: sagt nein, ist aber trotzdem da");

        }
    }
*/
//***************************

    if ( Found )
        return CurFaceIdx;

    return NOINDEX;
}

long TRISET_C::findDoubleEdge (long VIdx1, long VIdx2, long VIdx3, long FirstFaceIdx)

// Durchsucht das FaceArray nach dem ersten Auftreten einer Edge:
// (VIdx1, VIdx2), (VIdx2, VIdx3) oder (VIdx3, VIdx1)
// liefert NOINDEX, wenn nicht gefunden

{
    long FoundIdx;

    FoundIdx = findOrientatedEdge (VIdx1, VIdx2, FirstFaceIdx);
    if (FoundIdx!=NOINDEX) return FoundIdx;

    FoundIdx = findOrientatedEdge (VIdx2, VIdx3, FirstFaceIdx);
    if (FoundIdx!=NOINDEX) return FoundIdx;

    return findOrientatedEdge (VIdx3, VIdx1, FirstFaceIdx);
}

BOOL TRISET_C::existDoubleEdgeExcept (long VIdx1, long VIdx2, long VIdx3,
                                      INDEXARRAY_C *FaceIdxList
                                     )

// Durchsucht das FaceArray nach dem ersten Auftreten einer Edge:
// (VIdx1, VIdx2), (VIdx2, VIdx3) oder (VIdx3, VIdx1). Die Faces in FaceIdxList
// werden dabei ignoriert.
// liefert TRUE, wenn gefunden

{
    long FoundIdx;
    long FirstFaceIdx;
    BOOL FoundException;

    int  nExceptions = FaceIdxList->nElem();

    // fr jede der 3 Edges:

    for (int X=0; X<3; X++) {

        FirstFaceIdx = 0;

        do {
            switch ( X ) {

                case 0: FoundIdx = findOrientatedEdge (VIdx1, VIdx2, FirstFaceIdx);
                        break;
                case 1: FoundIdx = findOrientatedEdge (VIdx2, VIdx3, FirstFaceIdx);
                        break;
                case 2: FoundIdx = findOrientatedEdge (VIdx3, VIdx1, FirstFaceIdx);
                        break;
            }

            FoundException = FALSE;

            if (FoundIdx==NOINDEX)
                break;           // Edge nicht gefunden -> nchste Edge(X) suchen

            for (int I=0; I<nExceptions; I++) {

                if (FoundIdx==FaceIdxList->Val[I]) {

                    FoundException = TRUE;
                    break;       // weitersuchen
                }
            }
            FirstFaceIdx = FoundIdx + 1;

        } while (FoundException);

        if (FoundIdx!=NOINDEX) return TRUE;      // Eine gefundene Edge reicht fr TRUE
    }
    return FALSE;
}

void TRISET_C::delFace (long Index)
{
    FACE_S      *FP  = MacroGetFaceP (Index);
    long        *FIP = FP->Index;

    FIP[0] = FIP[1] = FIP[2] = 0L;

    // Speicher fr hidden verts freigeben:

    _clearHV (FP);

    ValidFaceCnt--;
}

/*------------------------------------------------------------------------+
|                                                                         |
| Methoden zum Zugriff auf VertexArrays ber Hash-Tabelle                 |
|                                                                         |
+------------------------------------------------------------------------*/

BOOL TRISET_C::createVertHashTab (long MaxEntries)
{
    if (VertexCnt!=0)
        Global.warning ("createVertHashTab: too late");

    destroyVertHashTab();

    int ArrayCnt = int (MaxEntries / VertHashEntriesPerArray);

    for (int I=0; I<ArrayCnt; I++) {

        VertHashArrayList[I] = (long*) malloc (VertHashArraySize);

        if (! VertHashArrayList[I]) {

            Global.warning ("Vertex hash table: Out of mem");
            destroyVertHashTab();
            return FALSE;
        }
        memset (VertHashArrayList[I], -1, VertHashArraySize);
    }

    int RestSize = int ( MaxEntries % VertHashEntriesPerArray ) * sizeof (long);

    if ( RestSize > 0) {

        VertHashArrayList[I] = (long*) malloc (RestSize);

        if (! VertHashArrayList[I]) {

            Global.warning ("Vertex hash table: Out of mem");
            destroyVertHashTab();
            return FALSE;
        }
        memset (VertHashArrayList[I], -1, RestSize);
    }

    VertHashSize = MaxEntries;
    return TRUE;
}

void TRISET_C::destroyVertHashTab()
{
    for (int I=0; I<MaxVertHashArrays; I++)
        if (VertHashArrayList[I]) {

            free (VertHashArrayList[I]);
            VertHashArrayList[I] = NULL;
        }
    VertHashSize = 0;
}

long TRISET_C::addVertexHash (Pptco3 &VP)

// liefert den Index der Vertex VP. Wenn VP noch nicht existiert, wird ein
// neuer Punkt angelegt.

{
    if (VertHashSize == 0L)
        return addVertex (VP);

    // Hash-Key generieren:

    long HashIdx = labs ((long)(VP.point.y * 50000.0 + VP.point.x) % VertHashSize);

    long Cnt = 0L;

    do {

        long *HashP = &VertHashArrayList [int (HashIdx / VertHashEntriesPerArray)]
                                         [int (HashIdx % VertHashEntriesPerArray)];

        if (*HashP==-1L) {

            *HashP = appendVertex (VP);
            return *HashP;
        }

        Pptco3 *TmpVP = MacroGetVertPtCoP (*HashP);

        if ( ptsEqual (VP.point, TmpVP->point) ) {

            return *HashP;
        }

        HashIdx++;
        HashIdx = HashIdx % VertHashSize;
        Cnt++;

    } while ( Cnt < 5000 );

    Global.warning ("addVertexHash: Hash table too small. Trying without...");
    destroyVertHashTab();
    return addVertex (VP);
}


/*------------------------------------------------------------------------+
|                                                                         |
| Methoden zur Verwaltung der Hidden Vertex Listen                        |
|                                                                         |
+------------------------------------------------------------------------*/


BOOL TRISET_C::addHV (long FaceIdx, long VertIdx)
{
    FACE_S *FP  = MacroGetFaceP (FaceIdx);

    if (FP->N>=FP->MaxN)
        if ( ! _newHVSize (FP, FP->MaxN + IndexArrayIncrement) )
            return FALSE;

    FP->HVP [FP->N++] = VertIdx;

    MacroGetVertP (VertIdx) -> FI = FaceIdx;

    return TRUE;
}

BOOL TRISET_C::adjustHVSize (long FaceIdx)
{
    FACE_S *FP  = MacroGetFaceP (FaceIdx);

    return _newHVSize (FP, FP->N);
}

BOOL TRISET_C::newHVSize (long FaceIdx, int aVertexCnt)
{
    return _newHVSize (MacroGetFaceP (FaceIdx), aVertexCnt);
}


/*------------------------------------------------------------------------+
|                                                                         |
| Methoden zur Verwaltung der BigArrays                                   |
|                                                                         |
+------------------------------------------------------------------------*/


void TRISET_C::compactFaces()

// Triset bereinigen: gelschte Faces (0,0,0) werden entfernt, nicht
// bentigte FaceArrays freigegeben.

{
    int    A, V;
    long   F;
    FACE_S *FP;
    long   *FIP;

    if (FaceCnt==0L) return;

    if (Echo)
        Global.StatWinP->status ("compacting triset face array(s) ...");

    long OldFaceCnt = FaceCnt;

    long NextFree   = 0L;
    long DelFaceCnt = 0L;

    // ungltige Faces lschen:

    long FirstFree = 0L;

    while ( !isFaceDeleted (FirstFree) ) {

        FirstFree++;
        if (FirstFree >= FaceCnt) goto FacesCompacted;
    }
    NextFree = FirstFree;

//  if (FirstFree==0L) DelFaceCnt++;
    DelFaceCnt = 1;

    for (F=FirstFree+1; F<FaceCnt; F++) {

        if ( isFaceDeleted (F) ) {

            DelFaceCnt++;
            }
        else {

            memcpy (MacroGetFaceP(NextFree), MacroGetFaceP(F), sizeof (FACE_S));
            NextFree++;
        }
    }

    FaceCnt = FaceCnt - DelFaceCnt;

    // Leere Blcke freigeben:

    for (A=int(FaceCnt >> Log2FacesPerArray)+1; A<=int(OldFaceCnt >> Log2FacesPerArray); A++) {

        if (FaceArrayList[A]) free (FaceArrayList[A]);
        FaceArrayList[A] = NULL;
    }

FacesCompacted:;

    // Vater-Face Eintrag der Vertizes updaten:

    for (F=0; F<FaceCnt; F++) {

        FP  = MacroGetFaceP (F);
        FIP = &FP->Index[0];

        MacroGetVertP (FIP[0])->FI = NULL;
        MacroGetVertP (FIP[1])->FI = NULL;
        MacroGetVertP (FIP[2])->FI = NULL;

        if (FP->HVP) {

            for (V=0; V<FP->N; V++) {

                MacroGetVertP (FP->HVP[V])->FI = F;
            }
        }
    }

    ValidFaceCnt = FaceCnt;

    if (Echo) {

        Global.StatWinP->TriFaces      = FaceCnt;
        Global.StatWinP->TriFacesValid = ValidFaceCnt;
        Global.StatWinP->status ("");
    }
}

void TRISET_C::clear()
{
    int I;

    FaceCnt          = 0;
    ValidFaceCnt     = 0;
    VertexCnt        = 0;
    VisibleVertexCnt = 0;

    for (I=0; I<MaxVertexArrays; I++) if (VertArrayList [I]) {
        free (VertArrayList [I]);
        VertArrayList [I] = NULL;
    }
    for (I=0; I<MaxFaceArrays; I++) if (FaceArrayList[I]) {
        free (FaceArrayList [I]);
        FaceArrayList [I] = NULL;
    }

    long TempS = VertHashSize;

    destroyVertHashTab();
    createVertHashTab (TempS);

    if (Echo && Global.StatWinP)
        updateStatus();
}

long TRISET_C::appendVertex (Pptco3 &V)
{
    int N = int (VertexCnt >> Log2VertsPerArray);

    if (VertArrayList[N]==NULL) {

        VertArrayList[N] = (VERTEX_T*) calloc (VertsPerArray, sizeof(VERTEX_T));

//      Global.debugMsg ("Vertex Array #%i (%p): %li bytes\n", N, VertArrayList[N],
//                       (long)VertsPerArray*sizeof(VERTEX_T) );

        if (!VertArrayList[N]) {
            Global.fatal ("TRISET::appendVertex: out of mem");
            return 0L;
        }
    }

    VERTEX_T *VP = MacroGetVertP (VertexCnt);

    VP->PC = V;
    VP->FI = NULL;

    VertexCnt++;
    VisibleVertexCnt++;

    updateStatus();

    return VertexCnt - 1;
}

long TRISET_C::addVertex (Pptco3 &Pt, long aStartSearch)
{
    Ppoint3 P = Pt.point;

    if (aStartSearch>VertexCnt) aStartSearch=VertexCnt;

    for (long I=aStartSearch; I<VertexCnt; I++) {

        Ppoint3 *PIP = MacroGetVertPtP (I);

        if (PIP->x==P.x)
            if (PIP->y==P.y)
                if (PIP->z==P.z)
                    break;
    }

    if (I==VertexCnt) {                // Vertex noch nicht vorhanden

        return appendVertex (Pt);
    }
    return I;
}

void TRISET_C::add3Vertices (Pptco3 &PC1, Pptco3 &PC2, Pptco3 &PC3,
                             long   &I1,  long   &I2,  long   &I3,
                             BOOL   &MatchedAll,
                             long   aStartSearch)
{
    Ppoint3 P1    = PC1.point;
    Ppoint3 P2    = PC2.point;
    Ppoint3 P3    = PC3.point;

    I1 = I2 = I3 = -1L;

    if (aStartSearch>VertexCnt)
        aStartSearch = VertexCnt;

    for (long I=aStartSearch; I<VertexCnt; I++) {

        Ppoint3 *PP = MacroGetVertPtP (I);

        if (PP->x==P1.x)
            if (PP->y==P1.y)
                if (PP->z==P1.z) {

                    I1 = I;
                }

        if (PP->x==P2.x)
            if (PP->y==P2.y)
                if (PP->z==P2.z) {

                    I2 = I;
                }

        if (PP->x==P3.x)
            if (PP->y==P3.y)
                if (PP->z==P3.z) {

                    I3 = I;
                }
    }

    MatchedAll = ( (I1 != -1) && (I2 != -1) && (I3 != -1) );

    if ( I1 == -1 )
        I1 = appendVertex (PC1);

    if ( I2 == -1 ) {

        if ( ptsEqual (P2, P1) )
            I2 = I1;
        else
            I2 = appendVertex (PC2);
    }

    if ( I3 == -1 ) {

        if ( ptsEqual (P3, P1) )
            I3 = I1;
        else {
            if ( ptsEqual (P3, P2) )
                I3 = I2;
            else
                I3 = appendVertex (PC3);
        }
    }
}

long TRISET_C::appendFace (long I1, long I2, long I3)

// Erzeugt ein Face mit den Eckpunkten I1,I2,I3 (Hidden Vertex Liste ist leer).
// Liefert den Index dieses Faces zurck.

{
    // Neuen Eintrag im FaceArray anlegen:

    int N = int (FaceCnt/FacesPerArray);

    if (FaceArrayList[N]==NULL) {      // neuen Block anlegen

        FaceArrayList[N]= (FACE_S*) calloc (FacesPerArray, sizeof(FACE_S));

//      Global.debugMsg ("Face Array #%i (%p): %li bytes\n", N, FaceArrayList[N],
//                       (long)FacesPerArray*sizeof(FACE_S) );

        if (!FaceArrayList[N]) {

            Global.fatal ("TRISET::appendFace: out of mem");
            return 0L;
        }
    }

    // Indizes der Eckvertizes bertragen:

    FACE_S *FP  = MacroGetFaceP (FaceCnt);
    long   *FIP = &(FP->Index[0]);

    FIP[0] = I1; FIP[1] = I2; FIP[2] = I3;

    // Hidden-Vertexliste initialisieren:

    FP->N    = 0;
    FP->MaxN = InitialIndexArraySize;
    FP->HVP  = NULL;

    if (FP->MaxN>0) {

        FP->HVP  = (long*) calloc (FP->MaxN, sizeof (long));

        if (!FP->HVP) {
            Global.fatal ("TRISET::appendFace: out of mem");
            return 0L;
        }
    }

    FaceCnt++;
    ValidFaceCnt++;

    updateStatus();

    return FaceCnt - 1;           // Index des neuen Faces zurckgeben
}


BOOL TRISET_C::addTriangle (Pptco3 &V1, Pptco3 &V2, Pptco3 &V3)
{
    long      I1, I2, I3;
    BOOL      MatchedAll;

        if (VertHashSize == 0L) {

            add3Vertices (V1, V2, V3, I1, I2, I3, MatchedAll, StartSearch);
            }
        else {

            I1 = addVertexHash (V1);
            I2 = addVertexHash (V2);
            I3 = addVertexHash (V3);
        }

        // connectIntervall() liefert z.T. entartete Dreiecke (zwei der Eck-
        // punkte sind identisch), wenn die beiden Polygonzge identische Punkte
        // haben !

        if ( (I1==I2) || (I1==I3) || (I2==I3) )
            return FALSE;

        appendFace (I1, I2, I3);

        // Statistik updaten;

        if ( ! (FaceCnt % 100))
            updateStatus();

    return TRUE;
}

BOOL TRISET_C::processFace (Pptco3 &V1, Pptco3 &V2, Pptco3 &V3)
{
    if (MapTriangles) {

        Global.TextureP->mapTriangle (V1, V2, V3);
    }

    if (StoreTriangles) {

        return addTriangle (V1, V2, V3);
    }

    return TRUE;
}

/*------------------------------------------------------------------------+
|                                                                         |
| Tesselierung von zwei Polylines                                         |
|                                                                         |
+------------------------------------------------------------------------*/

int TRISET_C::nearestPlVertex (Ppoint3 &RefVert, VERTEXARRAY_C &Pl, int Start, int End)
{
    int    NearestVert = -1;
    Pfloat MinDist     = MAXFLOAT;

    for (int I=Start; I<=End; I++) {

        Pfloat D = distance3 (RefVert, Pl.Val[I].point);

        if (D<MinDist) {

            MinDist     = D;
            NearestVert = I;
        }
    }

    return NearestVert;
}

BOOL TRISET_C::divideQuad (Pptco3 &A1, Pptco3 &A2, Pptco3 &B1, Pptco3 &B2)

// Erzeugt zwei Dreiecke, aus einem Viereck, indem entweder A1 und B2 oder
// A2 unt B1 verbunden werden (krzere Distanz wird gewhlt).
// Dabei wird auf eine gleichmige Orientierung der Dreiecke geachtet.

{
    if ( sqrDistance3 (A1.point, B2.point) < sqrDistance3 (A2.point, B1.point) ) {

        if (!Swapped) {
            processFace (A1, B2, B1);
            processFace (A1, A2, B2);
            }
        else {
            processFace (A1, B1, B2);
            processFace (A1, B2, A2);
        }
        }
    else {
        if (!Swapped) {
            processFace (A1, A2, B1);
            processFace (A2, B2, B1);
            }
        else {
            processFace (A1, B1, A2);
            processFace (A2, B1, B2);
        }
    }
    return TRUE;
}

static int  Depth    = 0;
static int  MaxDepth = 0;

void TRISET_C::connectIntervall (VERTEXARRAY_C&Pl1, int A1, int A2,
                                 VERTEXARRAY_C&Pl2, int B1, int B2)
{
    int I;

    Depth++;
    if (Depth>MaxDepth)
        MaxDepth = Depth;

    if (B1==B2) {
        for (I=A1; I<A2; I++)
            if (!Swapped)
                processFace (Pl2.Val[B1], Pl1.Val[I], Pl1.Val[I+1]);
            else
                processFace (Pl2.Val[B1], Pl1.Val[I+1], Pl1.Val[I]);
        Depth--;
        return;
    }
    if (A1==A2) {
        for (I=B1; I<B2; I++)
            if (!Swapped)
                processFace (Pl1.Val[A1], Pl2.Val[I+1], Pl2.Val[I]);
            else
                processFace (Pl1.Val[A1], Pl2.Val[I], Pl2.Val[I+1]);
        Depth--;
        return;
    }
    if ( (A2==A1+1) && (B2==B1+1) ) {
        divideQuad (Pl1.Val[A1], Pl1.Val[A2], Pl2.Val[B1], Pl2.Val[B2]);
        Depth--;
        return;
    }

    if ( (A2==A1+1) && (B2!=B1+1) ) {

        Swapped = Swapped ? FALSE : TRUE;        // 'Swapped' toggeln

        connectIntervall (Pl2, B1, B2, Pl1, A1, A2);

        Swapped = Swapped ? FALSE : TRUE;        // 'Swapped' toggeln

        Depth--;
        return;
    }

    int AMitte = (A1 + A2) / 2;        // Integerdivision !

//  int BMitte = (B1 + B2) / 2;

    int BMitte = nearestPlVertex (Pl1.Val[AMitte].point,
                                  Pl2, B1, B2);

    connectIntervall (Pl1, A1, AMitte, Pl2, B1, BMitte);
    connectIntervall (Pl1, AMitte, A2, Pl2, BMitte, B2);
    Depth--;
}

BOOL TRISET_C::addStripToTriset (VERTEXARRAY_C &Pl1, VERTEXARRAY_C &Pl2)

// Der Mittlere Punkt der ersten PL wird mit dem am nchsten liegenden Punkt
// der Zweiten verbunden. Rekursiver Aufruf fr  beide entstandenen Intervalle
// der ersten PL.

{
    char Str [100];

    strcpy (Str, "");

    if (MapTriangles)
        strcat (Str, "mapping colors ... ");
    if (StoreTriangles)
        strcat (Str, "creating faces ...");

    Global.StatWinP->status (Str);
    int N1 = Pl1.nElem();
    int N2 = Pl2.nElem();
    if ( (N1 < 2) || (N2 < 2)) return FALSE;

    // StartSearch gibt den Vertex-Index an, ab dem die neuen Vertizes auf
    // Gleichheit berprft werden sollen.
    // Je kleiner (VertexCnt-StartSearch), desto schneller der Algorithmus.
    // ACHTUNG: theoretisch ist es immmer mglich, da ein Punkt eines spter
    //          eingelesenen Polygonzugs auf einem bereits gelesenen liegt.

    StartSearch = MAX (VertexCnt - N1 - N2, 0);  // schnell
//  StartSearch = 0L;                            // auf Nummer sicher

    connectIntervall (Pl1, 0, N1-1, Pl2, 0, N2-1);

    Global.StatWinP->status ("");

    updateStatus();
    return TRUE;
}

/*------------------------------------------------------------------------+
|                                                                         |
| Tesselierung des gesamten Trisets                                       |
|                                                                         |
+------------------------------------------------------------------------*/

BOOL TRISET_C::getNeighbours (long         Center,     // <-  center vertex
                              INDEXARRAY_C *PGP,       //  -> Nachbarpolygon
                              INDEXARRAY_C *FAP,       //  -> Nachbarfaces
                              BOOL         &IsClosed   //  -> PG geschlossen ?
                             )

// Sucht alle Nachbarpunkte von <Center>. Die Punkte werden so sortiert, da
// ein Polygonzug entsteht. Die Punkte werden nach *PGP kopiert. Die Indizes
// der Nachbardreiecke werden in *FAP zurckgegeben.
// <IsClosed> wird auf FALSE gesetzt, wenn der Polygonzug offen ist. Daraus kann
// geschlossen werden, da <Center> ein Rand- oder Eckpunkt ist (der erste und
// letzte Nachbar in der List sind dann ebenfalls Randpunkte).
// Im Fehlerfall (z.B. Center ist nicht referenziert), wird FALSE zurckgegeben.

{
    long *F;

    const MaxNbFaces = 50;

    long B [MaxNbFaces];               // I Faces: Center, B[I], C[I]; I=0..N-1
    long C [MaxNbFaces];

    int NbFaceCnt = 0;                 // Anzahl der benachbarten Flchen

    PGP->clear();
    FAP->clear();

    // Vorauswahl:

    if (Center==0L) return FALSE;      // 0 macht Schwierigkeiten (auch Lschmarkierung)

    // alle Nachbardreiecke suchen. Indizes in *FAP ablegen. Die beiden Punkte
    // der Dreiecke, die ungleich <Center> sind, sind Nachbarpunkte.

    long     CurFaceIdx = -1;
    FACE_S  *CurFAP     = FaceArrayList[0];
    int      CurBlock   = 0;
    int      MaxBlock   = int (FaceCnt >> Log2FacesPerArray);
    long     RestLen    = ( CurBlock==MaxBlock ) ?
                            (FaceCnt & FaceOffsMask) : FacesPerArray ;

    do {
        // scanFaceArray (StartFaceP, Cnt, FoundIdx, VertIdx) sucht ab <StartFaceP>
        // die nchsten <Cnt> Faces nach dem Vertex <VertIdx> ab.
        // Wenn gefunden, ist <Cnt>        = Cnt - FoundIdx + 1,
        //                    <FoundIdx>   = Index des gefundenen Faces,
        //                    <StartFaceP> zeigt auf den nchsten Wert.

        if ( scanFaceArray ( &CurFAP, &RestLen, &CurFaceIdx, Center ) ) {

                F = MacroGetFaceIdxP (CurFaceIdx);

                if (F[0]==Center) {
                    B[NbFaceCnt]   = F[1];
                    C[NbFaceCnt++] = F[2];
                    FAP->append (CurFaceIdx);
                    }
                else if (F[1]==Center) {
                    B[NbFaceCnt]   = F[2];
                    C[NbFaceCnt++] = F[0];
                    FAP->append (CurFaceIdx);
                    }
                else if (F[2]==Center) {
                    B[NbFaceCnt]   = F[0];
                    C[NbFaceCnt++] = F[1];
                    FAP->append (CurFaceIdx);
                }
            }
        else {

            if ( CurBlock == MaxBlock ) {

                break;                 // fertig
                }
            else {                     // nchstes Array

                CurBlock++;
                CurFAP   = FaceArrayList[CurBlock];
                RestLen  = ( CurBlock==MaxBlock ) ?
                            (FaceCnt & FaceOffsMask) : FacesPerArray ;
            }
        }

        if (NbFaceCnt >= (MaxNbFaces-1) ) {
            Global.debugInfo ("TRISET_C::getNeighbours: Face buffer too small");
            return FALSE;
        }

    } while (TRUE);

    if (NbFaceCnt==0) return FALSE;        // Punkt war nicht referenziert (gelscht !)

    // wenn zwei der Nachbarfaces so hnlich sind, da sie (durch Rundungs-
    // fehler) die gleichen Eckpunkte haben, gibt es Schwierigkeiten. Wenn
    // dieser Fall eintritt, wird die Untersuchung abgebrochen

    long *FIdxP = FAP->Val;
    int  FCnt   = FAP->nElem();

    for (int K=0; K<FCnt-1; K++) {

        long *F1P = MacroGetFaceIdxP ( FIdxP[K] );

        for (int J=K+1; J<FCnt; J++) {

            long *F2P = MacroGetFaceIdxP (FIdxP [J]);

            if ( (F1P[0]==F2P[0]) || (F1P[1]==F2P[0]) || (F1P[2]==F2P[0]) )
                if ( (F1P[0]==F2P[1]) || (F1P[1]==F2P[1]) || (F1P[2]==F2P[1]) )
                    if ( (F1P[0]==F2P[2]) || (F1P[1]==F2P[2]) || (F1P[2]==F2P[2]) ) {

                        // bei AvoidDoubles solte das eigentlich nicht passieren:

                        if (Global.OptionsP->AvoidDoubles && Global.OptionsP->EnableInfos) {

                            Global.debugMsg
                               ("getNeighbours: Zwei Faces identisch (%li, %li)\n",
                                FIdxP[K], FIdxP[J]);
                            dumpFaceInfo (FIdxP[K]);
                            dumpFaceInfo (FIdxP[J]);
                        }

                        // Ecken Grn markieren:

                        if (Global.OptionsP->MarkFaces) {

                            markFace (FIdxP[K], 0, 255, 0);
                        }

                        return FALSE;
                    }

        }
    }

    // Randpunkte sortieren, damit ein geschlossenes Polygon entsteht:

    long Last  = B[0];
    long Next  = C[0];
    long First = Last;

    PGP->append (Last);

    BOOL Done  = FALSE;
    BOOL Found;

    do {
        Found = FALSE;

        for (int I=0; I<NbFaceCnt; I++) {

            if ( (B[I]==Next) && (C[I]!=Last) ) {

                Last  = Next;
                PGP->append (Last);
                Next  = C[I];
                Found = TRUE;
                break;
                }
            else if ( (C[I]==Next) && (B[I]!=Last) ) {

                Last  = Next;
                PGP->append (Last);
                Next  = B[I];
                Found = TRUE;
                break;
            }
        }

        if ( PGP->nElem() > NbFaceCnt + 1) {

            Global.debugInfo ("getNeighbours: PGP.N (%i) > NBFaceCnt", PGP->nElem());

            // Grn markieren:

            if (Global.OptionsP->MarkFaces) {

                Pcobundl *VCP = & (getVertPtCoP (Center)->colour);
                VCP->x = VCP->z = 0.0; VCP->y = 1.0;
            }

            //Global.CancelRequest = TRUE;
            return FALSE;
        }

        if ( (Next==First) || !Found )           // Polygon geschlossen oder Fehler
            Done = TRUE;

    } while (!Done);

    // redundante Plausibilittskontrolle

    if ( (Found != (Next==First)) || (Found  && (NbFaceCnt != PGP->nElem()) ) ) {

        int TempI;

        Global.debugInfo ("getNeighbours: was faul ! (Center: %li, Found=%i, Next=%li, First=%li, Faces:%i, Verts=%i)",
                      Center, Found, Next, First, NbFaceCnt, PGP->nElem());

        char *InfoStr = (char*)malloc (1000);
        char *TempStr = (char*)malloc (50);

        strcpy (InfoStr, "Faces:\n");

        for (TempI=0; TempI<NbFaceCnt; TempI++) {

            if (TempI>15) break;

            sprintf (TempStr, "%li: %li, %li\n", TempI, B[TempI], C[TempI]);
            strcat (InfoStr, TempStr);
        }
        Global.debugInfo  (InfoStr);

        Ppoint3 *PP = getVertPtP(Center);

        sprintf (InfoStr, "Neighbours of %li (%f,%f,%f):\n", Center, PP->x, PP->y, PP->z);

        for (TempI=0; TempI<PGP->nElem(); TempI++) {

            if (TempI>15) break;

            PP = getVertPtP ( PGP->Val[TempI] );

            sprintf (TempStr, "%li: %f,%f,%f\n", PGP->Val[TempI], PP->x, PP->y, PP->z);
            strcat  (InfoStr, TempStr);
        }
        Global.debugInfo  (InfoStr);

        // Grn markieren:

        if (Global.OptionsP->MarkFaces) {

            getVertPtCoP (Center)->colour.x = 0.0;
            getVertPtCoP (Center)->colour.y = 1.0;
            getVertPtCoP (Center)->colour.z = 0.0;
        }
        //Global.CancelRequest = TRUE;

        free (InfoStr);
        free (TempStr);

        return FALSE;
    }

    // Wenn das Polygon nicht geschlossen ist, kann es sein, da noch nicht
    // alle Punkte gefunden wurden. Es mu von <First> in die andere Richtung
    // gesucht werden:

    if (Next!=First) {     // offenes Polygon ?

        IsClosed = FALSE;

        PGP->append (Next);

        if ( PGP->nElem() <= NbFaceCnt ) { // noch nicht alle Punkte gefunden ?

            Last = C[0];               // andere Suchrichtung
            Next = B[0];

            do {
                Found = FALSE;
                Done  = FALSE;

                for (int I=0; I<NbFaceCnt; I++) {

                    if ( (B[I]==Next) && (C[I]!=Last) ) {

                        Last  = Next;
                        Next  = C[I];
                        PGP->insertFirst (Next);
                        Found = TRUE;
                        break;         // end for
                        }
                    else if ( (C[I]==Next) && (B[I]!=Last) ) {

                        Last  = Next;
                        Next  = B[I];
                        PGP->insertFirst (Next);
                        Found = TRUE;
                        break;         // end for
                    }
                }

                if ( PGP->nElem() > NbFaceCnt+1) {

                    Global.debugInfo ("getNeighbours: PGP (%i) > NBFaceCnt", PGP->nElem());
                    return FALSE;
                }

            } while ( Found );
        }
        } // endif (Next!=First)
    else
        IsClosed = TRUE;

    return TRUE;
}

BOOL TRISET_C::replacePatch (INDEXARRAY_C *OldFaceIdxP,
                             FACEARRAY_C  *NewFaceP)

// Ersetzt einen Teil des Trisets (N Faces) durch ein neues (N-2 Faces)
// NewFaceP enthlt die neuen Faces mit den Hidden Vertizes. Diese Daten
// werden in das Triset kopiert.
// Die Hidden Vertices mssen vorher iauf die neuen Faces verteilt worden sein!

{
    int I, Nnew, Nold;

    // alte Faces lschen:

    Nold = OldFaceIdxP->nElem();

    for (I=0; I<Nold; I++) {

        delFace (OldFaceIdxP->Val[I]);
    }

    // neue Faces anfgen, (auch hidden vertices kopieren):

    Nnew = NewFaceP->nElem();

    for (I=0; I<Nnew; I++) {

        FACE_S *SourceFP   = &(NewFaceP->Val[I]);
        long   *SourceIdxP = SourceFP->Index;
        long   *SourceHVP  = SourceFP->HVP;

        long   DestIdx     = appendFace (SourceIdxP[0], SourceIdxP[1], SourceIdxP[2]);

        if (!SourceHVP)
            continue;                  // dieses Face hat keine Hidden Verts

        int nHV = SourceFP->N;

        for (int HV=0; HV<nHV; HV++) {

            addHV (DestIdx, SourceHVP [HV]);
        }
    }

    if (Nold<Nnew)
        VisibleVertexCnt++;
    else if (Nold>Nnew)
        VisibleVertexCnt--;

    updateStatus();

    return TRUE;
}

BOOL TRISET_C::write()
{
    if ( ! Global.OptionsP->WriteGeometry ) return TRUE;

    if ( ! Global.GeomOutP->open (Global.OptionsP->GeomFile,
                                   Global.OptionsP->GeomSmall) ) {

        Global.warning ("Can't open '%s'.", Global.OptionsP->GeomFile);
        return FALSE;
    }

    if ( !Global.GeomOutP->writeTriset (Global.TrisetP) )
        return FALSE;

    Global.GeomOutP->close();

//  if (fileExist (BackupFSpec) remove (BackupFSpec);

    return TRUE;
}

BOOL TRISET_C::backup()
{
    REDUCE_C *RP = Global.ReduceP;

    if (!Global.StatWinP->enableBackup())
        return TRUE;

    if ( RP->CurEnergy > RP->SavedEnergy ) {

        char Str[100];

        sprintf (Str, "Saved energy: %f, current energy: %f\nOverwrite better result ?",
                      RP->SavedEnergy, RP->CurEnergy);

        int Answer = MessageBox (NULL, Str, "CONFIRM",
                                 MB_YESNO | MB_ICONQUESTION
                                );
        if (Answer!=IDYES) return TRUE;
    }

    BOOL Stat = TRUE;

    Global.StatWinP->enableBackup (FALSE);

    Global.debugMsg ("Backup: Nv: %li, Nf: %li\n", VisibleVertexCnt, ValidFaceCnt );
    Global.logMsg   ("Backup: Nv: %li, Nf: %li\n", VisibleVertexCnt, ValidFaceCnt );

    if (VertexCnt>0) {

        echo (FALSE);
        Stat = write();
        echo (TRUE);
    }

    if (Stat)
        RP->SavedEnergy = RP->CurEnergy;

    Global.StatWinP->enableBackup (TRUE);

    return Stat;
/*
    if ( ! Global.OptionsP->WriteGeometry ) return TRUE;

    char Drive[3], Dir[66], Name[9], Ext[5];

    fnsplit (Global.OptionsP->GeomFile, Drive, Dir, Name, Ext);
    fnmerge (BackupFSpec, Drive, Dir, "$comp3d$", Ext);

    if ( ! Global.GeomOutP->open (BackupFSpec,
                                  Global.OptionsP->GeomSmall) ) {

        return FALSE;
    }

    if ( !Global.GeomOutP->writeTriset (Global.TrisetP) )
        return FALSE;

    Global.GeomOutP->close();

    return TRUE;
*/
}

void TRISET_C::dumpVertexInfo (long VI)
{
    VERTEX_T *VP  = MacroGetVertP (VI);

    Global.debugMsg ("Vertex #%7li (%p): FI=%7li\n", VI, VP, VP->FI);
    Global.debugMsg ("  xyz: (%f, %f, %f), rgb: (%f, %f, %f)\n",
                     VP->PC.point.x, VP->PC.point.y, VP->PC.point.z,
                     VP->PC.colour.x, VP->PC.colour.y, VP->PC.colour.z
                    );
}

void TRISET_C::dumpFaceInfo (long FI)
{
    FACE_S   *FP  = MacroGetFaceP (FI);
    long     *FIP = FP->Index;

    Global.debugMsg ("Face #%li (%p): HV-list: %p, N: %i, MaxN:%i;  A=%f\n",
                     FI, FP, FP->HVP, FP->N, FP->MaxN,
                     areaTri3 (*getVertPtP (FIP[0]), *getVertPtP (FIP[1]), *getVertPtP (FIP[2]))
                    );

    dumpVertexInfo (FIP[0]);
    dumpVertexInfo (FIP[1]);
    dumpVertexInfo (FIP[2]);

    Global.debugMsg ("HV list: ");

    for (int I=0; I<FP->N; I++) Global.debugMsg ("%7i,", FP->HVP[I]);

    Global.debugMsg ("\n\n");
}

BOOL TRISET_C::isEqualOrientated()

// Prfen, ob alle Dreiecke gleichmig orientiert sind
// d.h. es darf keine Edge zweimal mit der gleiche Orientierung auftreten.

{
    BOOL SaveCancelRequest = Global.CancelRequest;
    Global.CancelRequest   = FALSE;

    compactFaces();

    long DoubleEdgeCnt = 0L;

    for (long F=0; F<FaceCnt-1; F++) {

        if (F % 100 == 0) {
            Global.StatWinP->status ("checking for double edges ... %li", F);
            Global.StatWinP->updateDlg();
            if (Global.CancelRequest) {

                Global.debugMsg ("scan for double edges: ***aborted***\n");
                goto EarlyOut;
            }
        }

        long *FIP = MacroGetFaceIdxP (F);

        long Idx = findDoubleEdge (FIP[0], FIP[1], FIP[2], F+1);

        if ( Idx != NOINDEX ) {

            DoubleEdgeCnt++;
            if (Global.OptionsP->EnableInfos) {

                Global.debugMsg ("found double edges: %li, %li\n", F, Idx);
                dumpFaceInfo (F);
                dumpFaceInfo (Idx);
            }
            if (Global.OptionsP->MarkFaces) {

                markFace (Idx, 255, 0, 255);
                markFace (F, 255, 0, 255);
            }
        }
    }

EarlyOut:
    Global.CancelRequest = SaveCancelRequest;
    Global.StatWinP->updateDlg();

    if (DoubleEdgeCnt>0)
        Global.debugMsg ("found %li double edges\n", DoubleEdgeCnt);
    else
        Global.debugMsg ("No double edges found!\n");

    Global.StatWinP->status ("");

    return (DoubleEdgeCnt==0);
}

BOOL TRISET_C::integrityCheck()
{
    long   F;
    FACE_S *FP;
    long   *FIP;

    BOOL WasFaul = FALSE;

    long ActVisibleVerts  = 0;
    long ActDelVerts      = 0;
    long ActValidFaces    = 0;
    int  MaxDelVerts      = 0;

    Global.StatWinP->status ("checking triset integrity ...");
    Global.debugMsg ("integrityCheck():\n");

    for (F=0; F<FaceCnt; F++) {

        FP  = MacroGetFaceP (F);
        FIP = FP->Index;

        if ( isFaceDeleted (F) )
            continue;

        ActValidFaces++;

        if ( (MacroGetVertP (FIP[0])->FI != 0) ||
             (MacroGetVertP (FIP[1])->FI != 0) ||
             (MacroGetVertP (FIP[2])->FI != 0) )
            {
            Global.debugMsg ("\tFace(%li) Edge Vertex is marked as hidden !\n", F);
            WasFaul = TRUE;
        }
        if (FP->HVP) {

            if (FP->N>MaxDelVerts) MaxDelVerts = FP->N;

            for (int HV=0; HV<FP->N; HV++) {

                ActDelVerts++;

                if ( MacroGetVertP (FP->HVP[HV])->FI != F ) {
                    Global.debugMsg
                        ("\tFace(%li): Hidden Vertex (%li) is marked as son of Face %li !\n",
                         F, FP->HVP[HV], MacroGetVertP (FP->HVP[HV])->FI);
                    WasFaul = TRUE;
                }
            }
        }
    }
    ActVisibleVerts = VertexCnt - ActDelVerts;

    Global.debugMsg ("\tVisible verts: %li, found: %li\n",
                     VisibleVertexCnt, ActVisibleVerts);
    Global.debugMsg ("\tMax HiddenVerts per Face: %i\n", MaxDelVerts);
    Global.debugMsg ("\tValid faces: %li, found: %li\n",
                     ValidFaceCnt, ActValidFaces);

    if ( (VisibleVertexCnt != ActVisibleVerts) ||
         (ValidFaceCnt !=  ActValidFaces) )
        WasFaul = TRUE;

    if (Global.OptionsP->CheckDoubleEdges)
        if ( !isEqualOrientated() )
            WasFaul = TRUE;

    Global.StatWinP->status ("");

    if (WasFaul) Global.beep();

    return !WasFaul;
}
