/**************************************************************************
*                                                                         *
* Name   :                             ( Headerdatei:                 )   *
*                                                                         *
* Klasse(n) definiert : ZOFREAD_C                                         *
* Objekt(e) deklariert:                                                   *
*                                                                         *
* Abgeleitet von      :                                                   *
*                                                                         *
* Aufgabe:                                                                *
*                                                                         *
**************************************************************************/

// --- Includes: ------------------------------------------------------

#include "cTypes.hpp"

#include <string.h>

#include "cGlobal.hpp"
#include "cZofRead.hpp"
#include "cMath.hpp"
#include "cOptions.hpp"
#include "cStatWin.hpp"
#include "cTools.hpp"
#include "cTriset.hpp"

// --- lokale Konstanten:----------------------------------------------

// --- lokale Typen: --------------------------------------------------

// --- lokale Variablen: ----------------------------------------------


/*------------------------------------------------------------------------+
|                                                                         |
| lokale Funktionen                                                       |
|                                                                         |
+------------------------------------------------------------------------*/

/*------------------------------------------------------------------------+
|                                                                         |
| Methoden der Klasse ZOFREAD_C                                           |
|                                                                         |
+------------------------------------------------------------------------*/

ZOFREAD_C::ZOFREAD_C()             // Konstruktor
{
    InFileP = NULL;

    DefColour.x = 0.8;
    DefColour.y = 0.8;
    DefColour.z = 0.8;
}

ZOFREAD_C::~ZOFREAD_C()            // Destruktor
{
    close();
}

BOOL ZOFREAD_C::open (char *FSpec)
{
    close();

    Line       = 0;
    FileSize   = 0;
    BytesRead  = 0;

    InFileP = fopen (FSpec, "rt");     // read text file

    if (InFileP) FileSize = fileSize (InFileP);

    return (InFileP!=NULL);
}

void ZOFREAD_C::close()
{
    if (!InFileP) return;

    fclose (InFileP);
}

void ZOFREAD_C::warning()
{
    Global.warning ("Read ZOF: illegal format in line %li", Line);
}

BOOL ZOFREAD_C::readLine (BOOL Warning)

// holt die nchste Zeile, die keine Leerzeile ist
// liefert FALSE, wenn Eof oder Fehler

{
    BOOL Stat;
    int  Pos;

    if (!InFileP) return FALSE;

    do {
        Stat = (fgets (Str, sizeof (Str), InFileP) != NULL);
        Line++;

        if (Stat) {

            BytesRead+=strlen (Str);
            Pos = -1;
            do Pos++; while (Str[Pos]=='\t' || Str[Pos]==' ');
        }

        if (Line%200 == 0) {

            Global.StatWinP->status ("reading ZOF ... (line %li)", Line);
            Global.StatWinP->Percentage = 100.0*BytesRead/FileSize;
            Global.StatWinP->updateDlg();
        }

    } while ( Stat && (Str[Pos]=='\n') ); // Leerzeilen berlesen

    if (!Stat && Warning)
        Global.warning ("Read ZOF: Unexpected EoF in line #%li", Line);

    return Stat;
}

BOOL ZOFREAD_C::searchTriset   (long &NVerts,   long &NFaces,
                                int  &VertFlag, int  &FaceFlag)
{
    if (!InFileP) return FALSE;

    while ( readLine() ) {

        if (strstr (strupr (Str), "TRISET3")) {

            sscanf (Str, "%*s %i %i", &FaceFlag, &VertFlag);

            if ( !readLine (TRUE) ) return FALSE;

            if (sscanf (Str, "%li %li", &NFaces, &NVerts)!=2) {

                warning();
                return FALSE;
            }
            return TRUE;
        }
    }

    // EoF  oder Fehler:

    return FALSE;
}

BOOL ZOFREAD_C::read (char *FSpec, TRISET_C *TP)
{
    long      L;
    long      VertCnt;
    long      FaceCnt;
    int       VertFlag;
    int       FaceFlag;
    Pptco3    V;
    long      I1, I2, I3;
    float     DummyF;
    int       N;

    BOOL      Status       = TRUE;
    int       SetCnt       = 0;
    long      TotalVertCnt = 0;
    long      TotalFaceCnt = 0;
    OPTIONS_C *OP          = Global.OptionsP;

    if ( !open (FSpec) ) {

        Global.warning ("Can't open '%s'", FSpec);
        return FALSE;
    }

    TRISET_C  *TempTP = new TRISET_C;
    if ( !TempTP ) return FALSE;

    TP->clear();
    TP->createVertHashTab (Global.OptionsP->VertHashTabSize);

    while ( searchTriset (VertCnt, FaceCnt, VertFlag, FaceFlag) ) {

	if ( (FaceFlag!=PFA_NONE) ||
             ( (VertFlag!=PVERT_NONE) && (VertFlag!=PVERT_COLOUR) &&
	       (VertFlag!=PVERT_NORMAL) && (VertFlag!=PVERT_DATA) &&
	       (VertFlag!=PVERT_COLOURDATA) && (VertFlag!=PVERT_NORMALDATA) &&
	       (VertFlag!=PVERT_COLOURNORMAL) && (VertFlag!=PVERT_COLOURNORMALDATA) ) ) {

            Global.warning ("Triset format not supported !\n Line: %li, vf:%i, ff:%i",
                            Line, VertFlag, FaceFlag);
            continue; // nchstes versuchen
        }

        // Defaultfarbe setzen:

        V.colour.x = DefColour.x;
        V.colour.y = DefColour.y;
        V.colour.z = DefColour.z;

        SetCnt++;

        // ------------------------------------------------
        // Verts lesen und in temporrem Triset ablegen
        // ------------------------------------------------

        TempTP->clear();

        for (L=0; L<VertCnt; L++) {

            if ( !readLine (TRUE) || Global.CancelRequest ) {

                Status = FALSE;
                goto ErrorExit;
            }

            switch (VertFlag) {

                case PVERT_NONE:
                    N = sscanf (Str, "%f %f %f",
                                &V.point.x,  &V.point.y,  &V.point.z);
                    if (N!=3) Status=FALSE;
                    break;

                case PVERT_COLOUR:
                    N = sscanf (Str, "%f %f %f %f %f %f",
                                &V.point.x,  &V.point.y,  &V.point.z,
                                &V.colour.x, &V.colour.y, &V.colour.z);
                    if (N!=6) Status=FALSE;
                    break;

		case PVERT_NORMAL:
                    N = sscanf (Str, "%f %f %f %f %f %f",
                                &V.point.x,  &V.point.y,  &V.point.z,
				&DummyF, &DummyF, &DummyF);
                    if (N!=6) Status=FALSE;
                    break;

		case PVERT_DATA:
                    N = sscanf (Str, "%f %f %f %f %f",
                                &V.point.x,  &V.point.y,  &V.point.z,
                                &DummyF, &DummyF);
                    if (N!=5) Status=FALSE;
                    break;

		case PVERT_COLOURDATA:
                    N = sscanf (Str, "%f %f %f %f %f %f %f %f",
                                &V.point.x,  &V.point.y,  &V.point.z,
                                &V.colour.x, &V.colour.y, &V.colour.z,
				&DummyF, &DummyF);
		    if (N!=8) Status=FALSE;
                    break;

		case PVERT_NORMALDATA:
                    N = sscanf (Str, "%f %f %f %f %f %f %f %f",
                                &V.point.x,  &V.point.y,  &V.point.z,
				&DummyF, &DummyF, &DummyF,
				&DummyF, &DummyF);
		    if (N!=8) Status=FALSE;
                    break;

                case PVERT_COLOURNORMAL:
                    N = sscanf (Str, "%f %f %f %f %f %f %f %f %f",
                                &V.point.x,  &V.point.y,  &V.point.z,
                                &V.colour.x, &V.colour.y, &V.colour.z,
				&DummyF, &DummyF, &DummyF);
		    if (N!=9) Status=FALSE;
                    break;

		case PVERT_COLOURNORMALDATA:
                    N = sscanf (Str, "%f %f %f %f %f %f %f %f %f %f %f",
                                &V.point.x,  &V.point.y,  &V.point.z,
				&V.colour.x, &V.colour.y, &V.colour.z,
				&DummyF, &DummyF, &DummyF,
				&DummyF, &DummyF);
                    if (N!=11) Status=FALSE;
                    break;

	    }

            if ( Status ) {

                if (Global.OptionsP->Filtering) {

                    V.colour.x = CLAMP (V.colour.x * OP->FactR, 0.0, 1.0);
                    V.colour.y = CLAMP (V.colour.y * OP->FactG, 0.0, 1.0);
                    V.colour.z = CLAMP (V.colour.z * OP->FactB, 0.0, 1.0);

                    // Y-clipping wird nicht durchgefhrt
                }

                TempTP->appendVertex (V);
                }
            else {

                warning();
                goto ErrorExit;
            }

        }

        TotalVertCnt += VertCnt;
        Global.StatWinP->InVerts    = TotalVertCnt;
        Global.StatWinP->InPls      = 0L;

        // ------------------------------------------------
        // Faces lesen und im Triset speichern
        // ------------------------------------------------

        TP->StoreTriangles = TRUE;
        TP->MapTriangles   = Global.OptionsP->TextureMapping;

        for (L=0; L<FaceCnt; L++) {

            if ( !readLine (TRUE) || Global.CancelRequest ) {

                Status = FALSE;
                goto ErrorExit;
            }

            N = sscanf (Str, "%li %li %li\n", &I1, &I2, &I3);

            if ( (N!=3) || (I1<=0) || (I2<=0) || (I3<=0) ||
                           (I1>VertCnt) || (I2>VertCnt) || (I3>VertCnt) ) {

                warning();
                Status=FALSE;
                goto ErrorExit;
            }

            TP->processFace (*TempTP->getVertPtCoP (I1-1),
                             *TempTP->getVertPtCoP (I2-1),
                             *TempTP->getVertPtCoP (I3-1) );
        }
        TotalFaceCnt += FaceCnt;
    }

ErrorExit:

    delete TempTP;

    TP->destroyVertHashTab();

    if (Status && Global.OptionsP->LogSettings)
        Global.logMsg ("\nZofRead\t: %li lines, %li bytes, %i sets, %li faces, %li verts\n",
                       Line-1, BytesRead, SetCnt, TotalFaceCnt, TotalVertCnt);

    return Status;
}
